import os, math, sys, getopt, string
import sympy
import copy
import glypher.Word as Word
from glypher.Widget import GlyphEntry, GlyphResponder
import random
from aobject.utils import debug_print
from aobject.paths import get_user_home
from gi.repository import Gdk, Gtk, GObject
import cairo
import threading
import numpy, numpy.fft
import scipy, scipy.interpolate, scipy.optimize
from third_party import ipython_view
from gi.repository import Pango

import pylab
from PIL import Image
import aobject.aobject as aobject

def format_obj_name (name) :
    obj = aobject.get_object_from_dictionary(name)
    return obj.get_aname_nice() + ' [' + obj.get_aname_root() + ']'

class AesPythonConsole(Gtk.Frame, aobject.AObject) :
    ipython_view = None

    def do_save(self, input_only=False) :
        chooser = Gtk.FileChooserDialog(\
                      title="Save text", action=Gtk.FileChooserAction.SAVE,
                      buttons=(Gtk.STOCK_CANCEL,Gtk.ResponseType.CANCEL,Gtk.STOCK_SAVE,Gtk.ResponseType.OK))

        chooser.set_current_folder(get_user_home())
        chooser.set_default_response(Gtk.ResponseType.OK)

        if input_only :
            chooser.set_current_name('.py')
        else :
            chooser.set_current_name('.txt')

        resp = chooser.run()
        self.grab_focus()

        if resp == Gtk.ResponseType.OK :
            filename = chooser.get_filename()
            chooser.destroy()
        else :
            chooser.destroy()
            return
        
        f = open(filename, 'w')
        if input_only :
            f.write(self.ipython_view.IP.extract_input_lines('0:-1'), raw=True)
        else :
            buff = self.ipython_view.get_buffer()
            st = buff.get_start_iter()
            en = buff.get_end_iter()

            f.write(buff.get_text(st, en))
        f.close()

        self.log(1, 'Saved to '+filename)

    def set_ui(self) :
        self.ui_action_group = Gtk.ActionGroup('ConsoleActions')
        self.ui_action_group.add_actions([('ConsoleMenu', None, 'Console'),
                                          ('ConsoleSaveText', None, 'Save text',
                                           None, None, lambda w : self.do_save(input_only=False)),
                                          ('ConsoleSaveInput', None, 'Save input',
                                           None, None, lambda w : self.do_save(input_only=True)),
                                         ])
        self.ui_ui_string = '''
            <ui>
                <menubar name="MenuBar">
                    <menu action="ConsoleMenu">
                        <menuitem action="ConsoleSaveText"/>
                        <menuitem action="ConsoleSaveInput"/>
                    </menu>
                </menubar>
            </ui>
        '''
        if ipython_view.version == '0.10' :
            self.ui_action_group.get_action('ConsoleSaveInput').set_sensitive(False)

    def __init__(self, env=None):
        GObject.GObject.__init__(self)
        self.ipython_view = ipython_view.IPythonView()
        aobject.AObject.__init__(self, "AesPythonConsole", env, view_object = True)
        self.set_aname_nice("Python Console " + (" ("+str(self.get_aname_num())+")" if self.get_aname_num()>1 else ""))
        cons_vbox = Gtk.VBox()
        V = self.ipython_view
        # with thanks to IPython cookbook
        V.modify_font(Pango.FontDescription("Mono 10"))
        V.set_wrap_mode(Gtk.WrapMode.CHAR)
        V.updateNamespace({'aobject' : aobject})
        V.updateNamespace({'AES' : aobject.get_object_from_dictionary})
        V.show()

        sw = Gtk.ScrolledWindow()
        sw.add(V)
        cons_vbox.pack_start(sw, True, True, 0)
        self.add(cons_vbox)

        self.source_action = lambda s :\
            V.write('AES(\''+s+'\')')

        views_hbox = Gtk.HBox()
        win = Gtk.VBox()
        views_hbox.pack_start(win, True, True, 0)
        win.pack_start(Gtk.Label('Objects', True, True, 0), False)

        s = Gtk.ScrolledWindow()
        win.pack_start(s, True, True, 0)

        dict_lsst = aobject.get_object_dictionary().get_liststore_by_am('aobject')
        dict_lsst.set_sort_column_id(1, Gtk.SortType.ASCENDING)

        dict_trvw = Gtk.TreeView(model=dict_lsst)

        dict_nice_crtx = Gtk.CellRendererText()
        dict_nice_tvcl = Gtk.TreeViewColumn('Nice Name', dict_nice_crtx)
        dict_nice_tvcl.add_attribute(dict_nice_crtx, 'text', 1)
        dict_nice_tvcl.set_expand(True)
        dict_trvw.append_column(dict_nice_tvcl)

        dict_name_crtx = Gtk.CellRendererText()
        dict_name_tvcl = Gtk.TreeViewColumn('Internal Name', dict_name_crtx)
        dict_name_tvcl.add_attribute(dict_name_crtx, 'text', 0)
        dict_name_tvcl.set_expand(True)
        dict_trvw.append_column(dict_name_tvcl)

        dict_trvw.connect('row-activated', lambda t, p, c : \
            self.ipython_view.write('AES(\''+\
            dict_lsst.get_value(dict_lsst.get_iter(p), 0)+'\')'))

        s.add(dict_trvw)
        s.set_size_request(-1, 200)

        win = Gtk.VBox()
        views_hbox.pack_start(win, True, True, 0)
        win.pack_start(Gtk.Label('Useful Variables', True, True, 0))

        s2 = Gtk.ScrolledWindow()
        win.pack_start(s2, True, True, 0)

        uv_lsst = aobject.get_object_dictionary().useful_vars
        uv_lsst.set_sort_column_id(0, Gtk.SortType.ASCENDING)

        uv_trvw = Gtk.TreeView(model=uv_lsst)

        uv_nice_crtx = Gtk.CellRendererText()
        uv_nice_tvcl = Gtk.TreeViewColumn('Object', uv_nice_crtx)
        #uv_nice_tvcl.add_attribute(uv_nice_crtx, 'text', 0)
        uv_nice_tvcl.set_cell_data_func(uv_nice_crtx, lambda c, r, t, i :
            uv_nice_crtx.set_property('text', \
                format_obj_name(uv_lsst.get_value(i, 0))))
        uv_nice_crtx.set_property('ellipsize', Pango.EllipsizeMode.MIDDLE)
        uv_nice_crtx.set_property('width-chars', 30)
        uv_nice_tvcl.set_expand(True)
        uv_trvw.append_column(uv_nice_tvcl)

        uv_desc_crtx = Gtk.CellRendererText()
        uv_desc_tvcl = Gtk.TreeViewColumn('Variable Description', uv_desc_crtx)
        uv_desc_tvcl.add_attribute(uv_desc_crtx, 'text', 2)
        uv_desc_tvcl.set_expand(True)
        uv_trvw.append_column(uv_desc_tvcl)

        uv_trvw.connect('row-activated', lambda t, p, c : \
            self.ipython_view.write('AES(\''+\
            uv_lsst.get_value(uv_lsst.get_iter(p), 0)+'\').'+\
            uv_lsst.get_value(uv_lsst.get_iter(p), 1)))

        s2.add(uv_trvw)
        s2.set_size_request(-1, 200)

        views_hbox.show_all()

        cons_vbox.pack_start(views_hbox, False, False, 0)
        self.show_all()

    #PROPERTIES
    def get_aesthete_properties(self):
        return { }
    #BEGIN PROPERTIES FUNCTIONS
    #END PROPERTIES FUNCTIONS
    def get_method_window(self) :
        win = Gtk.VBox()

        return win
