import os, math, sys, getopt, string
from .. import tablemaker
from gi.repository import Pango
import random
from gi.repository import Gdk, Gtk, GObject, GLib
import cairo
from ..tablemaker import PreferencesTableMaker
from aobject.paths import get_user_home, get_share_location
import threading
import matplotlib
import numpy, numpy.fft
import scipy, scipy.interpolate, scipy.optimize
from matplotlib.backends.backend_gtk3cairo import RendererGTK3Cairo as RendererCairo
import pylab
from PIL import Image
from aobject.aobject import *
from Pie import GlancerPie
from Plot import GlancerPlot
from Plot3d import GlancerPlot3D

def add_icons() :
    icon_factory = Gtk.IconFactory()

    icon_names = ('plot2d', 'pie', 'plot3d')

    for icon_name in icon_names :
        stock_id = 'aes-glancer-'+icon_name

        source = Gtk.IconSource()
        source.set_filename(
            paths.get_share_location() + 'images/icons/glancer/' + icon_name + '.svg')
        icon_set = Gtk.IconSet()
        icon_set.add_source(source)
        icon_factory.add(stock_id, icon_set)

    icon_factory.add_default()

glancer_plot_layout_signals = asignals.copy()
glancer_plot_layout_signals.update({ "draw" : "override" })
class GlancerPlotLayout(Gtk.Layout) :
    __gsignals__ = glancer_plot_layout_signals

    resize_conn = None

    # Priviledged child
    child = None

    def remove_all(self) :
        for child in self.get_children() :
            self.remove(child)

    def __init__(self) :
        GObject.GObject.__init__(self)
        self.connect_after('size-allocate', self.do_child_resize_event)
        #FIXME: horrendous hack to get resizing done. not that the size-allocate
        # seems to effect some sort of block preventing movement or resizing of
        # children during the callback. Any changes appear only at the next
        # event fire, leaving the motion consistently one iteration behind
        self.connect_after('draw', self.do_child_resize_event)
        self.surface_pattern =\
            cairo.ImageSurface.create_from_png(get_share_location()+'images/patterns/paper.png')
        self.surface_pattern = cairo.SurfacePattern(self.surface_pattern)
        self.surface_pattern.set_extend(cairo.EXTEND_REPEAT)
        #widget.connect('size-allocate', lambda w, a : self.frame.set(.5,.5))

    def add(self, child) :
        Gtk.Layout.add(self, child)
        self.child = child
        debug_print(self.get_allocation().width, self.get_allocation().height)
        child.set_size_request(self.get_allocation().width,
                               self.get_allocation().height)
        #self.resize_conn = child.connect('size-allocate', self.do_child_resize_event)
        self.do_child_resize_event(child, None)
        child.queue_draw()

    def do_child_resize_event(self, child, event) :
        debug_print_stack()
        children = self.get_children()
        if len(children) == 0 :
            return
        child = children[0]

        w = child.get_allocation().width
        h = child.get_allocation().height

        if w > 0 and h > 0 :
            wrat = w/float(self.get_allocation().width)
            hrat = h/float(self.get_allocation().height)
            rat = max(wrat, hrat)

            if abs(rat-1.) > 1e-5 and rat > 0 :
                #child.figure.set_dpi(child.figure.get_dpi()/rat)
                w = int(w/rat)
                h = int(h/rat)
                child.resize(w, h)
                child.set_size_request(w, h)

        al = self.get_allocation()
        x = max(int(.5*(al.width-w)), 0)
        y = max(int(.5*(al.height-h)), 0)

        if x != child.get_allocation().x or y != child.get_allocation().y :
            self.move(child, x, y)
            debug_print(x,y)

    def remove(self, child) :
        child.mpl_disconnect(self.resize_conn)
        Gtk.Layout.remove(self, child)

    def do_draw(self, cr) :
        ret = Gtk.Layout.do_draw(self, cr)

        cr = self.get_bin_window().cairo_create()
        cr.set_source(self.surface_pattern)
        cr.paint()

        children = self.get_children()
        if len(children) > 0 :
            child = children[0]
            al = child.get_allocation()
            cr.rectangle(al.x-5, al.y-5, al.width+10, al.height+10)
            cr.set_line_width(5.)
            cr.set_source_rgb(1.,.9,.8)
            cr.stroke()

        return ret

glancer_signals = asignals.copy()
glancer_signals.update({"draw" : "override", "loaded-sim" : (GObject.SignalFlags.RUN_LAST, None, ( GObject.TYPE_STRING, ))})
class Glancer(Gtk.Frame, AObject) :
    __gsignals__ = glancer_signals
    fig = None
    read_series = True
    can_save = True

    def savefig(self) :
        '''Save image of figure to file.'''

        chooser = Gtk.FileChooserDialog(\
                      title="Save Image", action=Gtk.FileChooserAction.SAVE,
                      buttons=(Gtk.STOCK_CANCEL,Gtk.ResponseType.CANCEL,Gtk.STOCK_SAVE,Gtk.ResponseType.OK))

        svg_ffil = Gtk.FileFilter()
        svg_ffil.set_name("Scalar Vector Graphics (SVG)")
        svg_ffil.add_mime_type("image/svg+xml")
        svg_ffil.add_pattern("*.svg")
        chooser.add_filter(svg_ffil)

        png_ffil = Gtk.FileFilter()
        png_ffil.set_name("Portable Network Graphics (PNG)")
        png_ffil.add_mime_type("image/png")
        png_ffil.add_pattern("*.png")
        chooser.add_filter(png_ffil)

        chooser.set_current_folder(get_user_home())
        chooser.set_default_response(Gtk.ResponseType.OK)
        chooser.set_current_name('.svg')
        resp = chooser.run()
        self.grab_focus()

        if resp == Gtk.ResponseType.OK :
            filename = chooser.get_filename()
            chooser.destroy()
        else :
            chooser.destroy()
            return

        if filename.endswith('.png') :
            chooser = Gtk.Dialog(\
                          title="Export PNG", parent=self.get_toplevel(), 
                          flags=Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT,
                          buttons=(Gtk.STOCK_CANCEL,Gtk.ResponseType.CANCEL,Gtk.STOCK_SAVE,Gtk.ResponseType.OK))
            chooser.set_default_response(Gtk.ResponseType.OK)

            vbox = Gtk.VBox()

            pref_tab = tablemaker.PreferencesTableMaker()

            dpi_entr = Gtk.Entry()
            pref_tab.append_row('DPI', dpi_entr)
            dpi_entr.set_text('200')
            dpi_entr.set_activates_default(True)

            vbox.pack_start(pref_tab.make_table(), True, True, 0)
            chooser.get_content_area().add(vbox)
            chooser.get_content_area().show_all()

            resp = chooser.run()
            chooser.grab_focus()

            if resp == Gtk.ResponseType.OK :
                f_dpi = float(dpi_entr.get_text())
                chooser.destroy()
            else :
                chooser.destroy()
                return

            canvas = self.fig.canvas
            w, h = self.plotter.get_size_inches()
            c_dpi = canvas.get_width_height()[0]/w
            ims = cairo.ImageSurface(cairo.FORMAT_RGB24, int(f_dpi*w), int(f_dpi*h))
            cr = cairo.Context(ims)

            cr.scale(f_dpi/c_dpi, f_dpi/c_dpi)

            self._render_to_canvas(cr)

            ims.write_to_png(filename)
    
            ims.finish()
        elif filename.endswith('.svg') :
            canvas = self.fig.canvas
            w, h = self.plotter.get_size_inches()
            c_dpi = canvas.get_width_height()[0]/w
            f_dpi = 72.0 # points per inch is const
            ims = cairo.SVGSurface(filename, f_dpi*w, f_dpi*h)
            cr = cairo.Context(ims)

            cr.scale(f_dpi/c_dpi, f_dpi/c_dpi)

            self._render_to_canvas(cr)

            ims.finish()

        self.log(1, 'Exported to '+filename)

    def _render_to_canvas(self, cr) :
        canvas = self.fig.canvas
        renderer = RendererCairo (canvas.figure.dpi)
        w, h = canvas.get_width_height()
        renderer.set_width_height (w*0.9, h*0.9)
        renderer.gc.ctx = cr
        canvas.figure.draw (renderer)

    def set_ui(self) :
        self.ui_action_group = Gtk.ActionGroup('GlancerActions')
        self.ui_action_group.add_actions([('GlancerMenu', None, 'Glancer'),
            ('GlancerSaveImage', None, 'Save image', None, None,
                lambda w : self.savefig()),
                                         ])
        self.ui_ui_string = '''
            <ui>
                <menubar name="MenuBar">
                    <menu action="GlancerMenu">
                        <menuitem action="GlancerSaveImage"/>
                    </menu>
                </menubar>
            </ui>
        '''

    def _insert_annotation(self) :
        self.plotter.make_insert_annotation_dialog()

    def _update_location_line_butt(self) :
        self.location_cancel_butt.set_sensitive(False)
        self.location_line_butt.set_sensitive(False)
        if self.plotter is None :
            self.location_line_butt.set_label('')
            return

        line = self.plotter.get_active_line()

        loc, loc_type = self.plotter.get_last_mouse_location()

        if loc is not None :
            loc_string = "(%.3lf, %.3lf)" % loc
            if loc_type == 'target' :
                loc_string = "<b>%s</b>" % loc_string
            self.location_labl.set_markup(loc_string)
        else :
            self.location_labl.set_text('')

        if line is not None :
            self.location_line_butt.set_sensitive(True)
            self.location_cancel_butt.set_sensitive(True)

            pos_string = ''
            pos_string += '<span foreground="%s">' % \
                         str(Gdk.Color(*line.get_colour()))
            pos_string += line.get_aname_nice()
            pos_string += '</span>'
            self.location_line_butt.get_label_widget().set_markup(pos_string)

    def _do_location_cancel_butt_clicked(self, butt) :
        self.plotter.set_active_line(None)

    def _do_location_line_butt_toggled(self, butt) :
        self.plotter.set_highlight(butt.get_active())

    plotter = None
    def __init__(self, env=None):
        GObject.GObject.__init__(self)

        vbox = Gtk.VBox()
        self.fig = matplotlib.pyplot.figure()

        toolbar = Gtk.Toolbar()
        toolbar_hbox = Gtk.HBox()
        toolbar_hbox.pack_start(toolbar, True, True, 0)

        location_labl_titm = Gtk.ToolItem()
        self.location_labl = Gtk.Label()
        location_labl_titm.add(self.location_labl)
        self.location_line_butt = Gtk.ToggleToolButton()
        self.location_line_butt.set_label_widget(Gtk.Label())
        self.location_line_butt.set_sensitive(False)

        annotate_tbut = Gtk.ToolButton(Gtk.STOCK_EDIT)
        #annotate_tbut.connect("clicked", lambda w : self._insert_annotation())
        toolbar.insert(annotate_tbut, -1)

        annotate_tbut = Gtk.ToolButton(Gtk.STOCK_EDIT)
        annotate_tbut.connect("clicked", lambda w : self._insert_annotation())
        toolbar.insert(annotate_tbut, -1)
        annotate_tbut = Gtk.ToolButton(Gtk.STOCK_EDIT)
        annotate_tbut.connect("clicked", lambda w :
                              self.frame.do_child_resize_event(None, None))
        toolbar.insert(annotate_tbut, -1)

        self.location_cancel_butt = Gtk.ToolButton(Gtk.STOCK_CLOSE)
        self.location_cancel_butt.set_sensitive(False)
        toolbar_hbox.pack_end(self.location_cancel_butt, False, False, 0)

        self.location_line_butt.set_sensitive(False)
        self.location_line_butt.connect("toggled",
                                        self._do_location_line_butt_toggled)
        toolbar_hbox.pack_end(self.location_line_butt, False, False, 0)

        toolbar_hbox.pack_end(location_labl_titm, False, False, 0)

        vbox.pack_start(toolbar_hbox, False, False, 0)

        self.frame = GlancerPlotLayout()
        vbox.pack_start(self.frame, True, True, 0)

        self.time_hbox = Gtk.HBox()
        self.time_hbox.pack_start(Gtk.Label('Time (as tuple, True, True, 0) :'), False, False, 0)
        #vbox.pack_start(self.time_hbox, False, False, 0)

        vbox.show_all()
        self.time_hbox.hide()

        self.add(vbox)

        self.set_size_request(0, 200)

        self.source_action = lambda s :\
            self.load_series(get_object_from_dictionary(s))

        AObject.__init__(self, "Glancer", env, view_object = True)
        self.set_aname_nice("Plot" + (" ("+str(self.get_aname_num())+")" if self.get_aname_num()>1 else ""))

    def load_series(self, source, series = None, vals = None):
        if self.plotter is None :
            dims = source.source_get_max_dim()
            if dims == 3:
                self.new_plotter("3D Plot", GlancerPlot3D)
            else :
                self.new_plotter("2D Plot", GlancerPlot)

        self.plotter.load_series(source, series=series, vals=vals)

        self.elevate()
        self.queue_draw()

    def aes_add_a(self, aname_root, **parameters) :
        if aname_root in ('GlancerLine', 'GlancerLegend', 'GlancerAnnotation',
                          'GlancerMark') :
            return self.plotter.aes_add_a(aname_root, **parameters)
        if aname_root == 'GlancerPlot' :
            self.new_plotter("2D Plot", GlancerPlot)
            return self.plotter
        return AObject.aes_add_a(self, aname_root, **parameters)

    def new_plotter(self, title, plotter) :
        self.check_clear()
        if self.plotter is not None :
            self.frame.remove_all()
            self.time_hbox.remove(self.time_hbox.get_children()[1])
            self.plotter.aes_remove()
            del self.plotter
        self.fig.clear()

        if self.plotter is not None :
            self.plotter.disconnect(self._active_line_conn)
            self.plotter.disconnect(self._location_conn)

        self.plotter = plotter(self.fig, queue_draw=self.queue_draw, env=self.get_aenv())

        self._active_line_conn = self.plotter.connect('active-line-changed',
                                                      lambda b, l :
                                                      self._update_location_line_butt())
        self._location_conn = self.plotter.connect('location-changed',
                                                      lambda b :
                                                   self._update_location_line_butt())

        self.absorb_properties(self.plotter, as_self=False)
        self.frame.add(self.plotter.canvas)
        self.frame.show_all()

        self.time_hbox.pack_start(self.plotter.time_entr, True, True, 0)

        global object_dictionary; object_dictionary.set_show(self, True)
        self.set_aname_nice(title)
        self.queue_draw()

    def do_draw(self, cr):
        ret = Gtk.Frame.do_draw(self, cr)
        if self.plotter is not None :
            self.plotter.canvas.draw()
        return ret

    def print_out(self, op, pc, pn) :
        w = pc.get_width(); h = pc.get_height()
        w1, h1 = self.plotter.canvas.get_width_height()
        r = 1
        if w/h > w1/h1 : w = h*w1/h1; r = h/h1
        else : h = w*h1/w1; r = w/w1
        #op.cancel()

        c = pc.get_cairo_context()
        c.scale(r, r)
        renderer = RendererCairo (self.plotter.canvas.figure.dpi)
        renderer.set_width_height (w1, h1)
        renderer.gc.ctx = pc.get_cairo_context()
        self.plotter.canvas.figure.draw (renderer)

    #PROPERTIES
    def get_auto_aesthete_properties(self):
        return { }
    #BEGIN PROPERTIES FUNCTIONS
    #END PROPERTIES FUNCTIONS

    def replot_all(self) :
        for line in self.lines :
            line.replot()

    def get_method_window(self) :
        win = Gtk.VBox()
        icon_table = Gtk.Table(3, 1)
        win.pack_start(icon_table, True, True, 0)

        new_butt = Gtk.ToggleButton()
        new_butt.set_image(Gtk.Image.new_from_stock(Gtk.STOCK_NEW,
                                                       Gtk.IconSize.BUTTON))
        new_butt.set_tooltip_text("New plot")
        icon_table.attach(new_butt, 0, 1, 0, 1)
        new_menu = Gtk.Menu()
        new_menu.attach_to_widget(new_butt, None)
        new_butt.connect("button_press_event", lambda w, e :\
            new_menu.popup(None, None, None, None, e.button, e.time) \
                if not w.get_active() else \
            new_menu.popdown())

        new_menu.set_title('New Plot')
        new_menu.modify_bg(Gtk.StateType.NORMAL, Gdk.Color(65535, 65535, 65535))

        new_labl = Gtk.Label()
        new_labl.set_markup('<b>New...</b>')
        new_meni = Gtk.MenuItem()
        new_meni.remove(new_meni.get_child())
        new_meni.add(new_labl)
        new_menu.append(new_meni)

        plot_butt = Gtk.MenuItem()
        plot_butt.remove(plot_butt.get_child())
        plot_butt.add(Gtk.Image.new_from_stock('aes-glancer-plot2d',
                                                       Gtk.IconSize.BUTTON))
        plot_butt.set_tooltip_text("New 2D plot")
        plot_butt.connect("activate", lambda o : self.new_plotter("2D Plot", GlancerPlot))
        new_menu.append(plot_butt)

        pie_butt = Gtk.MenuItem()
        pie_butt.remove(pie_butt.get_child())
        pie_butt.add(Gtk.Image.new_from_stock('aes-glancer-pie',
                                                       Gtk.IconSize.BUTTON))
        pie_butt.set_tooltip_text("New pie chart")
        pie_butt.connect("activate", lambda o : self.new_plotter("Pie Chart", GlancerPie))
        new_menu.append(pie_butt)

        plot_butt = Gtk.MenuItem()
        plot_butt.remove(plot_butt.get_child())
        plot_butt.add(Gtk.Image.new_from_stock('aes-glancer-plot3d',
                                                       Gtk.IconSize.BUTTON))
        plot_butt.set_tooltip_text("New 3D plot")
        plot_butt.connect("activate", lambda o : self.new_plotter("3D Plot", GlancerPlot3D))
        new_menu.append(plot_butt)

        new_menu.show_all()

        sim_butt = Gtk.Button()
        sim_butt.set_image(Gtk.Image.new_from_stock(Gtk.STOCK_INDEX,
                                                    Gtk.IconSize.BUTTON))
        sim_butt.set_tooltip_text("Plot from currently selected Source")
        sim_butt.set_sensitive(get_object_dictionary().selected_source \
                               is not None)
        icon_table.attach(sim_butt, 1, 2, 0, 1)
        sim_butt.connect("clicked", lambda o : self.load_from_sim(\
                            get_object_dictionary().selected_source))
        get_object_dictionary().connect(\
            'aesthete-selected-source-change',
            lambda tr : sim_butt.set_sensitive(True))

        win.set_border_width(5)
        win.show_all()

        return win

    def check_clear(self, force = False) :
        if self.plotter is not None :
            self.plotter.check_clear(force=force)

    def load_from_sim(self, aname) :
        if aname == None or aname == 'None' or aname == '' : return
        sim = get_object_from_dictionary(aname)

        self.check_clear()
        dim = sim.source_get_max_dim()

        if len(values) == 1 :
            self.load_series(sim)
        else :
            for point_set in values :
                points = point_set['values']
                series = range(0, len(points[0]))
                if dim > 1 :
                    trans = zip(*points)
                    if not self.read_series : trans[0] = series
                else :
                    trans = [series, points]
                self.load_series(None, trans[0], trans[1])

add_icons()

aname_root_catalog['Glancer'] = Glancer

