import os, math, sys, getopt, string
from gi.repository import Pango
import random
from gi.repository import Gdk, Gtk, GObject
import cairo
from ..tablemaker import PreferencesTableMaker
import threading
import numpy, numpy.fft
import scipy, scipy.interpolate, scipy.optimize
import pylab
from PIL import Image
from aobject.aobject import *

legend_locs = matplotlib.legend.Legend.codes    
legend_locs_rev = dict((legend_locs[k],k) for k in legend_locs)
print legend_locs_rev

class GlancerLegend(AObject) :
    font = None

    def get_useful_vars(self) :
        return {
                 'legend' : 'mpl Legend',
               }

    def supplement_context_menu(self, ctx_menu) :
        viz = self.get_visible()
        context_menu_remove = Gtk.MenuItem(("Hide" if viz else "Show") + " legend")
        context_menu_remove.connect("button-press-event", lambda o, e :
                                    self.set_visible(not viz))
        ctx_menu.add(context_menu_remove)

    def __init__ (self, axes, lines, managed_lines, queue_draw, remove, env=None):
        self.axes = axes
        self.lines = lines
        self.managed_lines = managed_lines
        self.queue_draw = queue_draw
        self.parent_remove = remove
        AObject.__init__(self, "GlancerLegend", env, elevate = False)
        self.set_loc('best')

    parent_remove = None
    def aes_remove(self) :
        self.axes.legend_ = None
        self.queue_draw()
        if self.parent_remove is not None :
            self.parent_remove(self)
        AObject.aes_remove(self)

    def update(self) :
        if self.visible and len(self.lines) > 0 :
            rows = filter(lambda r : r[2] is not None and\
                          r[2].get_show_in_legend(), self.managed_lines)
            handles_labels = [(row[2].line,row[2].line.get_label()) for row\
                              in rows]
            if len(handles_labels) == 0 :
                self.axes.legend_ = None
                self.queue_draw()
                return None

            handles, labels = zip(*handles_labels)
            self.axes.legend(handles, labels, loc=self.get_loc())

            if self.axes.legend_ is None :
                return None

            texts = self.axes.legend_.get_texts()
            for text in texts :
                label_props = text.get_fontproperties()
                font_to_mpl(label_props, self.font)
        else :
            self.axes.legend_ = None
        self.queue_draw()

    #PROPERTIES
    loc = None
    visible = True
    font = "Linux Libertine"
    def get_auto_aesthete_properties(self):
        return { 'loc' : (str,), 'visible' : (bool,), 'font' : (str,) }
    #BEGIN PROPERTIES FUNCTIONS
    def change_visible(self, visible) :
        self.visible = visible
        self.update()
    def change_loc(self, loc) :
        self.loc = loc
        self.update()
    def change_font(self, val) : 
        self.font = val
        self.update()

    def make_action_panel(self) :
        win = Gtk.VBox()
        win.aes_title = 'Legend'

        pref_tab = PreferencesTableMaker()

        visible_amtb = self.aes_method_toggle_button("visible", None,
                                                   preferencable=False)
        pref_tab.append_row('Visible', visible_amtb)

        pref_tab.append_heading("Appearance")
        legend_position_cmbo = Gtk.ComboBoxText()
        for loc in legend_locs :
            legend_position_cmbo.append_text(loc)
        self.aes_method_automate_combo_text(legend_position_cmbo, "loc")
        pref_tab.append_row("Position", legend_position_cmbo)

        pref_tab.append_row("Font", self.aes_method_font_button("font"))

        win.pack_start(pref_tab.make_table(), False, True, 0)
        win.show_all()

        return win

        #texts = self.legend.get_texts()
        #if len(texts) < 1 :
        #    return ''
        #label_props = texts[0].get_fontproperties()
        #return mpl_to_font(label_props)
    #END PROPERTIES FUNCTIONS

class GlancerAnnotation(AObject) :
    font = None

    parent_remove = None
    def aes_remove(self) :
        self.annotation.remove()
        if self.parent_remove is not None :
            self.parent_remove(self)
        AObject.aes_remove(self)

    def get_useful_vars(self) :
        return {'text' : 'Text', 'annotation' : 'mpl Annotation'}

    def aes_get_parameters(self) :
        return ['text', 'xy', 'xycoords']

    def supplement_context_menu(self, ctx_menu) :
        context_menu_remove = Gtk.MenuItem("Remove annotation")
        context_menu_remove.set_tooltip_text("Remove '%s' at %lf,%lf" % \
                                        (self.get_text(), self.get_xy()[0],
                                             self.get_xy()[1]))
        context_menu_remove.connect("button-press-event", lambda o, e :
                                    self.aes_remove())
        ctx_menu.add(context_menu_remove)
    def queue_draw(self) :
        #self.annotation.axes.draw_artist(self.annotation)
        self.queue_draw_plot()

    def __init__ (self, annotation, queue_draw, remove, env=None):
        self.annotation = annotation
        self.queue_draw_plot = queue_draw
        self.parent_remove = remove
        AObject.__init__(self, "GlancerAnnotation", env, elevate=False)

    def set_highlight(self, highlight) :
        self._highlight = highlight

        if highlight :
            self._highlight_old_bbox = self.annotation._bbox
            self._highlight_old_bbox_patch = self.annotation._bbox_patch

            self.annotation.set_bbox({'facecolor':'gray', 'alpha':0.5})
        else :
            self.annotation._bbox = self._highlight_old_bbox
            self.annotation._bbox_patch = self._highlight_old_bbox_patch

        self.queue_draw()

    @classmethod
    def create(cls, ax, text, xy, xycoords, queue_draw, remove, env=None) :
        ann = ax.annotate(text, xy=xy, xycoords=xycoords)
        return cls(ann, queue_draw, remove, env=env)

    def get_auto_aesthete_properties(self):
        return { 
                  'xy' : (string_to_float_tup,),
                 'xycoords' : (str,),
                 'text' : (str,),
                 'font' : (str,),
               }

    def get_xy(self) :
        return self.annotation.xy

    def change_xy(self, loc) :
        self.annotation.set_position(loc)
        self.annotation.xy = loc
        self.annotation.xytext = loc

    def get_text(self) :
        return self.annotation.get_text()

    def change_text(self, loc) :
        self.annotation.set_text(loc)
        self.queue_draw()

    def get_xycoords(self) :
        return self.annotation.xycoords

    def change_xycoords(self, var) :
        self.annotation.xycoords = var
        self.queue_draw()

    def change_font(self, val) : 
        label_props = self.annotation.get_fontproperties()
        font_to_mpl(label_props, val)
        self.queue_draw()

    def get_font(self) :
        label_props = self.annotation.get_fontproperties()
        return mpl_to_font(label_props)

    # Surely this is redundant in Plot.py? If it could somehow be a classmethod,
    # then we could use the same one twice...
    def make_action_panel(self) :
        win = Gtk.VBox()
        win.aes_title = 'Annotation'

        pref_tab = PreferencesTableMaker()

        text_entr = self.aes_method_entry_update('text')
        pref_tab.append_row('Text', text_entr)
        xy_entr = self.aes_method_tuple_entry_update('xy')
        pref_tab.append_row('Text location', xy_entr)
        xycoords_cmbo = Gtk.ComboBoxText.new_with_entry()
        xycoords_cmbo.append_text('data')
        xycoords_cmbo.append_text('figure points')
        xycoords_cmbo.append_text('figure pixels')
        xycoords_cmbo.append_text('figure fraction')
        xycoords_cmbo.append_text('axes points')
        xycoords_cmbo.append_text('axes pixels')
        pref_tab.append_row('Coordinates', xycoords_cmbo)
        self.aes_method_automate_combo_text(xycoords_cmbo, 'xycoords')

        pref_tab.append_row("Font", self.aes_method_font_button("font"))

        win.pack_start(pref_tab.make_table(), True, True, 0)
        win.show_all()

        return win

class GlancerMark(AObject) :
    font = None

    __gsignals__ = { "remove-request" : (GObject.SignalFlags.RUN_LAST,
                                         None, ()),
             }

    def detach(self) :
        if self.get_attached_to() is not None :
            self.emit('remove-request')
            self.disconnect(self._attached_to_remove_request_hid)

    parent_remove = None
    def aes_remove(self) :
        self.detach()
        if self.parent_remove is not None :
            self.parent_remove(self)
        self.mark.remove()
        AObject.aes_remove(self)

    def get_useful_vars(self) :
        return {'marker' : 'Marker', 'mark' : 'mpl (1pt) Line'}

    def aes_get_parameters(self) :
        return ['marker', 'xy']

    def supplement_context_menu(self, ctx_menu) :
        context_menu_remove = Gtk.MenuItem("Remove mark")
        context_menu_remove.set_tooltip_text("Remove %s at %lf,%lf" % \
                                        (self.get_marker(), self.get_xy()[0],
                                             self.get_xy()[1]))
        context_menu_remove.connect("button-press-event", lambda o, e :
                                    self.aes_remove())
        ctx_menu.add(context_menu_remove)

    def queue_draw(self) :
        #self.annotation.axes.draw_artist(self.annotation)
        self.queue_draw_plot()

    def __init__ (self, mark, queue_draw, remove, attached_to=None, point_id=None, env=None):
        self.mark = mark
        self.queue_draw_plot = queue_draw
        self.parent_remove_for_unattached = remove
        self.parent_remove = remove
        AObject.__init__(self, "GlancerMark", env, elevate=False)
        self.set_point_id(point_id)
        self.set_attached_to(attached_to)

    def set_highlight(self, highlight) :
        self._highlight = highlight

        if highlight :
            self._highlight_old_color = self.mark.get_color()
            self.mark.set_color('white')
        else :
            self.mark.set_color(self._highlight_old_color)

        self.queue_draw()

    @classmethod
    def create(cls, ax, marker, xy, queue_draw, remove, attached_to=None, point_id=None, env=None) :
        mark = ax.plot([xy[0]],[xy[1]], marker)[0]
        return cls(mark, queue_draw, remove, attached_to=attached_to, point_id=point_id, env=env)

    def get_auto_aesthete_properties(self):
        return { 
                  'xy' : (string_to_float_tup,),
                 'marker' : (str,),
                 'attached_to' : (get_object_from_dictionary,(AOBJECT_CAN_NONE,)),
                 'point_id' : (str,(AOBJECT_CAN_NONE,)),
               }

    def get_xy(self) :
        return (self.mark.get_xdata()[0], self.mark.get_ydata()[0])

    def change_xy(self, loc) :
        attached_to = self.get_attached_to()
        if attached_to is not None :
            loc = attached_to._get_closest_point(loc)
            self.set_point_id(loc[0])
            loc = loc[1:]

        self.mark.set_data([loc[0]],[loc[1]])
        self.queue_draw()

    def get_marker(self) :
        return self.mark.get_marker()

    def change_marker(self, marker) :
        self.mark.set_marker(marker)
        self.queue_draw()

    def change_attached_to(self, attached_to) :
        self.detach()
        self.attached_to = attached_to
        debug_print(attached_to)
        #FIXME: hanging aes_method_label!
        if attached_to is None :
            self._attached_to_labl.destroy()
            self._attached_to_labl = Gtk.Label()
            self._attached_to_labl.set_markup("<i>Unattached</i>")
            self._attached_to_labl.set_padding(3,0)
            self._attached_to_hbox.pack_end(self._attached_to_labl)
            self._attached_to_butt.set_visible(False)
            self.parent_remove = self.parent_remove_for_unattached
            for widget in self._detached_only :
                widget.set_sensitive(True)
        else :
            self._attached_to_labl.destroy()
            self._attached_to_labl = attached_to.aes_method_label('name_nice')
            self._attached_to_labl.set_padding(3,0)
            self._attached_to_hbox.pack_end(self._attached_to_labl)
            self._attached_to_butt.set_visible(True)
            self.parent_remove = None
            for widget in self._detached_only :
                widget.set_sensitive(False)

            self._attached_to_remove_request_hid = \
                self.connect('remove-request', attached_to.remove_mark)

        self._attached_to_labl.show_all()

    # Well... provided this is only called once, we can use the self.# =
    # It will be called only once by AObject, so we MUSTN'T call it at all
    def make_action_panel(self) :
        win = Gtk.VBox()
        win.aes_title = 'Mark'

        pref_tab = PreferencesTableMaker()

        text_entr = self.aes_method_entry_update('marker')
        pref_tab.append_row('Marker', text_entr)

        _attached_to_hbox = Gtk.HBox()
        self._attached_to_hbox = _attached_to_hbox
        self._attached_to_labl = Gtk.Label()
        self._attached_to_butt = Gtk.Button("Detach")
        self._attached_to_butt.connect("clicked", lambda o:
                                       self.set_attached_to(None))
        _attached_to_hbox.pack_end(self._attached_to_butt, False)
        _attached_to_hbox.pack_end(self._attached_to_labl)
        pref_tab.append_row('Attached', _attached_to_hbox)

        self._detached_only = []
        xy_entr = self.aes_method_tuple_entry_update('xy')
        pref_tab.append_row('Location', xy_entr)
        self._detached_only.append(xy_entr)

        win.pack_start(pref_tab.make_table(), True, True, 0)
        win.show_all()
        self._attached_to_butt.set_visible(False)

        self.change_attached_to(None)

        return win
