import os, math, sys, getopt, string
import lxml.etree as ET
from gi.repository import Pango
import random
from gi.repository import Gdk, Gtk, GObject
import cairo
from ..tablemaker import PreferencesTableMaker
import threading
import matplotlib
import numpy, numpy.fft
import scipy, scipy.interpolate, scipy.optimize
from matplotlib import cm
import pylab
from PIL import Image
from aobject.aobject import *
from ..glypher.Widget import GlyphImage
from ..glypher.Parser import parse_phrasegroup
from Legend import GlancerMark

class GlancerLineLike(AObject) :
    time = None
    def get_auto_aesthete_properties(self) :
        return {'time' : (string_to_float_tup, (AOBJECT_CAN_NONE,)) }

    def change_time(self, time) :
        self.time = time
        self.replot()
    def get_time(self) :
        return self.time

    def __init__(self, plot, line = None, source = None, resolution=None, env=None,
                 read_series=False, axes = None, aname="GlancerLine", time=None):

        self.read_series = read_series
        self.axes = axes
        self.resolution = resolution
        self.time = time

        if line is None and source is not None :
            self.source = source
            self._line_from_source()
        elif source is None and line is not None :
            self.line = line
        else :
            raise RuntimeError('Must specify one of line and source for Line')

        self.plot = plot
        self.source = source
        AObject.__init__(self, aname, env, elevate=False)

        self.set_aesthete_xml(self.source.get_aesthete_xml())

        default_label = self.source.get_aname_nice()
        if self.source.get_aesthete_xml() is None and len(default_label)>20 :
            loc = string.rfind(default_label, ' ', 0, 20)
            if loc != -1 :
                default_label = default_label[0:loc]+"..."
        self.change_property("label", default_label)


    def set_resolution(self, res) :
        self.change_property('resolution', res)

    def replot(self) :
        self._line_from_source()

    def aes_get_parameters(self) :
        d = AObject.aes_get_parameters(self)
        d.update({'source' : self.source})
        return d

    def redraw(self) :
        self.plot.do_legend()
        self.line.figure.canvas.draw()
    
    def aes_remove(self, parent_remove=True) :
        if (parent_remove) : self.plot.remove_line(self); self.plot.check_legend()
        if self.line in self.line.axes.lines :
            self.line.axes.lines.remove(self.line)
        for mark in self.marks :
            mark.aes_remove()
        if self._highlight_line is not None and \
                self._highlight_line.axes is not None and \
                self._highlight_line in self._highlight_line.axes.lines :
            self._highlight_line.remove()
            self._highlight_point.remove()
        AObject.aes_remove(self)
        if (parent_remove) : self.redraw()

    _highlight = False
    def set_highlight(self, highlight) :
        self._highlight = highlight

    label = None
    def set_label_for_button(self, obj=None) :
        text = None

        if obj is None :
            obj = self
        #if button is None :
        #    button = self.config_butt

        if obj is not None :
            rep = obj.get_arepr()
            if isinstance(rep, ET._ElementTree) :
                self.label = rep
                #button.remove(button.get_child())
                #ge = GlyphImage()
                #ge.set_xml(rep)
                #ge.show()
                #button.add(ge)
                return
            text = obj.get_aname_nice()

        if text is not None :
            self.label = text
            #button.set_label(text)
            #button.get_child().props.ellipsize = Pango.EllipsizeMode.END
            #button.get_child().modify_fg(Gtk.StateType.NORMAL,
            #             Gdk.Color(*self.get_colour()))
            #button.set_tooltip_markup("Line generated from <b>"+obj.get_aname_nice_long()+"</b>")

        self.plot.redraw_managed_lines()

class GlancerLine(GlancerLineLike) :
    line = None
    plot = None
    source = None
    resolution = None
    show_in_legend = True

    def _remake_highlight_line(self) :
        if self.line is None :
            return
        self._highlight_line = matplotlib.lines.Line2D(self.line.get_xdata(),
                                                       self.line.get_ydata())
        self._highlight_point = matplotlib.lines.Line2D([],[])
    def get_useful_vars(self) :
        return {
                 'line' : 'mpl Line',
               }
    def _line_from_source(self) :
        if self.source is None :
            return

        if self.source.needs_x_range and self.source.source_get_x_range() is None :
            x_range = self.axes.get_xlim()
        else :
            x_range = None

        values = self.source.source_get_values(multi_array=True,
                                           x_range=x_range,
                                           time=self.get_time(),
                                           resolution=self.get_resolution())[0]
        points = values['values']
        series = values['x']
        dim = self.source.source_get_max_dim()
        trans = [series, points]

        if self.line is None :
            self.line = self.axes.plot(trans[0], trans[1])[0]
        else :
            self.line.set_xdata(trans[0])
            self.line.set_ydata(trans[1])
        self._remake_highlight_line()
        self.update_marks()
        self.axes.figure.canvas.draw()

    def __init__(self, plot, line = None, source = None, resolution=None, env=None,
                 read_series=False, axes = None, aname="GlancerLine", time=None):
        self.marks = ()
        GlancerLineLike.__init__(self, plot, line=line, source=source,
                 resolution=resolution, env=env,
                 read_series=read_series, axes=axes, aname=aname, time=time)


    #PROPERTIES
    def get_auto_aesthete_properties(self):
        props = GlancerLineLike.get_auto_aesthete_properties(self)
        props.update({ 'label' : (str,),
             'marker' : (str,),
             'linewidth' : (float,(AOBJECT_CAN_NONE,)),
             'colour' : (string_to_float_tup,),
             'source' : (get_object_from_dictionary,),
             'markersize' : (float,(AOBJECT_CAN_NONE,)),
             'markevery' : (int,(AOBJECT_CAN_NONE,)),
             'markerfacecolor' : (string_to_float_tup,),
             'alpha' : (float,(AOBJECT_CAN_NONE,)),
             'zorder' : (int,(AOBJECT_CAN_NONE,)),
             'visible' : (bool,),
             'show_in_legend' : (bool,),
             'aa' : (bool,),
             'marks' : (string_to_aobject_tup,),
             'linestyle' : (str,),
             'resolution' : (int,(AOBJECT_CAN_NONE,)),
             'solid-capstyle' : (str,),
             'solid-joinstyle' : (str,),
             'dash-capstyle' : (str,),
             'dash-joinstyle' : (str,) } )
        return props

    #BEGIN PROPERTIES FUNCTIONS
    def get_label(self, val=None): return self.line.get_label()
    def get_marker(self, val=None): return self.line.get_marker()
    def get_linewidth(self, val=None): return self.line.get_linewidth()
    def get_colour(self, val=None):
        return mpl_to_tuple(self.line.get_color())
    def get_source(self, val=None) : return self.source
    def get_markersize(self, val=None): return self.line.get_markersize()
    def get_markevery(self, val=None): return self.line.get_markevery()
    def get_markerfacecolor(self, val=None):
        return mpl_to_tuple(self.line.get_markerfacecolor())
    def get_alpha(self, val=None): return self.line.get_alpha()
    def get_visible(self, val=None): return self.line.get_visible()
    def get_aa(self, val=None): return self.line.get_aa()
    def get_linestyle(self, val=None): return self.line.get_linestyle()
    def get_solid_capstyle(self, val=None): return self.line.get_solid_capstyle()
    def get_solid_joinstyle(self, val=None): return self.line.get_solid_joinstyle()
    def get_dash_capstyle(self, val=None): return self.line.get_dash_capstyle()
    def get_dash_joinstyle(self, val=None): return self.line.get_dash_joinstyle()
    def get_resolution(self, val=None): return self.resolution

    def change_resolution(self, val) : self.resolution = val; self.replot()
    def change_label(self, val) :
        self.line.set_label(val)
        self.set_aname_nice(val)
        self.plot.do_legend()
        self.redraw()
        self.set_label_for_button()
    def change_colour(self, val) :
        self.line.set_color(val)
        self.set_label_for_button()
        self.redraw()
    def change_marker(self, val) : self.line.set_marker(val); self.redraw()
    def change_linewidth(self, val) : self.line.set_linewidth(val); self.redraw()
    def change_markersize(self, val) : self.line.set_markersize(val); self.redraw()
    def change_markevery(self, val) : self.line.set_markevery(val); self.redraw()
    def change_markerfacecolor(self, val) : self.line.set_markerfacecolor(val); self.redraw()
    def change_alpha(self, val) : self.line.set_alpha(val); self.redraw()
    def change_zorder(self, val) :
        self.zorder = val
        if not self._highlight :
            self.line.set_zorder(val)
            self.redraw()
    def change_visible(self, val) :
        self.line.set_visible(val)
        self.set_show_in_legend(val)
        self.redraw()
    def change_show_in_legend(self, val) :
        self.show_in_legend = val
        self.redraw()
    def change_aa(self, val) : self.line.set_aa(val); self.redraw()
    def change_linestyle(self, val) : self.line.set_linestyle(val); self.redraw()
    def change_solid_capstyle(self, val) : self.line.set_solid_capstyle(val); self.redraw()
    def change_solid_joinstyle(self, val) : self.line.set_solid_joinstyle(val); self.redraw()
    def change_dash_capstyle(self, val) : self.line.set_dash_capstyle(val); self.redraw()
    def change_dash_joinstyle(self, val) : self.line.set_dash_joinstyle(val); self.redraw()

    #END PROPERTIES FUNCTIONS

    def _get_closest_point(self, loc) :
        #FIXME: this can be done easily and much more pythonically
        inv_trans_p = self.line.axes.transData.transform_point
        lox, loy = inv_trans_p(loc)
        ptx = self.line.get_xdata()[0]
        pty = self.line.get_ydata()[0]
        ptx, pty = inv_trans_p((ptx,pty))
        min_ind = 0
        min_dist = ((lox-ptx)**2+(loy-pty)**2)
        min_dist = 0 if min_dist < 1e-10 else min_dist**.5
        for i in range(len(self.line.get_xdata())) :
            ptx = self.line.get_xdata()[i]
            pty = self.line.get_ydata()[i]
            ptx, pty = inv_trans_p((ptx,pty))
            new_dist = ((lox-ptx)**2+(loy-pty)**2)**.5
            new_dist = 0 if new_dist < 1e-10 else new_dist**.5
            if min_dist > new_dist :
                min_ind = i
                min_dist = new_dist
        return min_ind, self.line.get_xdata()[min_ind], self.line.get_ydata()[min_ind]

    _data_loc = None
    def set_location(self, loc) :
        self._data_loc = self._get_closest_point(loc)
        if self._highlight :
            self._highlight_point.set_data([self._data_loc[1]],
                                           [self._data_loc[2]])
            if self._highlight_point not in self.line.axes.lines :
                self.line.axes.add_line(self._highlight_point)
        self.redraw()

    def supplement_context_menu(self, ctx_menu) :
        viz = self.get_visible()
        context_menu_aname_nice = Gtk.MenuItem(self.get_aname_nice())
        context_menu_aname_nice.set_sensitive(False)
        ctx_menu.add(context_menu_aname_nice)
        context_menu_visible = Gtk.MenuItem("Hide" if viz else "Show")
        context_menu_visible.connect("button-press-event", lambda o, e :
                                    self.set_visible(not viz))
        ctx_menu.add(context_menu_visible)
        context_menu_mark = Gtk.MenuItem("Mark line...")
        context_menu_mark.connect("button-press-event", lambda o, e :
                                    self.make_mark_dialog(self._data_loc[0]))
        context_menu_remove = Gtk.MenuItem("Remove line")
        context_menu_remove.set_tooltip_text("Remove %s" % \
                                        self.get_aname_nice_long())
        context_menu_remove.connect("button-press-event", lambda o, e :
                                    self.aes_remove())
        ctx_menu.add(context_menu_remove)
        ctx_menu.add(context_menu_mark)

    def get_data_location(self) :
        return self._data_loc

    def set_highlight(self, highlight) :
        GlancerLineLike.set_highlight(self, highlight)
        if highlight :
            self.line.set_zorder(1000)
            if self._highlight_line is not None :
                self._highlight_line.update_from(self.line)
                self._highlight_line.set_linewidth(self.get_linewidth()*10)
                self._highlight_line.set_linestyle('-')
                self._highlight_line.set_alpha(0.2)
                self._highlight_line.set_zorder(999)
                self._highlight_point.set_markersize(10*self.line.get_linewidth())
                self._highlight_point.set_marker('o')
                self._highlight_point.set_color(self.line.get_color())
                self._highlight_point.set_zorder(1001)
                self._highlight_point.set_alpha(0.6)
                data_loc = self.get_data_location()
                if data_loc is not None :
                    self._highlight_point.set_data([data_loc[1]], [data_loc[2]])
                    self.line.axes.add_line(self._highlight_point)
                self.line.axes.add_line(self._highlight_line)
        else :
            self.line.set_zorder(self.get_zorder())
            if self._highlight_line is not None and \
                    self._highlight_line.axes is not None and \
                    self._highlight_line in self._highlight_line.axes.lines :
                self._highlight_line.remove()
                self._highlight_point.remove()
        self.redraw()

    def remove_mark(self, mark) :
        marks = list(self.get_marks())
        marks.remove(mark)
        self.set_marks(tuple(marks))

    def update_marks(self) :
        for mark in self.marks :
            index = int(mark.get_point_id)
            mark.set_xy((self.line.get_xdata()[index],
                         self.line.get_ydata()[index]))

    def add_mark(self, mark, index) :
        mark_obj = GlancerMark.create(self.line.axes,
                                      mark,
                                      (self.line.get_xdata()[index],
                                       self.line.get_ydata()[index]),
                                      self.redraw,
                                      self.plot.remove_mark,
                                      attached_to=self,
                                      point_id=index,
                                      env=self.get_aenv())
        mark_obj.mark.set_zorder(self.line.get_zorder())

        # While it's attached to a line, a mark is technically owned by the plot
        self.absorber.absorb_properties(mark_obj, as_self=False)

        # Why do we use a tuple? So we can use mark as a property - good for
        # saving/opening
        marks = list(self.marks)
        marks.append(mark_obj)
        self.set_marks(tuple(marks))
        
    def make_mark_dialog(self, index) :
        ann_dial = Gtk.Dialog(\
                      title="Add mark", parent=self.get_aenv().toplevel, 
                      flags=Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT,
                      buttons=(Gtk.STOCK_CANCEL,Gtk.ResponseType.CANCEL,Gtk.STOCK_OK,Gtk.ResponseType.OK))
        ann_dial.set_default_response(Gtk.ResponseType.OK)

        vbox = Gtk.VBox()

        pref_tab = PreferencesTableMaker()

        text_entr = Gtk.Entry()
        pref_tab.append_row('Marker', text_entr)
        text_entr.set_text('o')
        xy_entr = Gtk.Entry()
        xy = self.line.get_xdata()[index], self.line.get_ydata()[index]
        xy_entr.set_text(str(xy) + ' (Point %d)'%index)
        xy_entr.set_sensitive(False)
        pref_tab.append_row('Location', xy_entr)

        vbox.pack_start(pref_tab.make_table(), True, True, 0)
        ann_dial.get_content_area().add(vbox)
        ann_dial.get_content_area().show_all()

        resp = ann_dial.run()
        ann_dial.grab_focus()

        if resp == Gtk.ResponseType.OK :
            marker = text_entr.get_text()
            ann_dial.destroy()
        else :
            ann_dial.destroy()
            return

        self.add_mark(marker, index)

    def make_action_panel(self) :
        #frame = Gtk.Frame(self.source)
        #hbox = Gtk.HBox()
        #config_butt = Gtk.Button(self.get_aname_nice()+'...')
        #config_butt.set_relief(Gtk.ReliefStyle.NONE)
        #config_butt.connect("clicked", lambda w : self.plot.set_active_line(self))
        #self.config_butt = config_butt
        #self.connect("aesthete-aname-nice-change", lambda o, a, v, l : self.set_label_for_button(config_butt,o))
        #self.connect("aesthete-aname-nice-change", lambda o, a, v, l : self.set_label_for_button(config_butt,o))
        #hbox.pack_start(config_butt, True, True, 0)

        #remove_butt = Gtk.Button(); remove_butt.add(Gtk.Image.new_from_stock(Gtk.STOCK_CLEAR, Gtk.IconSize.SMALL_TOOLBAR))
        #remove_butt.connect("clicked", lambda o : self.self_remove())
        #hbox.pack_start(remove_butt, False, False, 0)

        win = Gtk.VBox()
        nb = Gtk.Notebook()

        line_table_maker = PreferencesTableMaker()
        line_table_maker.append_row("Label",
                                    self.aes_method_entry_update("label"))
        line_table_maker.append_row("Colour",
                self.aes_method_colour_button("colour", "Set line colour"))
        line_table_maker.append_row("Thickness",
                self.aes_method_entry("linewidth",
                     wait_until_parsable_float=True))
        line_table_maker.append_row("Alpha",
                self.aes_method_entry_update("alpha"))
        line_table_maker.append_row("Z order",
                self.aes_method_entry_update("zorder"))
        line_table_maker.append_row("Visible",
                self.aes_method_toggle_button("visible", onoff=('On','Off')))
        line_table_maker.append_row("In legend",
                self.aes_method_toggle_button("show_in_legend"))
        if self.source.needs_resolution :
            line_table_maker.append_row("Resolution",
                self.aes_method_entry_update("resolution", 'Set'))
        line_table_maker.append_row("Antialiased",
                self.aes_method_toggle_button("aa", onoff=('On','Off')))
        nb.append_page(line_table_maker.make_table(), Gtk.Label(label="General"))

        marker_table_maker = PreferencesTableMaker()
        marker_table_maker.append_row("Marker",
                self.aes_method_entry_update("marker"))
        marker_table_maker.append_row("Size",
                self.aes_method_entry_update("markersize"))
        marker_table_maker.append_row("Freq",
                self.aes_method_entry_update("markevery"))
        marker_table_maker.append_row("Colour",
                self.aes_method_colour_button("markerfacecolor", "Set marker colour"))
        nb.append_page(marker_table_maker.make_table(), Gtk.Label(label="Marker"))

        style_table_maker = PreferencesTableMaker()
        style_table_maker.append_row("Line Style",
                self.aes_method_entry_update("linestyle"))
        style_table_maker.append_row("Solid Cap",
                self.aes_method_entry_update("solid-capstyle"))
        style_table_maker.append_row("Solid Join",
                self.aes_method_entry_update("solid-joinstyle"))
        style_table_maker.append_row("Dash Cap",
                self.aes_method_entry_update("dash-capstyle"))
        style_table_maker.append_row("Dash Join",
                self.aes_method_entry_update("dash-joinstyle"))
        nb.append_page(style_table_maker.make_table(), Gtk.Label(label="Style"))

        win.pack_start(nb, False, False, 0)

        if self.source is not None:
            win.pack_start(Gtk.HSeparator(), False, True, 0)
            update_hbox = Gtk.HBox()
            update_labl = Gtk.Label()
            update_labl.set_markup("from source <b>"+\
                self.source.get_aname_nice()+\
                "</b>")
            update_labl.set_tooltip_text(self.source.get_aname_nice_long())
            update_labl.set_alignment(1.0, 0.5)
            update_hbox.pack_start(update_labl, True, True, 0)
            update_butt = Gtk.Button()
            update_butt.add(Gtk.Image.new_from_stock(Gtk.STOCK_REFRESH,
                                                     Gtk.IconSize.BUTTON))
            update_butt.connect("clicked", lambda e : self.replot())
            update_hbox.pack_start(update_butt, False, False, 0)
            win.pack_start(update_hbox, False, False, 0)

        self.connect("aesthete-property-change",
                     lambda o, p, v, a : self.set_label_for_button())

        win.show_all()
        self.action_panel = win

        if self.env and self.env.action_panel :
            win.hide()
            win.aes_title = "Configure line"
            #config_butt.connect("clicked", lambda o : self.env.action_panel.to_action_panel(win))
        else :
            config_win = Gtk.Window()
            config_win.set_title("Configure line")
            remove_butt = Gtk.Button("Close"); remove_butt.connect("clicked", lambda o : config_win.hide())
            win.pack_start(remove_butt, True, True, 0)
            config_win.add(win)
            config_win.hide()
            #config_butt.connect("clicked", lambda o : config_win.show())
        #hbox.show_all()
        #frame.add(win)

        return win

class GlancerLine3D(GlancerLine) :
    def __init__(self, plot, line = None, source = None, resolution=None, env=None,
                 read_series=False, axes = None, aname="GlancerLine3D",
                 time=None):
        GlancerLine.__init__(self, plot, line=line, source=source,
                             resolution=resolution, env=env,
                 read_series=read_series, axes=axes, aname=aname, time=time)

    def _line_from_source(self) :
        if self.source is None :
            return

        x_range = self.axes.get_xlim() if self.source.needs_x_range else None
        values = self.source.source_get_values(multi_array=True,
                                               x_range=x_range,
                                               time=self.get_time())[0]

        points = values['values']
        series = values['x']
        dim = self.source.source_get_max_dim()
        if dim <3 :
            points = (points,)
        trans = [series, points]

        trans = (trans[0], trans[1][0], 0 if dim < 3 else trans[1][1])
        self.line = self.axes.plot(*trans)[0]
        # Can't replot as no set_zdata :(
        self.axes.figure.canvas.draw()

class GlancerSurface(GlancerLineLike) :
    line = None
    plot = None
    source = None
    resolution = None

    def get_useful_vars(self) :
        return {
                 'line' : 'mpl Poly3DCollection',
               }

    def __init__(self, plot, line = None, source = None, resolution=None, env=None,
                 read_series=False, axes = None, aname="GlancerSurface",
                 time=None):
        GlancerLineLike.__init__(self, plot, line=line, source=source,
                 resolution=resolution, env=env,
                 read_series=read_series, axes=axes, aname=aname, time=time)

    #PROPERTIES
    def get_aesthete_properties(self):
        return { 'label' : [self.change_label, self.get_label, True],
             'linewidth' : [self.change_linewidth, self.get_linewidth, True],
             'source' : [None, self.get_source, True],
             'alpha' : [self.change_alpha, self.get_alpha, True],
             'zorder' : [self.change_zorder, self.get_zorder, True],
             'visible' : [self.change_visible, self.get_visible, True],
             'resolution' : [self.change_resolution, self.get_resolution, True],
             'linestyle' : [self.change_linestyle, self.get_linestyle, True]}
    #BEGIN PROPERTIES FUNCTIONS
    def get_label(self, val=None): return self.line.get_label() if val==None else val
    def get_marker(self, val=None): return self.line.get_marker() if val==None else val
    def get_linewidth(self, val=None): return self.line.get_linewidth() if val==None else float(val)
    def get_source(self, val=None) : return self.source if val==None else val
    def get_alpha(self, val=None): return self.line.get_alpha() if val==None else float(val)
    def get_zorder(self, val=None): return self.line.get_zorder() if val==None else float(val)
    def get_visible(self, val=None): return self.line.get_visible() if val==None else (val=='True')
    def get_linestyle(self, val=None): return self.line.get_linestyle() if val==None else val
    def get_resolution(self, val=None): return self.resolution if val==None else int(val)

    def change_label(self, val) :
        self.line.set_label(val)
        self.set_aname_nice(val)
        self.plot.do_legend()
        self.redraw()
        self.line.set_label(val)
    def change_linewidth(self, val) : self.line.set_linewidth(val); self.redraw()
    def change_alpha(self, val) : self.line.set_alpha(val); self.redraw()
    def change_zorder(self, val) : self.line.set_zorder(val); self.redraw()
    def change_visible(self, val) : self.line.set_visible(val); self.redraw()
    def change_linestyle(self, val) : self.line.set_linestyle(val); self.redraw()
    def change_resolution(self, val) : self.resolution = val; self.replot()

    #END PROPERTIES FUNCTIONS
    #frame = Gtk.Frame(self.source)
    #hbox = Gtk.HBox()
    #config_butt = Gtk.Button(self.get_aname_nice()+'...')
    #config_butt.set_relief(Gtk.ReliefStyle.NONE)
    #self.connect("aesthete-aname-nice-change", lambda o, a, v, l : self.set_label_for_button(config_butt))
    #self.connect("aesthete-aname-nice-change", lambda o, a, v, l : self.set_label_for_button(config_butt))
    #self.config_butt = config_butt
    #hbox.pack_start(config_butt, True, True, 0)
    #remove_butt = Gtk.Button(); remove_butt.add(Gtk.Image.new_from_stock(Gtk.STOCK_CLEAR, Gtk.IconSize.SMALL_TOOLBAR))
    #remove_butt.connect("clicked", lambda o : self.self_remove())
    #hbox.pack_start(remove_butt, False, False, 0)

    def make_action_panel(self) :
        win = Gtk.VBox()
        nb = Gtk.Notebook()

        line_table_maker = PreferencesTableMaker()
        line_table_maker.append_row("Label",
                                    self.aes_method_entry_update("label", "Set"))
        line_table_maker.append_row("Thickness",
                self.aes_method_entry("linewidth",
                     wait_until_parsable_float=True))
        line_table_maker.append_row("Alpha",
                self.aes_method_entry_update("alpha", "Set"))
        line_table_maker.append_row("Z order",
                self.aes_method_entry_update("zorder", "Set"))
        line_table_maker.append_row("Visible",
                self.aes_method_toggle_button("visible", onoff=('On','Off')))
        if self.source.needs_resolution :
            line_table_maker.append_row("Resolution",
                self.aes_method_entry_update("resolution", 'Set'))
        nb.append_page(line_table_maker.make_table(), Gtk.Label(label="General"))

        style_table_maker = PreferencesTableMaker()
        style_table_maker.append_row("Line Style",
                self.aes_method_entry_update("linestyle", "Set"))
        nb.append_page(style_table_maker.make_table(), Gtk.Label(label="Style"))

        win.pack_start(nb, False, False, 0)

        if self.source is not None:
            win.pack_start(Gtk.HSeparator(), False, True, 0)
            update_hbox = Gtk.HBox()
            update_labl = Gtk.Label()
            update_labl.set_markup("from source <b>"+\
                self.source.get_aname_nice()+\
                "</b>")
            update_labl.set_alignment(1.0, 0.5)
            update_hbox.pack_start(update_labl, True, True, 0)
            update_butt = Gtk.Button()
            update_butt.add(Gtk.Image.new_from_stock(Gtk.STOCK_REFRESH,
                                                     Gtk.IconSize.BUTTON))
            update_butt.connect("clicked", lambda e : self.replot())
            update_hbox.pack_start(update_butt, False, False, 0)
            win.pack_start(update_hbox, False, False, 0)

        self.connect("aesthete-property-change",
                     lambda o, p, v, a : self.set_label_for_button())

        win.show_all()

        #if self.env and self.env.action_panel :
        #    win.hide()
        #    win.aes_title = "Configure line"
        #    config_butt.connect("clicked", lambda o : self.env.action_panel.to_action_panel(win))
        #else :
        #    config_win = Gtk.Window()
        #    config_win.set_title("Configure line")
        #    remove_butt = Gtk.Button("Close"); remove_butt.connect("clicked", lambda o : config_win.hide())
        #    win.pack_start(remove_butt, True, True, 0)
        #    config_win.add(win)
        #    config_win.hide()
        #    config_butt.connect("clicked", lambda o : config_win.show())
        #frame.add(win)

        return win
    def _line_from_source(self) :
        if self.source is None :
            return

        x_range = self.axes.get_xlim() if self.source.needs_x_range else None
        y_range = self.axes.get_ylim()
        values = self.source.source_get_values(multi_array=True,
                                               x_range=x_range,
                                               y_range=y_range,
                                               time=self.get_time(),
                                               resolution=self.get_resolution())[0]

        points = values['values']
        xs = values['x']
        ys = values['y']
        self.line = self.axes.plot_surface(xs, ys, points, rstride=1, cstride=1,
                                          cmap=cm.jet)
        # Can't replot as no set_zdata :(
        self.axes.figure.canvas.draw()
