import os, math, sys, getopt, string
import time
from aobject.utils import debug_print
from gi.repository import Pango
import random
from gi.repository import Gdk, Gtk, GObject
import cairo
from ..tablemaker import PreferencesTableMaker
import threading
import matplotlib
import numpy, numpy.fft
import scipy, scipy.interpolate, scipy.optimize
import pylab
from PIL import Image
from aobject.aobject import *
from Canvas import *
from Line import GlancerLine
from Scatter import GlancerScatter
from Legend import GlancerLegend, GlancerAnnotation, GlancerMark
from ..glypher import Widget

def lines_cdf(c, cr, tm, it, data) :
    obj = get_object_from_dictionary(tm.get_value(it,0))
    if obj is None : return
    if obj.label is None :
        cr.set_property('obj', obj)
    else :
        cr.set_property('obj', obj.label)

class LineGraphicCellRenderer(Gtk.CellRenderer) :
    __gproperties__ = {
        "aname": (GObject.TYPE_STRING, "aname", "AObject name", "", GObject.PARAM_READWRITE),
    }

    def __init__(self) :
        Gtk.CellRenderer.__init__(self)

    def do_set_property(self, pspec, value) :
        setattr(self, pspec.name, value)

    def do_get_property(self, pspec) :
        return getattr(self, pspec.name)

    def do_render(self, cr, widget, background_area, cell_area,
                  flags) :
        surface = cr.get_target()
        cr = cairo.Context(surface)

        line = get_object_from_dictionary(self.aname)

        if line is None :
            return
        
        cr.save()
        cr.translate(cell_area.x, cell_area.y)
        cr.rectangle(0, 0, cell_area.width, cell_area.height)
        cr.set_source_rgb(1, 1, 1)
        cr.fill()

        hv = cell_area.height/2.
        cr.rectangle(0, 0.9*hv, cell_area.width, (1.1)*hv)
        cr.set_source_rgb(*line.get_colour())
        cr.fill()

        cr.restore()

    line_height = 100
    def on_get_size(self, widget, cell_area=None):
        x = 0
        y = 0

        if cell_area is not None :
            x = cell_area.x
            y = cell_area.y

        w = 30
        h = 10

        return (x, y, w, h)

GObject.type_register(LineGraphicCellRenderer)

class GlancerPlotLike(AObject) :
    __gsignals__ = { "active-line-changed" : (GObject.SignalFlags.RUN_LAST,
                                              None, (GObject.TYPE_STRING,)),
                     "location-changed" : (GObject.SignalFlags.RUN_LAST,
                                           None, ()),
             }
    fig = None
    axes = None
    canvas = None

    lines = None
    legend_object = None

    plot_over = False
    title = ''
    xlabel_font = ''

    _last_location = None
    _last_location_type = 'normal'

    grid = get_preferencer().get_preference("GlancerPlot","grid")

    if grid is None :
        grid = False

    grid_color = get_preferencer().get_preference("GlancerPlot","grid_color")

    if grid_color is None :
        grid_color = 'k'

    read_series = True

    scroll_speed = 0.1

    _xlim_changed_conn = None

    def change_time(self, time) :
        self.time = time
        for line in self.lines :
            line.set_time(time)

    def _do_xlim_changed(self, ax) :
        for line in self.lines :
            if line.source.needs_x_range and \
               line.source.current_x_range != self.axes.get_xlim() :
                line.source.source_reload()
                line.replot()

    def insert_mark(self, marker, xy, attached_to=None, point_id=None) :
        ann_obj = GlancerMark.create(self.axes, marker, xy,self.queue_draw, 
                                      self.remove_mark,
                                     attached_to=attached_to, point_id=point_id,
                                           env=self.get_aenv())
        self.absorb_properties(ann_obj, as_self=False)
        self.marks.append(ann_obj)
        self.queue_draw()
        return ann_obj

    def remove_mark(self, mark) :
        self.marks.remove(mark)

    def insert_annotation(self, text, xy, xycoords) :
        ann_obj = GlancerAnnotation.create(self.axes, text, xy, xycoords,
                                           self.queue_draw,
                                           self.remove_annotation, env=self.get_aenv())
        self.absorb_properties(ann_obj, as_self=False)
        self.annotations.append(ann_obj)
        self.queue_draw()
        return ann_obj

    def remove_annotation(self, annotation) :
        self.annotations.remove(annotation)

    def set_xlim(self, left=None, right=None) :
        if left is None :
            left = self.get_axes_xmin()
        if right is None :
            right = self.get_axes_xmax()
        self.axes.set_xlim(left, right)

        self.queue_draw()

    def set_ylim(self, top=None, bottom=None) :
        if top is None :
            top = self.get_axes_ymax()
        if bottom is None :
            bottom = self.get_axes_ymin()
        self.axes.set_ylim(bottom, top)
        self.queue_draw()

    def get_useful_vars(self) :
        return {
                 'axes' : 'mpl Axes',
                 'fig' : 'mpl Fig',
                 'canvas' : 'mpl Canvas',
        }

    def aes_add_a(self, aname_root, **parameters) :
        debug_print(aname_root, parameters)
        if aname_root == 'GlancerLine' :
            source = parameters['source']
            return self.load_series(source)
        if aname_root == 'GlancerLegend' :
            return GlancerLegend(self.axes, self.lines, self.managed_lines_lsst,
                                 self.queue_draw, self.remove_legend, env=self.get_aenv()
                                 )
        if aname_root == 'GlancerAnnotation' :
            return self.insert_annotation(\
                                   parameters['text'],
                                   string_to_float_tup(parameters['xy']),
                                   parameters['xycoords'])
        if aname_root == 'GlancerMark' :
            # We trust this to be set in the properties. Problem is that the
            # attached_to object is not necessarily loaded when Mark is created.
            # Another good reason not to use parameters unless you really CANNOT
            # build the object without them (or it would be really, really slow
            # to change the property afterwards, AND the property is highly
            # unlikely to otherwise change)
            return self.insert_mark(\
                                   parameters['marker'],
                                   string_to_float_tup(parameters['xy']))

        return AObject.aes_add_a(self, aname_root, **parameters)

    def __init__(self, aname_root, fig, queue_draw, env=None):
        self.marks = []
        #self.canvas = GlancerCanvas(fig)
        self.canvas = mpl_Canvas(fig)
        self.queue_draw = queue_draw
        self.fig = fig

        self._xlim_changed_conn = self.axes.callbacks.connect('xlim_changed',
                                                    self._do_xlim_changed)

        self.canvas.mpl_connect('scroll_event', self.do_mpl_scroll_event)
        self.canvas.mpl_connect('button_press_event', self.do_mpl_button_press_event)
        self.canvas.mpl_connect('button_release_event',
                                self.do_mpl_button_release_event)
        self.canvas.mpl_connect('motion_notify_event',
                                self.do_mpl_motion_notify_event)
        self.lines = []
        self.annotations = []

        AObject.__init__(self, aname_root, env, view_object=False)
        self.time_entr = self.aes_method_entry_update('time')
        self.time_entr.show_all()

    def remove_legend(self, legend) :
        self.legend_object = None

    def do_legend(self, on=None, loc=None) :
        if self.legend_object is None :
            self.legend_object = GlancerLegend(self.axes, self.lines,
                                               self.managed_lines_lsst,
                                               self.queue_draw,
                                               self.remove_legend,
                                               env=self.get_aenv())
            self.absorb_properties(self.legend_object, as_self=False)

        if on is not None :
            self.legend_object.set_visible(on)
        if loc is not None :
            self.legend_object.set_loc(loc)
        self.legend_object.update()

    def check_legend(self) :
        self.legend_object.update()

    def check_clear(self, force = False) :
        if not self.plot_over or force :
            self.axes.clear()
            for line in self.lines :
                line.aes_remove(parent_remove = False)
            self.lines = []
            self.check_legend()
            self.queue_draw()

    def update_canvas_size(self) :
        w, h = self.get_size_inches()
        dpi = self.fig.get_dpi()
        h *= dpi
        w *= dpi
        #self.fig.canvas.resize(int(w), int(h))
        self.fig.canvas.set_size_request(int(w), int(h))
        #self.fig.canvas.resize(300, 300)
        #self.fig.canvas.set_size_request(300, 300)
        self.queue_draw()

    def replot_all(self) :
        for line in self.lines :
            line.replot()

    def remove_line(self, line) :
        self.lines.remove(line)
        for row in self.managed_lines_lsst :
            if row[2] == line :
                self.managed_lines_lsst.remove(row.iter)

    def append_line(self, line) :
        self.lines.append(line)
        self.managed_lines_lsst.append((line.get_aname(), line.get_aname_nice(),
                                        line, line.get_aname_nice_long()))

    def redraw_managed_lines(self) :
        self.lines_trvw.queue_draw()

class GlancerPlot(GlancerPlotLike) :
    max_dim = 2

    def __init__(self, fig, queue_draw, env=None):
        self.axes = fig.add_subplot(1,1,1)
        fig.canvas.mpl_connect('resize_event', lambda e :
                               debug_print("HI"))
        GlancerPlotLike.__init__(self, "GlancerPlot", fig, queue_draw, env=env)

    def restyle_by_order(self) :
        order = self.get_order()
        for i in range(len(order)) :
            order[i].set_astyle_id('line-%d' % i)
        self.line_style_index = len(order)
        for line in order :
            line.aes_restyle()
        self.do_legend()

    line_style_index = 0
    def load_series(self, source, series = None, vals = None):
        resolution = self.get_resolution()
        if source is None :
            mpl_line, = self.axes.plot(series, vals)
            line = GlancerLine(self, line=mpl_line, resolution=resolution,
                               env=self.get_aenv(), time=self.get_time())
        elif source.source_type() == 'line' :
            line = GlancerLine(self, source=source, axes=self.axes,
                           read_series=self.read_series,
                           resolution=resolution,
                           env=self.get_aenv(), time=self.get_time())
        else :
            line = GlancerScatter(self, source=source, axes=self.axes,
                           read_series=self.read_series,
                           env=self.get_aenv())

        line.set_astyle_id('line-%d' % self.line_style_index)
        line.aes_restyle(self.get_default_line_style())
        self.append_line(line)
        self.absorb_properties(line, as_self=False)
        self.line_style_index += 1

        self.do_legend()
        self.update_canvas_size()

        self.time_entr.get_parent().set_visible(source.get_time_cols() is not None)

        self.set_active_line(line)

        return line
    #PROPERTIES
    def get_auto_aesthete_properties(self) :
        return {
            'grid' : ( bool, ),
            'highlight' : ( bool, ),
            'legend' : (bool, ),
            'grid_color' : (string_to_float_tup, ),
            'plot_over' : (bool, ),
            'title' : (str, ),
            'xmultiplier' : (float,),
            'ymultiplier' : (float,),
            'read_series' : (bool,),
            'resolution' : (int,),
            'legend_loc' : (str,),
            'axes_axis_bgcolor' : (string_to_float_tup,),
            'figure_facecolor' : (string_to_float_tup,),
            'axes_xmin' : (float,),
            'axes_xlabel' : (str,),
            'axes_xmax' : (float,),
            'axes_ymin' : (float,),
            'axes_ylabel' : (str,),
            'axes_ymax' : (float,),
            'xhide_oom' : (bool,),
            'yhide_oom' : (bool,),
            'xtick_font' : (str,),
            'ytick_font' : (str,),
            'xlabel_font' : (str,),
            'ylabel_font' : (str,),
            'title_font' : (str,),
            'default_line_style' : (str, (AOBJECT_CAN_NONE,),),
            'order' : (string_to_aobject_tup,
                       (AOBJECT_CAN_NONE,)),
            'size_inches' : (string_to_float_tup,),
            'subplots_region' : (string_to_float_tup,),
            'time' : (string_to_float_tup, (AOBJECT_CAN_NONE,)),
            'time_args' : (int, (AOBJECT_CAN_NONE,)),


                }

    managed_lines_lsst = None
    #BEGIN PROPERTIES FUNCTIONS
    def get_order(self) :
        if self.managed_lines_lsst is None :
            debug_print("No managed_lines_lsst, drat.")
            return None
        return tuple([r[2] for r in  self.managed_lines_lsst])
    def change_order(self, val) :
        num = len(val)
        if self.managed_lines_lsst is None :
            debug_print("No managed_lines_lsst, drat.")
            return
        new_order = [None]*len(self.managed_lines_lsst)
        i = 0
        for row in self.managed_lines_lsst :
            if row[2] in val :
                new_order[val.index(row[2])] = i
            else :
                new_order[num] = i
                num += 1
            i += 1

        #Thanks to Manuel Kaufmann
        rows = []
        for row in self.managed_lines_lsst :
            rows.append([row[0],row[1],row[2],row[3]])
        self.managed_lines_lsst.clear()
        for i in range(len(new_order)) :
            self.managed_lines_lsst.append(rows[new_order[i]])

        # FIXME: PyGI lacks self.managed_lines_lsst.reorder(new_order)

        self.do_legend()
        debug_print(val)
    def get_subplots_region(self) :
        spp = self.fig.subplotpars
        return (spp.left, spp.bottom, spp.right, spp.top)
    def change_subplots_region(self, val) :
        self.fig.subplots_adjust(*val)
        self.queue_draw()
    size_inches = (4.,3.)
    def get_size_inches(self) :
        return self.size_inches
    def change_size_inches(self, val) :
        #self.fig.set_size_inches(*val, forward=True)
        #self.size_inches = val
        #debug_print(self.fig.get_size_inches())
        #self.fig.canvas.set_size_request(*self.fig.canvas.get_size())
        #self.queue_draw()
        self.size_inches = val
        self.update_canvas_size()
    def change_legend(self, val) :
        self.do_legend(val)
    def change_grid(self, val) :
        self.grid = val
        if self.grid : 
            self.axes.grid(color=self.grid_color)
        else : 
            self.axes.grid()
        self.queue_draw()
    def get_grid_color(self): return mpl_to_tuple(self.grid_color)
    def change_grid_color(self, val) :
        self.grid_color = val
        if self.grid :
            self.axes.grid(b=self.grid, color=val)
        self.queue_draw()
    def get_xmultiplier(self) : return 1.
    def get_ymultiplier(self) : return 1.
    def get_resolution(self) :
        if self.lines == [] :
            res_check = get_preferencer().get_preference("GlancerPlot","resolution")
            if res_check is not None :
                return int(res_check)
            else :
                return 10
        return self.lines[0].get_resolution()
    def get_axes_axis_bgcolor(self):
        return mpl_to_tuple(self.axes.get_axis_bgcolor())
    def get_figure_facecolor(self):
        return mpl_to_tuple(self.fig.get_facecolor())
    def get_axes_ylabel(self) : return self.axes.get_ylabel()
    def get_axes_ymin(self) : return self.axes.get_ylim()[0]
    def get_axes_ymax(self) : return self.axes.get_ylim()[1]
    def get_axes_xlabel(self) : return self.axes.get_xlabel()
    def get_axes_xmin(self) : return self.axes.get_xlim()[0]
    def get_axes_xmax(self) : return self.axes.get_xlim()[1]
    def get_xhide_oom(self) :
        return False
    def get_yhide_oom(self) :
        return False
    def get_xtick_font(self) :
        tick_props = self.axes.get_xaxis().get_major_ticks()[0].label1.get_fontproperties()
        return mpl_to_font(tick_props)
    def get_ytick_font(self) :
        tick_props = self.axes.get_yaxis().get_major_ticks()[0].label1.get_fontproperties()
        return mpl_to_font(tick_props)
    def get_xlabel_font(self) :
        label_props = self.axes.get_xaxis().get_label().get_fontproperties()
        return mpl_to_font(label_props)
    def get_ylabel_font(self) :
        label_props = self.axes.get_yaxis().get_label().get_fontproperties()
        return mpl_to_font(label_props)
    def get_title_font(self) :
        label_props = self.axes.title.get_fontproperties()
        return mpl_to_font(label_props)
    #These are exactly as Aesthete will define them...
    #def get_plot_over(self): return self.plot_over if val==None else (val=='True')
    #def change_plot_over(self, val): self.plot_over = val
    #def change_read_series(self, val) : self.read_series = val

    def change_legend_loc(self, val) : self.do_legend(True, loc=val)
    def change_title(self, val) :
        self.title = val
        self.axes.set_title(self.title, visible = (self.title!=''))
        self.queue_draw()
    def change_resolution(self, val) :
        for line in self.lines : 
            line.set_resolution(val)
        self.queue_draw()
    def change_xhide_oom(self, val) : 
        self.axes.get_xaxis().major.formatter.hide_oom = val
        self.queue_draw()
    def change_yhide_oom(self, val) : 
        self.axes.get_yaxis().major.formatter.hide_oom = val
        self.queue_draw()
    def change_ytick_font(self, val) : 
        ticks = self.axes.get_yaxis().get_major_ticks()
        for tick in ticks :
            font_to_mpl(tick.label1.get_fontproperties(), val)
        self.queue_draw()
    def change_xtick_font(self, val) : 
        ticks = self.axes.get_xaxis().get_major_ticks()
        for tick in ticks :
            font_to_mpl(tick.label1.get_fontproperties(), val)
        self.queue_draw()
    def change_xlabel_font(self, val) : 
        label_props = self.axes.get_xaxis().get_label().get_fontproperties()
        font_to_mpl(label_props, val)
        self.queue_draw()
    def change_ylabel_font(self, val) : 
        label_props = self.axes.get_yaxis().get_label().get_fontproperties()
        font_to_mpl(label_props, val)
        self.queue_draw()
    def change_title_font(self, val) : 
        label_props = self.axes.title.get_fontproperties()
        font_to_mpl(label_props, val)
        self.queue_draw()
    def change_xmultiplier(self) : self.axes.xaxis.get_major_formatter().multiplier = val; self.queue_draw()
    def change_ymultiplier(self) : self.axes.yaxis.get_major_formatter().multiplier = val; self.queue_draw()
    def change_axes_axis_bgcolor(self, val) : self.axes.set_axis_bgcolor(val); self.queue_draw()
    def change_axes_xlabel(self, val) : self.axes.set_xlabel(val); self.queue_draw()
    def change_axes_xmin(self, val) : self.set_xlim(left=val); self.queue_draw()
    def change_axes_xmax(self, val) : self.set_xlim(right=val); self.queue_draw()
    def change_axes_ylabel(self, val) : self.axes.set_ylabel(val); self.queue_draw()
    def change_axes_ymin(self, val) : self.set_ylim(bottom=val); self.queue_draw()
    def change_axes_ymax(self, val) : self.set_ylim(top=val); self.queue_draw()
    def change_figure_facecolor(self, val) : self.fig.set_facecolor(val); self.queue_draw()
    #END PROPERTIES FUNCTIONS


    def _show_legend_config(self) :
        if self.legend_object is None :
            self.do_legend()
        if self.legend_object is not None :
            self.legend_object.to_action_panel()

    def do_set_default_line_style(self) :
        chooser = Gtk.Dialog(\
                      title="Pick default line style",
                             parent=self.get_aenv().toplevel, 
                      flags=Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT,
                      buttons=(Gtk.STOCK_CANCEL,Gtk.ResponseType.CANCEL,Gtk.STOCK_SAVE,Gtk.ResponseType.OK))
        chooser.set_default_response(Gtk.ResponseType.OK)

        vbox = Gtk.VBox()

        styles_lsst = Gtk.ListStore(str, str)
        styles = get_styles_by_aname_root('GlancerLine')
        for style in styles :
            styles_lsst.append([style,
                                get_style_from_dictionary(style).name_nice])

        style_cmbo = Gtk.ComboBox(styles_lsst)
        style_cllr = Gtk.CellRendererText()
        style_cmbo.pack_start(style_cllr, True, True, 0)
        style_cmbo.add_attribute(style_cllr, 'text', 1)
        vbox.pack_start(style_cmbo, True, True, 0)

        existing_chkb = Gtk.CheckButton("Apply to existing lines")
        existing_chkb.set_active(True)
        vbox.pack_start(existing_chkb, True, True, 0)

        chooser.get_content_area().add(vbox)
        chooser.get_content_area().show_all()

        resp = chooser.run()
        chooser.grab_focus()

        if resp == Gtk.ResponseType.OK :
            style = style_cmbo.get_active_iter()

            if style is not None :
                style = styles_lsst.get(style, 0)
                style = style[0]
                debug_print(style)

            self.set_default_line_style(style)

            if existing_chkb.get_active() :
                for line in self.lines :
                    line.aes_restyle(style)

            chooser.destroy()
        else :
            chooser.destroy()
            return

    def get_method_window(self) :
        #fram = Gtk.Frame()
        #fram.modify_bg(Gtk.StateType.NORMAL, Gdk.Color(1, 1, 1))
        win = Gtk.VBox()
        who_algn = Gtk.Alignment.new(0.5, 0.5, 1., 1.)
        who_algn.set_property("top_padding", 10)
        who_hbox = Gtk.HBox(spacing=5)
        who_hbox.pack_start(Gtk.Image.new_from_stock('aes-glancer-plot2d',
                                                     Gtk.IconSize.BUTTON),
                            False, False, 0)
        who_hbox.pack_start(Gtk.Label("2D Plot"), False, False, 0)
        who_algn.add(who_hbox)
        win.pack_start(who_algn, True, True, 0)

        icon_table = Gtk.Table(1, 5)
        win.pack_start(icon_table, True, True, 0)

        # Visual Config
        config_butt = Gtk.Button()
        config_butt.set_image(Gtk.Image.new_from_stock(Gtk.STOCK_PAGE_SETUP,
                                                       Gtk.IconSize.BUTTON))
        config_butt.set_tooltip_text("Appearance preferences...")
        icon_table.attach(config_butt, 0, 1, 0, 1)
        config_vbox = self.methods_make_visual_config()
        config_vbox.aes_title = "Configure plot appearance"
        config_butt.connect("clicked", lambda o :
                            self.env.action_panel.to_action_panel(config_vbox))

        # Import Config
        #legend_amtb = self.aes_method_toggle_button("legend", None,
        #                                           preferencable=False)
        legend_butt = Gtk.Button()
        legend_butt.set_image(Gtk.Image.new_from_stock(Gtk.STOCK_JUSTIFY_RIGHT,
                                                    Gtk.IconSize.BUTTON))
        legend_butt.connect("clicked", lambda o : self._show_legend_config())
        legend_butt.set_tooltip_text("Toggle legend")
        icon_table.attach(legend_butt, 1, 2, 0, 1)
        plot_over_amtb = self.aes_method_toggle_button("plot_over", None,
                                                     preferencable=False)
        plot_over_amtb.set_image(Gtk.Image.new_from_stock(Gtk.STOCK_DND_MULTIPLE,
                                                    Gtk.IconSize.BUTTON))
        plot_over_amtb.set_tooltip_text("Toggle overlay of new plots")
        icon_table.attach(plot_over_amtb, 2, 3, 0, 1)

        style_menu = Gtk.Menu()
        style_tbut = Gtk.ToggleButton()
        style_menu.attach_to_widget(style_tbut, None)
        style_tbut.set_image(Gtk.Image.new_from_stock(Gtk.STOCK_COLOR_PICKER,
                                                      Gtk.IconSize.BUTTON))
        style_tbut.set_tooltip_text("Style menu")
        style_tbut.connect("button_press_event", lambda w, e :\
            style_menu.popup(None, None, None, None, e.button, e.time) \
                if not w.get_active() else \
            style_menu.popdown())

        style_menu.set_title('Style menu')
        style_menu.modify_bg(Gtk.StateType.NORMAL, Gdk.Color(65535, 65535, 65535))

        pick_style_tbut = Gtk.MenuItem()
        pick_style_tbut.remove(pick_style_tbut.get_child())
        pick_style_tbut.add(Gtk.Image.new_from_stock(Gtk.STOCK_INDEX,
                                                      Gtk.IconSize.BUTTON))
        pick_style_tbut.set_tooltip_text("Set default line style")
        pick_style_tbut.connect("activate", lambda o : self.do_set_default_line_style())
        style_menu.append(pick_style_tbut)

        restyle_tbut = Gtk.MenuItem()
        restyle_tbut.remove(restyle_tbut.get_child())
        restyle_tbut.add(Gtk.Image.new_from_stock(Gtk.STOCK_COLOR_PICKER,
                                                      Gtk.IconSize.BUTTON))
        restyle_tbut.set_tooltip_text("Restyle lines by legend order")
        restyle_tbut.connect("activate", lambda o : self.restyle_by_order())
        style_menu.append(restyle_tbut)

        style_menu.show_all()
        icon_table.attach(style_tbut, 3, 4, 0, 1)

        clear_butt = Gtk.Button()
        clear_butt.set_image(Gtk.Image.new_from_stock(Gtk.STOCK_CLEAR,
                                                      Gtk.IconSize.BUTTON))
        clear_butt.set_tooltip_text("Clear all lines")
        icon_table.attach(clear_butt, 4, 5, 0, 1)
        clear_butt.connect("clicked", lambda o : self.check_clear(force=True))

        replot_butt = Gtk.Button()
        replot_butt.set_image(Gtk.Image.new_from_stock(Gtk.STOCK_REFRESH,
                                                       Gtk.IconSize.BUTTON))
        replot_butt.set_tooltip_text("Replot all lines")
        replot_butt.connect("clicked", lambda o : self.replot_all())
        icon_table.attach(replot_butt, 5, 6, 0, 1)

        self.managed_lines_lsst = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING,
                                 GObject.TYPE_PYOBJECT, GObject.TYPE_STRING)
        self.managed_lines_lsst.connect("rows-reordered",
                                        self._do_managed_lines_rows_reordered)
        lines_trvw = Gtk.TreeView(self.managed_lines_lsst)
        lines_scrw = Gtk.ScrolledWindow()
        lines_scrw.set_min_content_height(100)
        lines_scrw.add(lines_trvw)
        lines_trvw.set_size_request(-1, 300)
        lines_trvw.set_reorderable(True)
        lines_trvw.set_can_focus(True)
        self.lines_trvw = lines_trvw

        graphic_trvc = Gtk.TreeViewColumn()
        graphic_trvc.set_expand(False)
        graphic_cllr = LineGraphicCellRenderer()
        graphic_trvc.pack_start(graphic_cllr, True)
        graphic_trvc.add_attribute(graphic_cllr, 'aname', 0)
        lines_trvw.append_column(graphic_trvc)

        lines_trvc = Gtk.TreeViewColumn('Lines')
        lines_trvc.set_expand(True)
        lines_trvw.append_column(lines_trvc)
        lines_cllr = Widget.GlyphCellRenderer()
        lines_trvc.pack_start(lines_cllr, True)
        lines_trvc.set_cell_data_func(lines_cllr, lines_cdf)

        del_trvc = Gtk.TreeViewColumn()
        del_trvc.set_expand(False)
        self._managed_lines_del_trvc = del_trvc
        lines_trvw.append_column(del_trvc)
        del_cllr = Gtk.CellRendererPixbuf()
        del_cllr.set_property('stock-id', Gtk.STOCK_CLEAR)
        del_trvc.pack_start(del_cllr, False)

        lines_trvw.connect("row-activated",
                           self._do_managed_lines_row_activated)

        win.pack_start(lines_scrw, True, True, 0)

        #lines_trvw.connect("button-press-event", do_source_treeview_popup)
        #lines_trvw.connect("row-activated", lambda t, p, c : \
        #                     aobject.get_object_dictionary().try_active_source_action())
        #lines_trvw.connect("cursor-changed", lambda trvw :
        #                     aobject.get_object_dictionary().selected_source_change(\
        #                            lines_lsst.get_value(\
        #                                lines_lsst.get_iter(trvw.get_cursor()[0]), 0)))
        #lines_trvw.set_tooltip_column(3)
        #lines_lsst.connect("row-changed", connect_lines_reload_signal, lines_trvw, lines_cllr)
        #lines_lsst.connect("row-changed", \
        #        lambda tm, p, i : lines_trvw.set_cursor(p, lines_trvc))

        #fram.add(win)
        win.show_all()

        return win

    #FIXME : rows-reordered not fired on DND
    def _do_managed_lines_rows_reordered(self, lsst, path, it, new_order) :
        self.do_legend()
        debug_print("Order changed")

    def _do_managed_lines_row_activated(self, trvw, path, view_column) :
        aobj = self.managed_lines_lsst[path][2]
        if view_column == self._managed_lines_del_trvc :
            aobj.aes_remove()
        else :
            self.set_active_line(aobj)
        debug_print(view_column)

    def methods_make_visual_config(self) :
        config_vbox = Gtk.VBox()

        config_ntbk = Gtk.Notebook()

        general_table_maker = PreferencesTableMaker()
        general_table_maker.append_heading("Title")
        general_table_maker.append_row("Text",
            self.aes_method_entry_update("title"))
        general_table_maker.append_row("Font", self.aes_method_font_button("title_font", "Set title font"))
        general_table_maker.append_heading("Geometry")
        general_table_maker.append_row("Size (in)",
            self.aes_method_tuple_entry_update("size_inches"))
        general_table_maker.append_row("Plot area",
            self.aes_method_tuple_entry_update("subplots_region"))
        general_table_maker.append_heading("Colours")
        general_table_maker.append_row("Face", self.aes_method_colour_button("figure_facecolor", "Set figure colour"))
        general_table_maker.append_row("Axes",self.aes_method_colour_button("axes_axis_bgcolor", "Axes Background Colour"))
        general_table_maker.append_heading("Detail")
        general_table_maker.append_row("Resolution",
            self.aes_method_entry_update("resolution"),
            tip="Only applies to continuous Sources")
        general_table_maker.append_heading("Grid")
        general_table_maker.append_row("Show Grid", self.aes_method_check_button("grid", None))
        general_table_maker.append_row("Grid Colour",self.aes_method_colour_button("grid_color", "Grid Colour"))
       

        config_tabl = general_table_maker.make_table()
        config_tabl_vbox = Gtk.VBox(); config_tabl_vbox.pack_start(config_tabl, False, False, 0)
        config_ntbk.append_page(config_tabl_vbox, Gtk.Label(label="General"))

        axes = { 'x' : "X" , 'y' : "Y" }
        for axis in axes :
            axes_table_maker = PreferencesTableMaker()
            axes_table_maker.append_heading("Labeling")
            axes_table_maker.append_row(" Label",
                self.aes_method_entry_update("axes_"+axis+"label"))
            axes_table_maker.append_row(" Axis",
                self.aes_method_font_button(axis+"label_font", "Set "+axes[axis]+" axis font"))
            axes_table_maker.append_row(" Tick", self.aes_method_font_button(axis+"tick_font", "Set "+axes[axis]+" axis font"))
            axes_table_maker.append_heading("Data Limits")
            axes_table_maker.append_row(" Minimum Value",
                self.aes_method_entry_update("axes_"+axis+"min"))
            axes_table_maker.append_row(" Maximum Value",
                self.aes_method_entry_update("axes_"+axis+"max"))

            config_tabl = axes_table_maker.make_table()
            config_tabl_vbox = Gtk.VBox(); config_tabl_vbox.pack_start(config_tabl, False, False, 0); 
            config_ntbk.append_page(config_tabl_vbox, Gtk.Label(label=axes[axis]+" Axis"))

        config_vbox.pack_start(config_ntbk, True, True, 0)
        config_vbox.show_all()
        return config_vbox

    def do_mpl_scroll_event(self, event) :
        debug_print('hi')
        '''Handle scrolling ourselves.'''

        if event.inaxes != self.axes :
            return False

        self.axes.set_autoscale_on(False)
        xl = self.axes.get_xlim()
        yl = self.axes.get_ylim()

        ec = (event.xdata, event.ydata)

        # event.step tells direction
        spd = (1+self.scroll_speed) ** (-event.step)

        # unfortunately, this seems to be the only sensible way to
        # get to the modifiers. Phrased oddly, but says do_x if we're
        # not told to only do y, and v.v.
        do_specific = event.guiEvent.state & Gdk.ModifierType.CONTROL_MASK
        do_x = not (do_specific and (event.guiEvent.state & Gdk.ModifierType.SHIFT_MASK))
        do_y = not (do_specific and do_x)

        if do_x :
            self.axes.set_xlim(ec[0] - (ec[0]-xl[0])*spd,
                               ec[0] - (ec[0]-xl[1])*spd)
        if do_y :
            self.axes.set_ylim(ec[1] - (ec[1]-yl[0])*spd,
                               ec[1] - (ec[1]-yl[1])*spd)


        self.queue_draw()
        return True

    _move_type = None
    _move_object = None
    _move_from = None
    _move_from_xl = None
    _move_from_yl = None
    _move_from_data = None
    _focused_object = None
    _focused_object_type = None
    _button_press_time = None
    def do_mpl_button_press_event(self, event) :
        '''Check button presses.'''

        if event.inaxes != self.axes :
            return False
        
        m_control = event.guiEvent.state & Gdk.ModifierType.CONTROL_MASK
        move = False

        self._button_press_time = long(round(time.time()*1000))

        _object_type = 'axes'
        _object = None

        for ab in self.annotations :
            if _object is not None : break
            if ab.annotation.contains(event)[0] :
                _object_type = 'annotation'
                _object = ab
        for mark in self.marks :
            if _object is not None : break
            if mark.mark.contains(event)[0] :
                _object_type = 'mark'
                _object = mark

        for line in self.lines :
            if _object is not None : break
            for mark in line.marks :
                if mark.mark.contains(event)[0] :
                    _object_type = 'mark'
                    _object = mark
            if _object is None and line.line.contains(event)[0] :
                _object_type = 'line'
                _object = line

        if _object is None and self.legend_object is not None\
                and self.legend_object.axes.legend_ is not None \
                and self.legend_object.axes.legend_.legendPatch.contains(event)[0] :
            _object = self.legend_object
            _object_type = 'legend'

        if event.button == 1 :
            if _object_type == 'annotation' or _object_type == 'mark':
                _object.set_highlight(True)
                move = True
                self._move_from_data = _object.get_xy()
            else :
                self.axes.set_autoscale_on(False)
                self._move_from_xl = self.axes.get_xlim()
                self._move_from_yl = self.axes.get_ylim()
                move = True
        elif event.button == 2 :
            if m_control :
                self.axes.autoscale_view()
                self.axes.set_autoscale_on(True)
                self.queue_draw()
            elif _object is not None :
                _object.to_action_panel()

            self.queue_draw()
            return True

        if move :
            self._move_from = (event.x, event.y)
            self._move_type = _object_type
            self._move_object = _object
        self._focused_object = _object
        self._focused_object_type = _object_type

        return False

    def change_highlight(self, val) :
        self.highlight = val
        line = self.get_active_line()
        if line is not None :
            line.set_highlight(val)

    _active_line = None
    def get_active_line(self) :
        return self._active_line

    def set_active_line(self, line) :
        if line is not None :
            line.to_action_panel()

        if line == self._active_line :
            return True

        if line is None :
            self._active_line = None
        elif line not in self.lines :
            return False
        else :
            if self._active_line is not None :
                self._active_line.set_highlight(False)
            line.set_highlight(self.get_highlight())
            self._active_line = line

        self.emit('active-line-changed', line.get_aname())

        return True

    def get_last_mouse_location(self) :
        return (self._last_location, self._last_location_type)

    def make_insert_mark_dialog(self, xy=None) :
        ann_dial = Gtk.Dialog(\
                      title="Add mark", parent=self.get_aenv().toplevel, 
                      flags=Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT,
                      buttons=(Gtk.STOCK_CANCEL,Gtk.ResponseType.CANCEL,Gtk.STOCK_OK,Gtk.ResponseType.OK))
        ann_dial.set_default_response(Gtk.ResponseType.OK)

        vbox = Gtk.VBox()

        pref_tab = PreferencesTableMaker()

        text_entr = Gtk.Entry()
        pref_tab.append_row('Marker', text_entr)
        xy_entr = Gtk.Entry()
        if xy is not None :
            xy_entr.set_text(str(xy))
        pref_tab.append_row('Location', xy_entr)

        vbox.pack_start(pref_tab.make_table(), True, True, 0)
        ann_dial.get_content_area().add(vbox)
        ann_dial.get_content_area().show_all()

        resp = ann_dial.run()
        ann_dial.grab_focus()

        if resp == Gtk.ResponseType.OK :
            text = text_entr.get_text()
            xy = string_to_float_tup(xy_entr.get_text())
            ann_dial.destroy()
        else :
            ann_dial.destroy()
            return

        self.insert_mark(text, xy)

    def make_insert_annotation_dialog(self, xy=None) :
        ann_dial = Gtk.Dialog(\
                      title="Add annotation", parent=self.get_aenv().toplevel, 
                      flags=Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT,
                      buttons=(Gtk.STOCK_CANCEL,Gtk.ResponseType.CANCEL,Gtk.STOCK_OK,Gtk.ResponseType.OK))
        ann_dial.set_default_response(Gtk.ResponseType.OK)

        vbox = Gtk.VBox()

        pref_tab = PreferencesTableMaker()

        text_entr = Gtk.Entry()
        pref_tab.append_row('Text', text_entr)
        xy_entr = Gtk.Entry()
        if xy is not None :
            xy_entr.set_text(str(xy))
        pref_tab.append_row('Text location', xy_entr)
        xycoords_cmbo = Gtk.ComboBoxText.new_with_entry()
        xycoords_cmbo.append_text('data')
        xycoords_cmbo.append_text('figure points')
        xycoords_cmbo.append_text('figure pixels')
        xycoords_cmbo.append_text('figure fraction')
        xycoords_cmbo.append_text('axes points')
        xycoords_cmbo.append_text('axes pixels')
        xycoords_cmbo.set_active(0)
        pref_tab.append_row('Coordinates', xycoords_cmbo)

        vbox.pack_start(pref_tab.make_table(), True, True, 0)
        ann_dial.get_content_area().add(vbox)
        ann_dial.get_content_area().show_all()

        resp = ann_dial.run()
        ann_dial.grab_focus()

        if resp == Gtk.ResponseType.OK :
            text = text_entr.get_text()
            xy = string_to_float_tup(xy_entr.get_text())
            xycoords = xycoords_cmbo.get_active_text()
            ann_dial.destroy()
        else :
            ann_dial.destroy()
            return

        self.insert_annotation(text, xy, xycoords)

    def _make_context_menu(self, event, supplement=None) :
        if self._button_press_time is None :
            return
        context_menu = Gtk.Menu()

        if supplement is not None :
            supplement.supplement_context_menu(context_menu)
            context_menu.add(Gtk.SeparatorMenuItem())
        else :
            context_menu_aname_nice = Gtk.MenuItem(self.get_aname_nice())
            context_menu_aname_nice.set_sensitive(False)
            context_menu.add(context_menu_aname_nice)
        
        if self._focused_object is not None and self._focused_object != supplement:
            self._focused_object.supplement_context_menu(context_menu)
            context_menu.add(Gtk.SeparatorMenuItem())

        context_menu_rename = Gtk.MenuItem("Annotate...")
        context_menu_rename.connect("button-press-event", lambda o, e :
                                    self.make_insert_annotation_dialog(self._last_location))
        context_menu.add(context_menu_rename)
        context_menu_mark = Gtk.MenuItem("Mark...")
        context_menu_mark.connect("button-press-event", lambda o, e :
                                    self.make_insert_mark_dialog(self._last_location))
        context_menu.add(context_menu_mark)

        context_menu.popup( None, None, None, None, event.guiEvent.button,
                           event.guiEvent.time)
        context_menu.show_all()
        debug_print(self._button_press_time)

    def do_mpl_button_release_event(self, event) :
        '''Check button releases.'''

        if event.button == 2 or event.button == 1:
            if self._move_type == 'annotation' or self._move_type == 'mark' :
                self._move_object.set_highlight(False)

            self._move_type = None
            self._move_object = None
            self._move_from = None
            self._move_from_xl = None
            self._move_from_yl = None
            self._move_from_data = None

            self.queue_draw()
            return True
        elif event.button == 3 :
            if self.get_highlight() and self.get_active_line() is not None :
                self._make_context_menu(event, supplement=self.get_active_line())
            else :
                self._make_context_menu(event)

        self._button_press_time = None

        return False

    def do_mpl_motion_notify_event(self, event) :
        '''Check motion notifications.'''

        if event.inaxes != self.axes :
            return False

        do_specific = event.guiEvent.state & Gdk.ModifierType.CONTROL_MASK
        do_x = not (do_specific and (event.guiEvent.state & Gdk.ModifierType.SHIFT_MASK))
        do_y = not (do_specific and do_x)

        self._last_location = (event.xdata, event.ydata)
        self._last_location_type = 'normal'

        if self.get_highlight() :
            line = self.get_active_line()
            if line is not None :
                line.set_location(self._last_location)
                dloc = line.get_data_location()
                self._last_location = (dloc[1],dloc[2])
                self._last_location_type = 'target'
            self.queue_draw()
        elif self._move_type == 'axes' :
            dx = (event.x-self._move_from[0])
            dy = (event.y-self._move_from[1])
            l,b,r,t = self.axes.bbox.extents
            el,er = self.axes.get_xlim()
            eb,et = self.axes.get_ylim()
            dx = dx*(er-el)/(r-l)
            dy = dy*(et-eb)/(t-b)

            x, y = (0,0)
            if do_x :
                self.axes.set_xlim(self._move_from_xl[0]-dx,
                                   self._move_from_xl[1]-dx)
                x = dx
            if do_y :
                self.axes.set_ylim(self._move_from_yl[0]-dy,
                                   self._move_from_yl[1]-dy)
                y = dy

            self._last_location = (dx, dy)
            self._last_location_type = 'target'

            self.queue_draw()
        elif self._move_type == 'annotation' or self._move_type == 'mark' :
            dx = (event.x-self._move_from[0])
            dy = (event.y-self._move_from[1])
            l,b,r,t = self.axes.bbox.extents
            el,er = self.axes.get_xlim()
            eb,et = self.axes.get_ylim()
            dx = dx*(er-el)/(r-l)
            dy = dy*(et-eb)/(t-b)

            x = self._move_from_data[0]+dx
            y = self._move_from_data[1]+dy
            self._move_object.set_xy((x,y))

            self._last_location = (x,y)
            self._last_location_type = 'target'
            self.emit('location-changed')

            self.queue_draw()

        self.emit('location-changed')
        return True
