from ..glypher.Widget import GlyphBasicGlypher
import uuid, subprocess
import sys
from StringIO import StringIO
from aobject.utils import debug_print
from ..glypher.PhraseGroup import GlypherBracketedPhrase
from gi.repository import Gtk
from aobject.aobject import *
from ..glypher.Word import make_word
from ..glypher import Dynamic
from ..glypher import glypher as g
from sympy.core.symbol import Symbol
from ..sources.Source import Source
import lxml.etree as ET

import Command

class UFLEditor(Gtk.Frame, AObject) :
    line_height = 35.
    default_height = 560
    default_width = 360
    can_save = True

    def do_aes_save(self, uri=None) :
        AObject.do_aes_save(self, uri)

        if uri.endswith('.ufl.aes') :
            ufl_uri = uri[:-8]+'.aes.ufl'
        else :
            ufl_uri = uri+'.ufl'

        gf = Gio.File(uri=ufl_uri)
        f = gf.replace('', False)
        f.write(self.gmg.main_phrase.to_string('ufl'))
        f.close()

    def get_aesthete_xml(self) :
        return self.gmg.get_xml()

    def set_aesthete_xml(self, xml) :
        #xml = ET.ElementTree(ET.fromstring(xml))
        self.gmg.set_xml(xml)

    def _do_python_view_key_press_event(self, python_view, event):
        keyname = Gdk.keyval_name(event.keyval)
        m_control = bool(event.get_state() & Gdk.ModifierType.CONTROL_MASK)
        
        if m_control and keyname=='Escape' :
            self.toggle_python_view()
            return True

        return False

    def _do_gmg_key_press_event(self, gmg, event):
        keyname = Gdk.keyval_name(event.keyval)
        m_control = bool(event.get_state() & Gdk.ModifierType.CONTROL_MASK)
        
        if m_control and keyname=='Escape' :
            self.toggle_python_view()
            return True

        ret = GlyphUFLInside_.do_key_press_event(gmg, event)
        return ret

    def toggle_python_view(self) :
        if self.python_view_sw.get_visible() :
            self.python_view_sw.hide()
            self.gmg_sw.show()
            self.gmg.grab_focus()
        else :
            py = self.gmg.main_phrase.to_string("ufl")
            self.python_view.get_buffer().set_text(py)
            self.python_view_sw.show()
            self.gmg_sw.hide()
            self.python_view.grab_focus()

    def __init__(self, env = None):
        self.gmg = GlyphUFLInside_(env=env, evaluable=False)

        GObject.GObject.__init__(self)
        AObject.__init__(self, "UFLEditor", env, view_object=True)
        self.connect("aesthete-property-change", lambda o, p, v, a : self.queue_draw())
        self.set_aname_nice('UFL')

        self.vbox = Gtk.VBox()

        self.gmg.do_key_press_event = lambda e :\
            self._do_gmg_key_press_event(self.gmg, e)
        self.gmg.caret.shortcut_families = ("FamilyUFL", "FamilyGlassmaker", "FamilyGeneral")
        self.gmg.main_phrase.move()
        self.gmg.main_phrase.line_length = self.default_width
        self.absorb_properties(self.gmg)
        sw = Gtk.ScrolledWindow()
        sw.add_with_viewport(self.gmg)
        self.vbox.pack_start(sw, True, True, 0)
        self.gmg_sw = sw

        self.python_view = Gtk.TextView()
        self.python_view.connect("key-press-event",
                                 self._do_python_view_key_press_event)
        sw = Gtk.ScrolledWindow()
        sw.add_with_viewport(self.python_view)
        self.vbox.pack_start(sw, True, True, 0)
        self.python_view_sw = sw

        sw = Gtk.ScrolledWindow()
        self.output_view = Gtk.TextView()
        sw.add(self.output_view)
        self.vbox.pack_start(sw, False, False, 0)
        sw.set_size_request(-1, 100)
        self.output_view.get_buffer().set_text("UFL OUTPUT")

        self.add(self.vbox)
        self.show_all()

        self.python_view_sw.hide()

        self.log(1, "New UFLEditor initalized")

    def compile_current(self) :
        ufl_code = self.gmg.main_phrase.to_string("ufl")

        tmp_name = 'aes_%s.ufl' % uuid.uuid4()
        f = open('/tmp/%s'%tmp_name, 'w')
        f.write(ufl_code)
        f.close()

        sp = subprocess.Popen(["ffc", "-l", "dolfin","/tmp/%s"%tmp_name],stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        buff, err_buff = sp.communicate()
        sp.wait()
        self.output_view.get_buffer().set_text(buff+"\n----\n"+err_buff)
        
        if sp.returncode == 0 and self.current_uri is not None :
            if self.current_uri.endswith('.aes.ufl') :
                h_uri = self.current_uri[:-8]+'.h'
            elif self.current_uri.endswith('.ufl') :
                h_uri = self.current_uri[:-4]+'.h'
            else :
                h_uri = self.current_uri+'.h'
            tmp_h_name = tmp_name[:-4]+'.h'
            header_file = open(tmp_h_name, 'r')
            header_string = header_file.read()
            header_file.close()

            hf = Gio.File(h_uri)
            f = hf.replace('', False)
            f.write(header_string)
            f.close()
        else :
            self.log(0, 'FFC exited code %d' % sp.returncode)

    def get_method_window(self) :
        win = Gtk.VBox()
        
        run = Gtk.Button('Compile')
        win.pack_start(run, False, False, 0)
        run.connect("clicked", lambda w : self.compile_current())

        win.show_all()

        return win

class GlyphUFLInside_(GlyphBasicGlypher) :
    def reset_main_phrase(self, block=None) :
        GlyphBasicGlypher.reset_main_phrase(self)

        self.caret.remove_boxes()

        if block is None :
            self.block = Command.GlassmakerBlock(self.main_phrase,
                                       num_ops=1)
        else :
            self.block = block

        self.block.set_deletable(False)
        self.block.set_highlight_group(False)

        self.main_phrase.set_attachable(False)
        self.main_phrase.set_enterable(False)
        self.main_phrase.append(self.block)
        self.block.set_recommending(self.block.get_target('pos0'))
        self.caret.try_suggestion()

aname_root_catalog['UFLEditor'] = UFLEditor
g.to_string_fallbacks['ufl'] = 'python'
