from ..glypher.Widget import GlyphBasicGlypher
import pickle
from ..glypher import glypher as g
from aobject.utils import debug_print
from ..glypher.PhraseGroup import GlypherBracketedPhrase
from ..glypher.Entity import GlypherEntity
from gi.repository import Gtk
from aobject.aobject import *
from ..glypher.Word import make_word
from ..glypher.Phrase import GlypherPhrase
from sympy.core.symbol import Symbol
from sympy import sympify
from ..sources.Source import Source
from ..glypher.Widget import display

class GlypherSourceReference(GlypherBracketedPhrase) :
    toolbox = {'symbol' : '[S]',
               'category' : 'References',
               'shortcut' : None,
               'alternatives' : 'source_reference',
               'priority' : None }
    source = None

    lhs_target = 'expression'
    rhs_target = 'source'

    def __init__(self, parent, source=None, name=None) :
        GlypherBracketedPhrase.__init__(self, parent, bracket_shapes=('[',']'), auto=False)
        self.mes.append('reference')
        self.mes.append('source_reference')

        # We store source (inefficiently) in a target phrase as this allows
        # automatic handling for alternatives
        source_phr = GlypherPhrase(self)
        self.add_target(source_phr, 'source')
        source_phr.adopt(make_word(source, self))

        source_name = make_word(name, self)
        source_name.set_bold(True)
        source_name.set_auto_italicize(False)
        source_name.set_italic(False)

        self['expression'].adopt(source_name)

        self.set_recommending(self["expression"])
        self['expression'].set_font_size_scaling(0.6)
        source_name.set_enterable(False)
        self['expression'].IN().set_enterable(False)
        self.set_rgb_colour((0.5, 0.4, 0))
    
    def get_xml(self, name=None, top=True, targets=None, full=False) :
        root = GlypherEntity.get_xml(self, name=name, top=top,
                                              targets=targets, full=full)
        
        root.set('type', 'source_reference')
        root.set('gluer_id', self['source'].to_string())
        root.set('display_name', self['expression'].to_string())

        return root

    @classmethod
    def parse_element(cls, parent, root, names, targets, operands, recommending, lead,
                  add_entities, am=None, top=True, args=None) :

        source = root.get('gluer_id')
        name = root.get('display_name')
        fn = cls(parent, source, name)

        return fn

    def get_sympy(self) :
        return Symbol('gluer_'+str(self['source'].to_string()))

    def get_python(self, true_mode="python") :
        return 'gluer_'+self['source'].to_string()

class GlypherSourceEmbedding(GlypherSourceReference) :
    toolbox = {'symbol' : '<S>',
               'category' : 'Matrices',
               'shortcut' : None,
               'alternatives' : 'source_reference',
               'priority' : None }
    source = None

    def __init__(self, parent, source=None, name=None) :
        GlypherSourceReference.__init__(self, parent, source, name)
        # Change self from a source_reference to a source_embedding
        self.mes[-1] = 'source_embedding'
        self.set_rgb_colour((0.5, 0.0, 0.4))
    
    def get_xml(self, name=None, top=True, targets=None, full=False) :
        root = GlypherEntity.get_xml(self, name=name, top=top,
                                              targets=targets, full=full)
        
        root.set('type', 'source_embedding')
        root.set('gluer_id', str(self['source'].to_string()))
        root.set('display_name', str(self['source'].to_string()))

        return root

    @classmethod
    def parse_element(cls, parent, root, names, targets, operands, recommending, lead,
                  add_entities, am=None, top=True, args=None) :

        source = root.get('gluer_id')
        name = root.get('display_name')
        fn = cls(parent, source, name)

        return fn

    def get_sympy(self) :
        print self['source'].to_string()
        return Symbol('gluee_'+str(self['source'].to_string()))

    def get_python(self, true_mode="python") :
        return 'gluee_'+self['source'].to_string('python')

class Gluer(Gtk.VBox, AObject) :
    line_height = 35.
    default_height = 560
    default_width = 360

    def get_aesthete_xml(self) :
        xml = self.gmg.get_xml()
        fs = float(self.gmg.main_phrase.get_scaled_font_size())
        xml.getroot().set('width', str(self.gmg.main_phrase.get_width()/fs))
        xml.getroot().set('height', str(self.gmg.main_phrase.get_height()/fs))
        return xml

    def get_gluer_xml(self, xml) :
        self.gmg.set_xml(xml)

    def aes_get_parameters(self) :
        return dict([('gluer_%d'%s, get_object_from_dictionary(self.sources[s])) for s in self.sources])

    def get_auto_aesthete_properties(self) :
        return {
            'base' : (get_object_from_dictionary, (AOBJECT_CAN_NONE,))
        }

    @classmethod
    def aes_load_a(cls, env=None, **params) :
        gluer = cls(env)

        for param in params :
            gluer.add_source(params[param], num=int(param[len('gluer_'):]))

        return gluer

    def __init__(self, env=None):
        GObject.GObject.__init__(self)
        self.gmg = GlyphBasicGlypher(env=env, evaluable=False)
        self.gmg_x = GlyphBasicGlypher(env=env, evaluable=False)
        AObject.__init__(self, "Gluer", env, view_object=True)
        self.connect("aesthete-property-change", lambda o, p, v, a : self.queue_draw())

        self.vbox = Gtk.VBox()

        self.sources = {}

        for gmg in (self.gmg, self.gmg_x) :
            gmg.main_phrase.move()
            gmg.main_phrase.set_enterable(True)
            gmg.main_phrase.set_attachable(True)
            gmg.main_phrase.line_length = self.default_width
            self.absorb_properties(gmg)
            self.vbox.pack_start(gmg, True, True, 0)
        self.gmg.modify_bg(Gtk.StateType.NORMAL, Gdk.Color('#FFDDDD'))
        self.gmg_x.modify_bg(Gtk.StateType.NORMAL, Gdk.Color('#FFFFDD'))

        self.source_action = self.insert_source_ref

        self.add(self.vbox)
        self.show_all()

        self.log(1, "New Gluer initalized")

    sources = None
    def insert_source_ref(self, source) :
        num = self.add_source(source)
        obj = get_object_from_dictionary(source)
        ref = GlypherSourceReference(None, str(num),
                                     obj.get_aname_nice())
        self.gmg.caret.insert_entity(ref)
        self.gmg.grab_focus()

    def add_source(self, source, num=None) :
        if num is None :
            num = len(self.sources)
        self.sources[num] = source

        if self.base_cmbo.get_active_iter() is None :
            mdl = self.base_cmbo.get_model()
            it = mdl.get_iter_first()
            while it is not None and \
                    mdl.get_value(it,0) != source :
                it = mdl.iter_next(it)

            if it is not None :
                self.base_cmbo.set_active_iter(it)

        return num

    def get_method_window(self) :
        win = Gtk.VBox()

        make_hbox = Gtk.HBox()
        make_hbox.set_spacing(5)
        make_entr = Gtk.Entry()
        make_entr.set_tooltip_text("Name for new Source")
        make_entr.set_width_chars(10)
        make_hbox.pack_start(make_entr, True, True, 0)
        make_butt = Gtk.Button()
        make_butt.set_tooltip_text("Generate Source")
        make_butt.set_image(Gtk.Image.new_from_stock(Gtk.STOCK_GOTO_LAST,
                                                     Gtk.IconSize.BUTTON))
        make_butt.connect("clicked", lambda b : self.make_new_source(make_entr.get_text()))
        make_hbox.pack_start(make_butt, False, False, 0)
        make_hbox.pack_start(Gtk.Label('Base', True, True, 0), False)
        base_cmbo = Gtk.ComboBox( get_object_dictionary().get_liststore_by_am('Source') )
        base_cmbo.set_size_request(100, -1)
        make_hbox.pack_start(base_cmbo, True, True, 0)
        base_cllr = Gtk.CellRendererText(); base_cmbo.pack_start(base_cllr, True, True, 0); base_cllr.props.ellipsize = Pango.EllipsizeMode.END;
        base_cmbo.add_attribute(base_cllr, 'text', 1)
        self.aes_method_automate_combo(base_cmbo, 'base', 0)
        self.base_cmbo = base_cmbo

        win.pack_start(make_hbox, False, False, 0)

        icon_table = Gtk.Table(1, 1)

        sim_butt = Gtk.Button()
        sim_butt.set_image(Gtk.Image.new_from_stock(Gtk.STOCK_INDEX,
                                                    Gtk.IconSize.BUTTON))
        sim_butt.set_tooltip_text("Reference currently selected Source")
        sim_butt.set_sensitive(False)
        icon_table.attach(sim_butt, 0, 1, 0, 1)
        sim_butt.connect("clicked", lambda o : self.insert_source_ref(\
                            get_object_dictionary().selected_source))
        get_object_dictionary().connect(\
            'aesthete-selected-source-change',
            lambda tr : sim_butt.set_sensitive(True))


        win.pack_start(icon_table, False, False, 0)

        win.show_all()

        return win

    def make_new_source(self, aname_nice) :
        sy = self.gmg.get_sympy()
        xsy = self.gmg_x.get_sympy()

        base_source = self.get_base()
        if base_source is None :
            return
        #base_source = \
        #    self.base_cmbo.get_model().get_value(base_source, 0)

        cs = ComposedSource('ComposedSource', sympy_object=sy,
                            x_sympy_object=xsy, sources=self.sources.copy(),
                            xml=self.get_aesthete_xml(), base=base_source)
        cs.set_aname_nice(aname_nice)
        self.gmg.grab_focus()

tol = 1e-7
class ComposedSource(Source) :
    resolution = None
    sympy_object = None
    x_sympy_object = None

    def get_x_sympy_object_pickled(self) :
        return pickle.dumps(self.x_sympy_object)

    def change_x_sympy_object_pickled(self, var) :
        f = pickle.loads(var)
        self.x_sympy_object = f
        self.source_reload()

    def get_sympy_object_pickled(self) :
        return pickle.dumps(self.sympy_object)

    def change_sympy_object_pickled(self, var) :
        f = pickle.loads(var)
        self.sympy_object = f
        self.source_reload()

    def aes_get_parameters(self) :
        d = dict([('gluer_%d'%s, get_object_from_dictionary(self.sources[s])) for s in self.sources])
        return d
        #d['sympy_object_pickled'] = self.get_sympy_object_pickled()

    def get_auto_aesthete_properties(self) :
        d = Source.get_auto_aesthete_properties(self)
        d.update({ \
            'base' : (get_object_from_dictionary, (AOBJECT_CAN_NONE,)),
            'sympy_object_pickled' : (str, (AOBJECT_CAN_NONE,)),
            'x_sympy_object_pickled' : (str, (AOBJECT_CAN_NONE,)),
            'gluer_xml' : (cast_xml, (AOBJECT_CAN_NONE,)),
        })
        return d

    @classmethod
    def aes_load_a(cls, env=None, **params) :
        sources = filter(lambda p : p.startswith('gluer_'), params)
        sources = dict((int(s[len('gluer_')]), params[s]) for s in sources)
        #sympy_object = pickle.loads(parameters['sympy_function_pickled'])

        comp_src = cls('ComposedSource', sources, env=env)

        #for param in params :
        #    gluer.add_source(params[param], num=int(param[len('gluer_'):]))

        return comp_src

    def __init__(self, stem, sources, sympy_object=None, x_sympy_object=None, base=None, xml=None, env=None, show = False, reloadable = False, resolution = 300) :
        self.resolution = resolution
        self.sources = sources
        self.sympy_object = sympy_object
        self.x_sympy_object = x_sympy_object
        Source.__init__(self, stem, env, show, reloadable)
        self.set_base(base)
        self.set_gluer_xml(xml)

    def change_base(self, base) :
        self.base = base
        self.source_reload()

    def source_type(self) :
        return self.base.source_type()

    def get_domain_cols(self) :
        if self.get_base() is None :
            return (0,)
        return tuple(range(len(self.get_base().get_domain_cols())))

    def source_get_values(self, time = None, multi_array = False, x_range =
                          (0,1), resolution = 300) :

        #FIXME: compat hack
        if self.sympy_object is None or self.x_sympy_object is None :
            if self.get_aesthete_xml() is not None :
                obj = display.make(self.get_aesthete_xml())
                print obj.to_string()
                print obj.get_sympy()
                self.change_sympy_object_pickled(pickle.dumps(obj.get_sympy()))
                xobj = sympify("x_0")
                self.change_x_sympy_object_pickled(pickle.dumps(xobj))
            print "%s has no sympy object yet!" % self.get_aname_nice()
        debug_print(self.x_sympy_object.args)
        debug_print(self.x_sympy_object)
        debug_print(self.sympy_object.args)
        debug_print(self.sympy_object)
        sub_dict = {}
        sub_dict_dom = {}
        for arg in list(self.sympy_object.atoms(Symbol))+list(self.x_sympy_object.atoms(Symbol)) :
            name = None
            if str(arg).startswith('gluer_') :
                name = str(arg)[len('gluer_'):]
            if str(arg).startswith('gluee_') :
                name = str(arg)[len('gluee_'):]
            if name is not None :
                sub_dict[name] =\
                    get_object_from_dictionary(self.sources[int(name)])

        #if len(sub_dict) == 0:
        #    raise RuntimeError('Need some other Source!')

        basis_source = self.get_base()
        if basis_source is None :
            return

        bvals = basis_source.source_get_values(time=time)[0]

        value_dict = {}
        for source in sub_dict :
            value_dict[source] =\
            sub_dict[source].source_get_values(time=time)[0]

        p = []
        xarr = []
        xaarr = []

        for source in value_dict :
            if len(value_dict[source]['values']) < len(bvals['values']) :
                raise RuntimeError( \
                    """Components don\'t match!
                    (length %d for %s < %d for %s)""" % \
                    (len(value_dict[source]), sub_dict[source].get_aname_nice(),
                    len(bvals), basis_source.get_aname_nice()))

        #cols = basis_source.get_domain_cols()
        for i in range(0, len(bvals['xa'][0])) :
            sub_val_dict = {}
            xa = bvals['xa']
            for j in range(len(xa)) :
                sub_val_dict['x_%d' % j] = xa[j][i]
            for source in value_dict :
                xs = value_dict[source]['xa']
                if True in [abs(xs[j][i]-xa[j][i]) > tol for j in range(len(xa))] :
                    raise RuntimeError(
                        """Components don\'t match!
                        (%s vs %s at point %d)""" %\
                        (sub_dict[source].get_aname_nice(),
                         basis_source.get_aname_nice(), i))
                ys = value_dict[source]['values']

                if isinstance(ys[0], list) :
                    sub_val_dict['gluer_'+source] = ys[0][i]
                    for j in range(len(ys)) :
                        sub_val_dict['gluer_%s_%d'%j] = ys[j][i]
                else :
                    sub_val_dict['gluer_'+source] = ys[i]
            #sub_val_dict['gluee_'+source] = value_dict[source]
            x = self.x_sympy_object.subs(sub_val_dict)
            y = self.sympy_object.subs(sub_val_dict)
            p.append(float(y))
            xarr.append(float(x))
            xaarr.append((float(x),))

        if multi_array :
            return [{'values':p,'xa':xaarr,'x':xarr,'name':self.get_aname_nice()}]
        return p

    def source_get_max_dim(self) :
        return 1 + len(self.get_domain_cols())

aname_root_catalog['Gluer'] = Gluer.aes_load_a
aname_root_catalog['ComposedSource'] = ComposedSource.aes_load_a
g.add_phrasegroup_by_class('source_reference', GlypherSourceReference)
g.add_phrasegroup_by_class('source_embedding', GlypherSourceEmbedding)
