from aobject import paths
from PhraseGroup import GlypherPhraseGroup
from Phrase import GlypherPhrase
from Spacer import GlypherSpace
import cairo
import glypher as g
from aobject.utils import debug_print, debug_print_stack

class GlypherPNGEntity(GlypherSpace) :
    def __init__(self, filename, parent=None) :

        self.surface_pattern =\
            cairo.SVGSurface.create_from_png(paths.get_share_location()+'images/'+filename)
        self.surface_pattern = cairo.SurfacePattern(self.surface_pattern)

        sw = self.surface_pattern.get_surface().get_width()
        sh = self.surface_pattern.get_surface().get_height()
        
        GlypherSpace.__init__(self, parent, dims=(1,float(sh)/sw),
                              as_line_height_scale=True, blank=False)
        self.mes.append('svg_entity')
    def draw(self, cr) :
        if not self.get_visible() or self.get_blank() :
            return
        cr.save()
        #l, t = self.get_local_offset()
        l = self.config[0].bbox[0]+self.padding[0]
        t = self.config[0].bbox[1]+self.padding[1]
        cr.translate(l, t)
        w = self.get_ref_width() - self.padding[2]
        h = self.get_ref_height() - self.padding[3]
        sw = self.surface_pattern.get_surface().get_width()
        sh = self.surface_pattern.get_surface().get_height()
        
        #cr.move_to(l, t)
        cr.scale(w/sw, w/sw)
        cr.rectangle(0, 0, sw, sh)
        cr.set_source(self.surface_pattern)
        cr.fill()
        cr.restore()

class GlypherSVGEntity(GlypherSpace) :
    def __init__(self, filename, parent=None) :

        raise NotImplementedError("This entity is unavailable until we can rely on gir version of rsvg")
        self.rsvg_handle =\
            Rsvg.Handle(file=paths.get_share_location()+'images/'+filename)

        sw = self.rsvg_handle.props.width
        sh =  self.rsvg_handle.props.height        
        GlypherSpace.__init__(self, parent, dims=(1,float(sh)/sw),
                              as_line_height_scale=True, blank=False)
        self.mes.append('svg_entity')
        
    def draw(self, cr) :
        if not self.get_visible() or self.get_blank() :
            return
        cr.save()
        #l, t = self.get_local_offset()
        l = self.config[0].bbox[0]+self.padding[0]
        t = self.config[0].bbox[1]+self.padding[1]
        cr.translate(l, t)
        w = self.get_ref_width() - self.padding[2]
        h = self.get_ref_height() - self.padding[3]
        sw = self.rsvg_handle.props.width
        sh =  self.rsvg_handle.props.height
        
        #cr.move_to(l, t)
        cr.scale(w/sw, w/sw)
        cr.rectangle(0, 0, sw, sh)
        self.rsvg_handle.render_cairo(cr)
        cr.restore()

class GlypherSVGPhraseGroup(GlypherPhraseGroup) :
    def __init__(self, parent, filename) :
        GlypherPhraseGroup.__init__(self, parent, [], area=(0,0,0,0),
                                    phraser=GlypherPhrase, highlight_group=False)
        self.mes.append('svg_phrasegroup')
        self.filename = filename
        svg_entity = GlypherSVGEntity(filename, self)
        self.append(svg_entity)
        self.set_enterable(False)
        self.set_recommending(self)

    def get_sympy(self) :
        return None

    @classmethod
    def parse_element(cls, parent, root, names, targets, operands, recommending, lead,
                  add_entities, am = None, top = True, args=None) :
        filename = root.get('file')
        return cls(parent, 'phrasegroups/'+filename)

    def get_xml(self, name=None, top=True, targets=None, full=False) :
        debug_print(top)
        root = GlypherPhraseGroup.get_xml(self, name=name, top=top, targets=targets,
                                   full=full)
        root.set('file', self.filename)
        return root

g.add_phrasegroup_by_class('svg_phrasegroup', GlypherSVGPhraseGroup)
