import glypher as g
import Dynamic
import sympy
import re
import draw
from Phrase import *
from PhraseGroup import *
from Symbol import *
from sympy import physics, sympify
from sympy.core.symbol import Symbol as sySymbol
from sympy.core.basic import Basic
from sympy.core.numbers import *
from sympy.core.sets import *
from ..unicode_to_latex import unicode_to_latex, latex_to_unicode
import Parser
import Script

class GlypherWord (GlypherEntity) :
    title = "Word"
    info_text = '''
Basic conceptual unit - combination of unicode characters (GlypherSymbols)
together representing a sympy Symbol
        '''
    defined_function = False
    defined_symbol = False
    let_function = False
    let_matrix = False
    wildcard = False

    has_indexed_positions = True
    def get_indexed_length(self) :
        return len(self.shape)

    mid = None
    shape = u''
    ink = False
    def get_shape(self) :
        return self.shape
    def set_shape(self, shape) :
        if shape is None :
            shape = ''
        debug_print(shape)
        self.shape = unicode(shape)
        self.set_redraw_required()
    def get_italic(self) : return self.get_p('italic')
    def set_italic(self, italic, quiet=False) :
        self.set_p('italic', italic)
        self.set_redraw_required()
        if not quiet : self.recalc_bbox()
    def get_ink(self) : return self.ink
    def set_ink(self, ink) :
        self.ink = ink
        self.set_redraw_required()
    def get_mid(self) : return self.mid
    def set_mid(self, mid) :
        self.mid = mid
        self.set_redraw_required()

    rendering = None

    def find_nearest(self, point, fall_through=True,
                     enterable_parent=False,
                     attachable=False,
                     row=0,
                     avoid=None) :
        dist, ent, ni = GlypherEntity.find_nearest(self, point,
                    fall_through, enterable_parent,
                    attachable, row, avoid=None)

        abspos = -self.get_local_offset()[0]
        abspos -= self.config[0].get_bbox()[0]
        closest = (0, self.get_width()+abspos)
        if dist > -1 and ent is self :
            for i in range(self.get_indexed_length()) :
                cur = self.get_caret_position(True, None, i)
                debug_print(closest[1], point[0], cur[0]+abspos)
                if abs(closest[1] - point[0]) > abs(cur[0]+abspos - point[0]) :
                    closest = (i, cur[0]+abspos)

        debug_print(closest, point)
        return dist, ent, closest[0]
    
    def set_active(self, active) :
        '''
        Set this entity as active (and tell any parent)
        '''
        self.active = active

        if self.included() :
            self.get_parent().set_child_active(active, self)

    def check_combination(self, shape, go_up=True) :
        cmb = self.to_string() + unicode(shape)
        chg = cmb in g.combinations
        if chg :
            self.set_shape(g.combinations[cmb])
            self.check_shape()
            self.recalc_bbox()
            if self.included() :
                self.get_parent().child_change()
        return chg

    def get_xml(self, name = None, top = True, targets = None, full = False) :
        root = GlypherEntity.get_xml(self, name, top)
        if not self.get_auto_italicize() :
            root.set('italic', self.get_italic())
        root.set('shape', self.get_shape())
        root.set('ink', str(self.get_ink()))
        return root

    def get_python(self, true_mode="python") :
        return self.to_string()

    def to_latex(self) :
        if not self.get_visible() : 
            return ""
        elif self.get_blank() :
            return " "

        try :
            latex = unicode_to_latex(unicode(self.get_shape()))
        except :
            latex = str(self.get_shape())

        return latex

    def to_string(self, mode="string", true_mode=None) :
        if not self.get_visible() : return unicode("")
        elif self.get_blank() : return unicode(" ")

        uc = u'\u25a1'
        try :
            uc = unicode(self.get_shape())
        except UnicodeDecodeError :
            raise(RuntimeError('Could not change ' + self.get_shape() + ' to unicode'))
        return unicode(self.get_shape())

    def recalc_basebox(self, config = None, only_recalc_self = False) :
        b = self.config[0].get_bbox()
        mid = self.get_mid()
        if mid is None : mid = (b[3]-b[1])*0.5
        self.config[0].basebox = (b[0], (b[0]+b[2])*0.5, b[2], b[1], b[3]-mid, b[3])

    #def get_non_ink_extent(self) :
    #    if not self.get_ink() :
    #        return None

    #    (rh,rw,mid,delta,cp) = self.rendering.draw(self.get_font_size()*10,\
    #        ink=True, italic=self.get_italic(), bold=self.get_ip('bold'), font_name=self.get_ip('font_name'))
    #    c = self.get_font_size_combined_coeff()/10.
    #    self.cp = (cp[0]*c, cp[1]*c)

    #    return ( self.config[0].basebox[0]-c*delta[0],
    #             self.config[0].basebox[2]-c*delta[1],
    #             self.config[0].basebox[3]-c*delta[2],
    #             self.config[0].basebox[5]-c*delta[3] )

    def get_ink_extent(self) :
        (rh,rw,mid,delta,cp) = self.rendering.draw(self.get_font_size()*10,\
            ink=True, italic=self.get_italic(), bold=self.get_ip('bold'), font_name=self.get_ip('font_name'))
        c = self.get_font_size_combined_coeff()/10.
        self.cp = (cp[0]*c, cp[1]*c)

        return ( self.config[0].basebox[0]+c*delta[0],
                 self.config[0].basebox[3]+c*delta[2],
                 self.config[0].basebox[2]+c*delta[1],
                 self.config[0].basebox[5]+c*delta[3] )

    def get_caret_position(self, inside=False, pos=None, index=None) :
        if index is None :
            return GlypherEntity.get_caret_position(self, inside=inside, pos=pos)

        bcp = GlypherEntity.get_caret_position(self, inside=inside, pos=[self.config[0].bbox[0], self.config[0].bbox[3]])
        (rh,rw,mid,delta,cp) = self.rendering.draw(self.get_font_size()*10,\
            ink=True, italic=self.get_italic(), bold=self.get_ip('bold'),
                                                font_name=self.get_ip('font_name'),
                                                index=index)
        c = self.get_font_size_combined_coeff()/10.
        self.cp = (cp[0]*c, cp[1]*c)
        return (bcp[0]+self.cp[0], bcp[1]+self.cp[1])

    def set_bold(self, bold) :
        self.set_ip('bold', bold)
    def get_bold(self) :
        return self.get_ip('bold')
    def cast(self, size, pos, code = None, align = ('l','b'), bild = False) :
        shape = self.get_shape()
        ink = self.get_ink()
        italic = self.get_italic()
        bold = self.get_bold()
        font_name = self.get_ip('font_name')
        self.rendering = Renderers.find_rendering(shape, code, italic=italic,
                                                  bold=bold)

        (rh,rw,mid,delta, cp) = self.rendering.draw(self.get_font_size()*10,\
            ink=ink, italic=italic, bold=bold, font_name=font_name)
        if not ink :
            (irh,irw,imid,idelta,icp) = self.rendering.draw(self.get_font_size()*10,\
                ink=True, italic=italic, bold=bold, font_name=font_name)
            rw = max(rw, irw)
        c = self.get_font_size_combined_coeff()/10.
        self.cp = (cp[0]*c, cp[1]*c)
        #(rh,rw,mid) = self.rendering.draw(size, ink=ink)
        rh *= c
        rw *= c
        mid *= c
        self.set_mid(mid)
        p = self.padding
        #rh += p[1]+p[3]
        #rw += p[0]+p[2]
        self.set_ref_width(rw)
        self.set_ref_height(rh)
        #if align[0] == 'l' :
        #    self.ref_bbox[0] = pos[0]
        #    self.ref_bbox[2] = pos[0] + rw
        #elif align[0] == 'c' :
        #    self.ref_bbox[0] = pos[0] - rw/2
        #    self.ref_bbox[2] = pos[0] + rw/2
        #else :
        #    self.ref_bbox[0] = pos[0] - rw
        #    self.ref_bbox[2] = pos[0]

        #if align[1] == 'b' :
        #    self.ref_bbox[3] = pos[1]
        #    self.ref_bbox[1] = pos[1] - rh
        #elif align[1] == 'm' :
        #    self.ref_bbox[1] = pos[1] - rh/2
        #    self.ref_bbox[3] = pos[1] + rh/2
        #else :
        #    self.ref_bbox[1] = pos[1]
        #    self.ref_bbox[3] = pos[1] + rh
        #self.ref_bbox = list(self.bbox)
    
    def recalc_bbox(self, quiet=False, old_bbox=None) :
        if self.get_shape() is None : return
        self.cast(self.get_scaled_font_size(), self.get_anchor(), \
            None, self.get_align(), None)
        #traceback.print_stack()
        GlypherEntity.recalc_bbox(self, quiet=quiet)

        # Sort out zero-width symbols (we really should be able to handle these)
    
    def draw(self, cr) :
        if self.draw_offset != (0,0) :
            cr.save()
            cr.translate(*self.draw_offset)
        self.draw_topbaseline(cr, (0,1,1))
        if not self.get_visible() or self.get_blank() : return
        #if self.get_scaled_font_size() != self.size : #TODO: find a more efficient way of doing this
        #    self.size = self.get_scaled_font_size()
            #self.cast(self.shape, self.get_scaled_font_size(), self.loc, self.text, None, self.align, self.italic)
        #debug_print(self.format_me())

        try :
            fvar = g.find_variable(self.get_sympy(sub=False))
            if self.get_visible() and not self.get_blank() and \
                    fvar[0] == g.VAR_SET_SYM :
                cr.save()
                cr.set_source_rgba(1.0,0.8,0,0.5)
                #cr.rectangle(self.config[0].bbox[0], self.config[0].bbox[1],
                #             self.get_width(), self.get_height())
                area = (self.config[0].bbox[0],
                        self.config[0].bbox[2],
                        self.config[0].bbox[1],
                        self.config[0].bbox[3])
                draw.trace_rounded(cr, area, 7)
                cr.fill()
                cr.restore()
        except :
            pass

        if g.show_rectangles :
            colour = (0.8,0.8,0.8)
        elif self.in_selected() :
            colour = (0.8, 0.2, 0.2)
        else :
            colour = self.get_rgb_colour()

        if not self.rendering :
            self.cast(self.get_scaled_font_size(), self.get_anchor(), \
                None, self.get_align(), None)
        cr.save()
        cr.translate(self.config[0].bbox[0]+self.padding[0], self.config[0].bbox[1]+self.padding[1])
        cr.scale(self.get_font_size_combined_coeff()/10.,
                 self.get_font_size_combined_coeff()/10.)
        self.rendering.draw(self.get_font_size()*10, cr, self.config[0].bbox[0]+self.padding[0], self.config[0].bbox[1]+self.padding[1],\
            colour=colour,\
            ink=self.get_ink(), italic=self.get_italic(), bold=self.get_ip('bold'), font_name=self.get_ip('font_name'))
        cr.restore()
        if g.show_rectangles :
            cr.save()
            cr.set_source_rgba(0.5,0.5,1.0,0.2)
            cr.rectangle(self.config[0].bbox[0], self.config[0].bbox[1], self.config[0].bbox[2]-self.config[0].bbox[0], self.config[0].bbox[3]-self.config[0].bbox[1])
            cr.stroke()
            cr.set_source_rgba(0.5, 0.0, 1.0, 0.4)
            cr.move_to(self.config[0].basebox[0]-5, self.config[0].basebox[4])
            cr.line_to(self.config[0].basebox[2]+5, self.config[0].basebox[4])
            for i in (0,2) :
                cr.move_to(self.config[0].basebox[i], self.config[0].basebox[4]-2)
                cr.line_to(self.config[0].basebox[i], self.config[0].basebox[4]+2)
            cr.move_to(self.config[0].basebox[0]-5, self.config[0].basebox[3])
            cr.line_to(self.config[0].basebox[2]+5, self.config[0].basebox[3])
            cr.move_to(self.config[0].basebox[0]-5, self.config[0].basebox[5])
            cr.line_to(self.config[0].basebox[2]+5, self.config[0].basebox[5])
            cr.stroke()
            cr.restore()
        if self.draw_offset != (0,0) :
            #cr.translate(-self.draw_offset[0], -self.draw_offset[1])
            cr.restore()

        if self.show_decoration() :
            self.decorate(cr)

    def change_alternative(self, dir) :
        alts = g.find_alternatives(self.get_shape())
        if alts == None :
            if self.included() :
                return self.get_parent().change_alternative(dir)
            else :
                return False
        ind = alts[0].index(self.get_shape())
        self.set_shape(alts[0][(len(alts[0]) + ind + dir)%len(alts[0])])
        self.check_shape()
        self.set_recommending(self)
        self.recalc_bbox()
        return True

    def draw_alternatives(self, cr) :
        if not self.get_visible() : return

        a = g.find_alternatives(self.get_shape(), generate_altbox=True)
        if a is None : return
        alts, altbox = (a[0], a[1])
        #cr.save()
        #cr.set_line_width(2.0)
        #cr.set_source_rgba(1.0,1.0,1.0)
        #cr.rectangle(self.config[0].bbox[0]-4, self.config[0].bbox[1]-4, self.config[0].bbox[2]-self.config[0].bbox[0]+8, self.config[0].bbox[3]-self.config[0].bbox[1]+8)
        #cr.fill_preserve()
        #cr.set_source_rgb(0.7,0.7,0.7)
        #cr.stroke()
        cr.save()
        cr.translate(self.config[0].bbox[2]+self.get_local_offset()[0]+20, self.config[0].bbox[1]+self.get_local_offset()[1])
        if alts != None : altbox.draw(cr, anchor=(0,0),\
            size=self.get_scaled_font_size(), rgb_colour=self.get_rgb_colour(), active=self.get_shape())
        cr.restore()
        #self.draw(cr)
        #cr.restore()

    @classmethod
    def parse_element(cls, parent, root, names, targets, operands, recommending, lead,
                  add_entities, am = None, top = True, args=None) :
        shape = root.get('shape')
        italic = root.get('italic')
        
        if italic is not None :
            italic = italic=='True'

        ink = root.get('ink')

        if ink is not None :
            ink = ink=='True'

        new_phrasegroup = cls(parent, shape=shape, italic=italic,
                                               ink=ink)

        return new_phrasegroup
    def is_function(self) :
        return self.defined_function or self.let_function 

    def set_auto_italicize(self, auto_italicize) : self.set_p('auto_italicize', auto_italicize)
    def get_auto_italicize(self) : return self.get_p('auto_italicize')

    def delete_index(self, index) :
        shape = self.get_shape()
        self.set_shape(shape[:index-1]+shape[index:])
        self.check_shape()
        self.recalc_bbox()

    def __init__(self, parent, area = (0,0,0,0), shape=u'', italic=False,
                 ink=False) :
        GlypherEntity.__init__(self, parent)
        self.add_properties({
            'align' : ('c', 'm'),
            'italic' : False,
            'mid' : 0.5, 'have_alternatives' : True})

        self.topline = area[1]
        self.baseline = area[3]
        self.set_ink(ink)
        self.set_italic(italic)
        self.set_shape(shape)
        self.check_shape()
        
        self.mes.append('symbol')
        self.set_ref_width(area[2] - area[0])
        self.set_ref_height(area[3] - area[1])
        self.recalc_basebox()
        self.config[0].reset()
        pos = (self.config[0].bbox[0], self.config[0].bbox[3])
        self.old_bbox = [pos[0], pos[1], pos[0], pos[1]]

        self.recalc_bbox()

        self.mes.append('word')
        self.add_properties({'auto_italicize' : True, 'wordlike' : True})

    def split(self, position) :
        area = (self.config[0].bbox[0], self.get_topline(), self.config[0].bbox[2], self.get_baseline())
        wanc = area[0]
        word1 = GlypherWord(self.get_up(), area)
        word2 = GlypherWord(self.get_up(), area)
        ents = self.shape
        bboxes = [self.get_caret_position(True, None, i)[0] for i in\
                    range(self.get_indexed_length())]
        i = 0
        for ent in ents :
            if bboxes[i]-wanc < position : word1.append(ent)
            else : word2.append(ent)
            i += 1
        word1.recalc_bbox()
        word2.recalc_bbox()
        if len(word1.shape) == 0 : return (None,word2)
        elif len(word2.shape) == 0 : return (word1,None)
        return (word1, word2)

    # We assume that, if a word is not leading with a number, it shouldn't be considered as such
    def is_leading_with_num(self) : return self.is_num()

    def is_num(self) : return all_numbers.match(self.get_shape())

    def is_roman(self) : return all_roman.match(self.get_shape())

    def to_latex(self) :
        if not self.get_visible() : return ""
        elif self.get_blank() : return " "

        me_str = self.to_string()
        if me_str in g.interpretations :
            me_str = g.interpretations[me_str]["latex"]
        else :
            me_str = GlypherEntity.to_latex(self)

        return str(me_str)

    #def _get_symbol_string(self, sub = True) :
    #    if self.is_num() :
    #        return None

    #    return GlypherPhrase._get_symbol_string(sub=sub)

    def get_sympy(self, sub = True, ignore_func = False) :
        ignore_func = ignore_func or not sub

        if not ignore_func :
            fvar = g.find_variable(self.get_sympy(ignore_func=True))

            if fvar is not None :
                if fvar[0] == g.VAR_DEF_FUN :
                    name = self.get_sympy(ignore_func=True)
                    args = fvar[1]
                    f = Dynamic.Function(str(self.get_shape()))
                    return f(*args)
                if fvar[0] == g.VAR_DEF_UNI :
                    return self.get_sympy(ignore_func=True)*fvar[1]
                if fvar[0] == g.VAR_SET_SYM :
                    return fvar[1]
                else :
                    return fvar[1]

        #if self.let_function and not ignore_func and \
        #        self.get_sympy(ignore_func=True) in g.let_functions :
        #    name = self.get_sympy(ignore_func=True)
        #    return g.let_functions[name]
        #elif self.defined_function and not ignore_func and \
        #        self.get_sympy(ignore_func=True) in g.define_functions :
        #elif self.let_matrix and not ignore_func and \
        #        self.get_sympy(ignore_func=True) in g.let_matrices :
        #    name = self.get_sympy(ignore_func=True)
        #    return g.let_matrices[name]

        me_str = self._get_symbol_string(sub)
        if self.is_num() :
            sym = sympify(self.to_string())
        else :
            me_str = str(unicode_to_latex(me_str))
            sym = sySymbol(me_str)
            #if sub and sym in g.var_table :
            #    sym = g.var_table[sym]

        #if sym in g.define_symbols and not ignore_func :
        #    sym = g.define_symbols[sym]

        return sym
            #return sySymbol(self.to_string("sympy"))

    def get_symbol_extents(self) :
        return (self.config[0].bbox[0], self.config[0].bbox[2])
        ####
        if len(self.shape) == 0 :
            return (self.config[0].bbox[0], self.config[0].bbox[2])

        extents = [self.get_entities()[0].config[0].bbox[0],
                   self.get_entities()[0].config[0].bbox[2]]

        for sym in self.get_entities() :
            ie = sym.config[0].bbox
            if ie[0] < extents[0] :
                extents[0] = ie[0]
            if ie[2] > extents[1] :
                extents[1] = ie[2]

        return extents

    #def draw(self, cr) :
    #    try :
    #        fvar = g.find_variable(self.get_sympy(sub=False))
    #        if self.get_visible() and not self.get_blank() and \
    #                fvar[0] == g.VAR_SET_SYM :
    #            cr.save()
    #            cr.set_source_rgba(1.0,0.8,0,0.5)
    #            #cr.rectangle(self.config[0].bbox[0], self.config[0].bbox[1],
    #            #             self.get_width(), self.get_height())
    #            area = (self.config[0].bbox[0],
    #                    self.config[0].bbox[2],
    #                    self.config[0].bbox[1],
    #                    self.config[0].bbox[3])
    #            draw.trace_rounded(cr, area, 7)
    #            cr.fill()
    #            cr.restore()
    #    except :
    #        pass

    #    GlypherPhrase.draw(self, cr)

    def process_key(self, name, event, caret) :
        mask = event.get_state()
        fvar = g.find_variable(self.get_sympy(sub=False, ignore_func=True))
        if name == 'Return' and \
                self.included() and \
                fvar[0] == g.VAR_SET_SYM :
            #sy = self.get_sympy(sub=False, ignore_func=True)
            new_pg = interpret_sympy(self.get_parent(), fvar[1])
            self.get_parent().exchange(self, new_pg)
            return True
        
        return GlypherEntity.process_key(self, name, event, caret)

    def _adjust_bbox(self, bbox) :
        '''
        Override this to expand (or contract) bbox after it has been set by
        contained elements.
        '''

        ie = self.get_ink_extent()
        if ie is None :
            return
        if ie[2] > bbox[2] :
            bbox[2] = ie[2]

    #FIXME: this is a bit of a hack... or is it? (cf BinaryExpression should_mul)
    def get_entities(self) :
        return self.shape

    def append(self, shape, after=None) :
        '''Runs a few checks to see what kind of Word we now have.'''

        debug_print(self.to_string(), shape)
        if isinstance(shape, GlypherSymbol) :
            shape = shape.get_shape()

        if after is None :
            if self.shape is not None :
                new_shape = self.shape+unicode(shape)
            else :
                new_shape = unicode(shape)
        else :
            new_shape = self.shape[:after]+unicode(shape)+self.shape[after:]
        self.set_shape(new_shape)
        self.check_shape()
        new_shape = self.get_shape()

        # Check whether we have a special combination
        if new_shape in g.specials and self.included() :
            pg = Parser.make_phrasegroup(self.get_parent(),
                                         g.specials[shape])
            self.get_parent().exchange(self, pg)
            self.set_recommending(pg)
            return

    def check_shape(self) :
        shape = self.get_shape()

        # Check whether we have a collapsible combination
        if shape in g.combinations :
            self.set_shape(g.combinations[shape])

        try :
            symp = self.get_sympy(ignore_func=True)
        except :
            symp = None

        should_be_set = False

        try :
            fvar = g.find_variable(self.get_sympy(ignore_func=True))
        except :
            fvar = None

        if isinstance(symp, Basic) and fvar is not None :
            should_be_set = True
            # If we have a Let function, show it up
            if fvar[0] == g.VAR_LET_FUN and not self.let_function :
                self.let_function = True
                self.set_bold(True)
                self.set_rgb_colour((0.5, 0.2, 0.5))
            # If we have a Define function, show it up
            if fvar[0] == g.VAR_DEF_UNI and not self.defined_symbol :
                self.defined_symbol = True
                self.set_rgb_colour((0.6, 0.3, 0.0))
            # If we have a Define function, show it up
            if fvar[0] == g.VAR_DEF_FUN and not self.defined_function :
                self.defined_function = True
                self.set_bold(True)
                self.set_rgb_colour((0.2, 0.4, 0.4))
            # If we have a Let matrix, show it up
            if fvar[0] == g.VAR_LET_MAT and not self.let_matrix :
                self.let_matrix = True
                self.set_bold(True)
            if fvar[0] == g.VAR_DEF_SYM :
                # If we have a Wildcard, show it up
                if isinstance(fvar[1], Dynamic.Wild) :
                    if not self.wildcard :
                        self.wildcard = True
                        self.set_rgb_colour((0.8, 0.6, 0.0))
                # If we have a pre-defined symbol, show it up
                elif not self.defined_symbol :
                    self.defined_symbol = True
                    self.set_rgb_colour((0.4, 0.4, 0.4))

        # Otherwise, cancel those settings
        if not should_be_set and \
                (self.defined_function or self.let_function or self.wildcard or \
                 self.defined_symbol) :
            self.defined_function = False
            self.let_function = False
            self.set_bold(False)
            self.set_rgb_colour(None)

        if self.get_auto_italicize() :
         if len(self.shape) > 1 and self.is_roman() :
            self.set_italic(False)
         else :
            if not self.get_italic() and all_roman.match(shape) :
                self.set_italic(True)

        self.index = len(self.shape)-1

        if len(self.shape) > 1 :
            self.index = 1

        self.recalc_bbox()

    def decorate(self, cr) :
        #self.draw_topbaseline(cr)
        if not self.get_visible() :
            return
        #if self.active and self.shows_active :
        if g.additional_highlighting and self.get_attached() and self.get_shows_active() :
            cr.save()
            #cr.set_line_width(4.0)
            #cr.rectangle(self.bbox[0]-2, self.bbox[1]-2, self.bbox[2]-self.bbox[0]+4, self.bbox[3]-self.bbox[1]+4)
            cr.move_to(self.config[0].bbox[0]-2, self.config[0].bbox[3]+2)
            draw.draw_blush( cr, self.config[0].bbox[2]-self.config[0].bbox[0]+4, (0.5,0.5,0.5), 8)
            #cr.line_to(self.config[0].bbox[2]+2, self.config[0].bbox[3]+2)
            #cr.set_source_rgba(0.5, 0.5, 0.5, 0.6)
            #cr.stroke()
            cr.restore()
        elif self.get_attached() or (len(self.shape) > 1 and not self.is_num()) :
            cr.save()
            #cr.set_line_width(4.0)
            cr.move_to(self.config[0].bbox[0]+2, self.config[0].bbox[3]-2)
            #draw_blush( cr, self.config[0].bbox[2]-self.config[0].bbox[0]+4, (0.7,0.9,0.7), 2)
            cr.line_to(self.config[0].bbox[2]-2, self.config[0].bbox[3]-2)
            col = (0.7, 0.7, 0.9, 1.0) if self.get_attached() else (0.7, 0.9, 0.7, 1.0)
            cr.set_source_rgba(*col)
            cr.stroke()
            cr.restore()

alternatives = {
    1 : ('planck', 'hbar'),

    2 : ('second', 'millisecond', 'microsecond', 'nanosecond', 'picosecond'),
    3 : ('gee','gram', 'kilogram', 'microgram', 'milligram'),
    4 : ('liter', 'milliliter', 'centiliter', 'decaliter'),
    5 : ('meter', 'kilometer', 'u0', 'mole', 'micron', 'millimeter', 'centimeter'),
    6 : ('ampere',),
    7 : ('kelvin',),
    8 : ('speed_of_light', 'candela'),
    9 : ('hertz',),
    10: ('newton',),
    11: ('exp1', 'e0'),
    12: ('G',),
    }
alternatives_keys = {
    'h' : 1, 's' : 2, 'g' : 3, 'l' : 4, 'm' : 5,
    'A' : 6, 'K' : 7, 'C' : 8, 'H' : 9, 'N' : 10,
    'e' : 11,'G' : 12,
    }
alternatives_current_defaults = {}

class GlypherConstant(GlypherPhraseGroup) :
    get_sympy_code = None
    value = None
    toolbox = None

    @classmethod
    def parse_element(cls, parent, root, names, targets, operands, recommending, lead,
                  add_entities, am = None, top = True, args=None) :
        ent = cls(parent, GlypherSymbol(parent, root.find('symbol').text))
        if root.find('sub') is not None :
            sub = Parser.parse_element(parent, root, names, targets, operands,
                                       recommending, lead, add_entities, am=am,
                                       top=False, args=args)
            ent = Script.GlypherScript.subscript(parent, area=(0,0,0,0),
                expression=ent, subscript=sub)

        return ent

    def get_sympy(self) :
        if self.get_sympy_code is not None :
            return Dynamic.eval_for_sympy(self, self.get_sympy_code)
        return self.value

    def __init__(self, parent, entity) :
        GlypherPhraseGroup.__init__(self, parent, [], area=[0,0,0,0])
        self.set_rgb_colour((0.5, 0.5, 0.8))
        self.mes.append('constant')
        self.add_properties({'enterable' : False,
                             'have_alternatives' : False})
        self.suspend_recommending()
        self.append(entity)
        #self.set_default_entity_xml()

constants = {}
class GlypherConstant_(GlypherConstant) :
    value = None
    symbol = None
    me = None
    constant_have_alternatives = False

    parse_element = None

    def __init__(self, parent) :
        GlypherPhraseGroup.__init__(self, parent)
        self.set_rgb_colour((0.5, 0.5, 0.8))
        self.mes.append('constant')
        if self.me is not None :
            self.mes.append(self.me)
        self.add_properties({'enterable' : False,
                             'have_alternatives' :
                             self.constant_have_alternatives})
        self.suspend_recommending()
        if self.symbol is not None :
            self.append(GlypherEntity.xml_copy(parent, self.symbol))
        self.set_default_entity_xml()

    def change_alternative(self, dir = 1) :
        ret = GlypherPhrase.change_alternative(self, dir=dir)
        if ret:
            for ind in alternatives :
                alts = alternatives[ind]
                if self.mes[-1] in alts :
                    i = alternatives_current_defaults[ind]
                    alternatives_current_defaults[ind] = (i+dir) % len(alts)
                    break

        return ret

    @classmethod
    def new_from_symbol(cls, name, symbol, value, italicize=False,
                        cat="Constants", alternatives_cat=None) :
        new_sym = GlypherSymbol(None, symbol, italic=italicize)
        return cls.new_from_entity(name, new_sym, value, cat=cat,
                                   alternatives_cat=alternatives_cat)

    @classmethod
    def new_from_symbol_sub(cls, name, symbol, sub,
            value, italicize=False, it_sub=None, cat="Constants",
                            alternatives_cat=None) :
        if it_sub is None :
            it_sub = italicize
        new_sym = GlypherSymbol(None, symbol, italic=italicize)
        new_sub = GlypherSymbol(None, sub, italic=it_sub)
        new_dec = Script.GlypherScript.subscript(None, area=(0,0,0,0),
            expression=new_sym, subscript=new_sub)
        return cls.new_from_entity(name, new_dec, value, cat=cat,
                                   alternatives_cat=alternatives_cat)

    @classmethod
    def new_from_entity(cls, name, symbol, value, cat="Constants",
                        alternatives_cat=None) :
        if cat is not None :
            toolbox = {'symbol' : symbol.to_string(),
                           'category' : cat,
                           'priority' : None,
                           'shortcut' : None,
                       'alternatives' : None
                      }
        else :
            toolbox = None
        new_dict = {'me':name, 'value':value, 'symbol':symbol,
                    'toolbox':toolbox,'altname':name,'alternatives_cat':alternatives_cat}
        new_sym = type('GlypherConstant_'+str(name), (cls,object,), new_dict)

        for alts in alternatives :
            alts = alternatives[alts]
            if name in alts :
                new_sym.constant_have_alternatives = True
                new_sym.alternatives = alts
                new_sym.altname = name
        g.add_phrasegroup_by_class(name, new_sym, alt_cat=alternatives_cat)
        return new_sym

constants['planck'] = GlypherConstant_.new_from_symbol('planck', 'h',
                                                      physics.units.planck,
                                                      italicize=True,
                                                      cat="Physical Constants",
                                                      alternatives_cat='planck')
constants['hbar'] = GlypherConstant_.new_from_symbol( 'hbar', u'\u0127',
                                                    physics.units.hbar,
                                                    italicize=True,
                                                    cat="Physical Constants",
                                                    alternatives_cat='planck')
constants['speed_of_light'] = GlypherConstant_.new_from_symbol( 'speed_of_light',
                                                              'c',
                                                              physics.units.speed_of_light,
                                                              italicize=True,
                                                              cat="Physical Constants")
constants['G'] = GlypherConstant_.new_from_symbol( 'G', 'G',
                                                 physics.units.G,
                                                 italicize=True,
                                                 cat="Physical Constants")
constants['gee'] = GlypherConstant_.new_from_symbol( 'gee', 'g',
                                                   physics.units.gee,
                                                   italicize=True,
                                                 cat="Physical Constants",
                                                   alternatives_cat="gram")
constants['e0'] = GlypherConstant_.new_from_symbol_sub( 'e0', 'e', '0',
                                                      physics.units.e0,
                                                      italicize=True,
                                                 cat="Physical Constants")
constants['u0'] = GlypherConstant_.new_from_symbol_sub( 'u0', u'\u03BC', '0',
                                                      physics.units.u0,
                                                      italicize=True,
                                                 cat="Physical Constants")
constants['Z0'] = GlypherConstant_.new_from_symbol_sub( 'Z0', 'Z', '0',
                                                      physics.units.Z0,
                                                      italicize=True,
                                                 cat="Physical Constants")

constants['exp1'] = GlypherConstant_.new_from_symbol( 'exp1',
                                                    'e',Exp1(),
                                                    italicize=False,
                                                   cat="Mathematical Constants")
constants['exponential_e'] = GlypherConstant_.new_from_symbol( 'exp1', 'e',
                                                             Exp1(),
                                                             italicize=False,
                                                   cat="Mathematical Constants")
constants['empty_set'] = GlypherConstant_.new_from_symbol(
                                                                  'empty_set',
                                                                  u'\u2205',
                                                                  EmptySet(),
                                                                  italicize=False,
cat="Sets")

#constants['complex'] = GlypherConstant_.new_from_symbol(
#                                                                  'Complex',
#                                                                  u'\u2102',
#                                                                  Exp1(),
#                                                                  italicize=False)
#        (u'\u2102', '\\Complex'), \
#        (u'\u2124', '\\Integer'), \
#        (u'\u2115', '\\Natural'), \
#        (u'\u211A', '\\Rational'), \
#        (u'\u211D', '\\Real'), \

constants['infinity'] = GlypherConstant_.new_from_symbol(
                                       'infinity',
                                       u'\u221e',
                                       Infinity(),
    italicize=False,
                                                   cat="Mathematical Constants")
constants['imaginary_unit'] = GlypherConstant_.new_from_symbol(
                                       'imaginary_unit',
                                       'i',
                                       ImaginaryUnit(),
    italicize=False,
    cat="Complex")

constants['realR'] = GlypherConstant_.new_from_symbol(
                                       'realR',
                                       u'\u211D',
                                       None, italicize=False,
cat="Sets")

constants['rationalQ'] = GlypherConstant_.new_from_symbol(
                                       'rationalQ',
                                       u'\u211A',
                                       None, italicize=False,
cat="Sets")

constants['complexC'] = GlypherConstant_.new_from_symbol(
                                       'complexC',
                                       u'\u2102',
                                       None, italicize=False,
cat="Sets")
constants['integerZ'] = GlypherConstant_.new_from_symbol(
                                       'integerZ',
                                       u'\u2124',
                                       None, italicize=False,
cat="Sets")

units = {}
class GlypherUnit_(GlypherConstant_) :
    value = None

    def __init__(self, parent) :
        GlypherConstant_.__init__(self, parent)
        self.set_rgb_colour((0.5, 0.8, 0.6))
        self.suspend_recommending()

def auto_make_unit(name, symbol, cat=None, alternatives_cat=None) :
  units[name] = GlypherUnit_.new_from_symbol(name, symbol,
                                            physics.units.__dict__[name],
                                            cat=cat,
                                            alternatives_cat=alternatives_cat)

auto_make_unit('meter', 'm', cat="Units", alternatives_cat="meter")
auto_make_unit('gram', 'g', cat="Units", alternatives_cat="gram")
auto_make_unit('second', 's', cat="Units", alternatives_cat="second")
auto_make_unit('ampere', 'A', cat="Units")
auto_make_unit('kelvin', 'K', cat="Units")
auto_make_unit('mole', u'\u33d6', cat="Units")
auto_make_unit('candela', u'\u33c5', cat="Units")
auto_make_unit('liter', u'\u2113', cat="Units", alternatives_cat="liter")

auto_make_unit('hertz', u'\u3390', cat="Units")
auto_make_unit('newton', 'N', cat="Units")

auto_make_unit('millisecond', u'\u33b3', alternatives_cat='second')
auto_make_unit('microsecond', u'\u33b2', alternatives_cat='second')
auto_make_unit('nanosecond', u'\u33b1', alternatives_cat='second')
auto_make_unit('picosecond', u'\u33b0', alternatives_cat='second')

auto_make_unit('kilogram', u'\u338f', alternatives_cat='gram')
auto_make_unit('microgram', u'\u338d', alternatives_cat='gram')
auto_make_unit('milligram', u'\u338e', alternatives_cat='gram')

auto_make_unit('milliliter', u'\u3396', alternatives_cat='liter')
#auto_make_unit('decaliter', 'dl')

auto_make_unit('kilometer', u'\u339e', alternatives_cat='meter')
auto_make_unit('micron', u'\u339b', alternatives_cat='meter')
auto_make_unit('millimeter', u'\u339c', alternatives_cat='meter')
auto_make_unit('centimeter', u'\u339d', alternatives_cat='meter')

def make_word(string, parent, area = (0,0,0,0), auto_italicize=True) :
    '''Make a Word from a string.'''

    if string[0] == '\\' :
        string = latex_to_unicode(string)

    string = unicode(string)
    word = GlypherWord(parent, area, shape=string)
    word.set_auto_italicize(auto_italicize)
    word.check_shape()
    word.recalc_bbox()
    return word

all_numbers = None
all_roman = None
def word_init() :
    global all_numbers, all_roman
    all_numbers = re.compile(g.is_floating_point_num_regex)
    all_roman   = re.compile('[A-Za-z]')

import Interpret
def interpret_sympy(p, sy) :
    return Interpret.interpret_sympy(p, sy)

g.add_phrasegroup_by_class('constant', GlypherConstant)
g.add_phrasegroup_by_class('word', GlypherWord)
