from ..glypher.BinaryExpression import \
    GlypherBinaryExpression,\
    GlypherBinaryRowExpression, GlypherAdd
from aobject.utils import debug_print
from ..glypher import Parser
from ..glypher import glypher as g
from ..glypher.Symbol import GlypherSymbol
from ..glypher.Spacer import GlypherSpace
from ..glypher.Phrase import GlypherPhrase
from ..glypher.PhraseGroup import GlypherPhraseGroup
from ..glypher.Word import make_word


class GlassmakerIndent(GlypherPhrase) :
    def __init__(self, parent) :
        GlypherPhrase.__init__(self, parent)
        self.append(GlypherSpace(parent, (1., 0.4)))
        phrase = GlypherPhrase(parent)
        self.append(phrase)
        self.set_IN(phrase)
        self.mes.append('indent')

class GlassmakerCommand(GlypherBinaryExpression) :
    toolbox = { 'symbol' : 'C',
                'category' : 'Utilities',
                'shortcut' : None,
                'alternatives' : None,
                'priority' : None }
    stop_for_binary_expression_exceptions = ('blockcommand', 'block')
    def __init__(self, parent, lhs=None, rhs=None, num_ops=2) :
        space_symbol = GlypherSpace(parent, (0.1,0.4))
        #space_symbol = make_word(' ',self)
        GlypherBinaryExpression.__init__(self, parent, symbol=space_symbol,
                                         no_brackets=True,
                                         num_ops=num_ops, lhs=lhs, rhs=rhs,
                                         variable_ops=True)
        self.poss[0].set_rgb_colour((0.8,0.4,0.0))
        self.add_properties({'allow_unary' : True, 'symbol_shape' : ' ',
                             'no_brackets' : True, 'num_ops' : 2, 'gravity' : True})
        self.mes.append('space_array')
        self.mes.append('command')

    def get_python(self, true_mode="python") :
        return " ".join([arg.to_string(true_mode) for arg in self.get_args()])

class GlassmakerBlock(GlypherBinaryRowExpression) :
    stop_for_binary_expression_exceptions = ()
    toolbox = { 'symbol' : 'C{}',
                'category' : 'Utilities',
                'shortcut' : None,
                'alternatives' : None,
                'priority' : None }
    poser = GlypherPhrase

    @classmethod
    def parse_element(cls, parent, root, names, targets, operands, recommending, lead,
                  add_entities, am=None, top=True, args=None) :
        indent = root.get('indent') == 'True'
        num_ops = root.get('num_ops')
        if num_ops is None :
            num_ops = 2
        else :
            num_ops = int(num_ops)

        fn = cls(parent, num_ops=num_ops, indent=indent)
        return fn

    def get_xml(self, name = None, top = True, targets = None, full = False) :
        root = GlypherBinaryRowExpression.get_xml(self, name, top, targets, full)

        root.set('indent', str(self.get_p('indent')))

        return root

    def __init__(self, parent, area=(0,0,0,0), lhs=None, rhs=None,
                 subtract=False, num_ops=2, indent=False) :
        if indent :
            self.poser = GlassmakerIndent

        symbol = GlypherSymbol(None, '')
        GlypherBinaryRowExpression.__init__(self, parent, symbol, area, lhs=lhs, rhs=rhs, num_ops=num_ops,\
            allow_unary=True, no_brackets=True)
        self.add_properties({'gravity': True, 'indent': False})
        self.set_p('indent', indent)
        self.mes.append('space_array')
        self.mes.append('block')

    def get_python(self, true_mode="python") :
        tab_count = len(filter(lambda anc : anc.am('indent'),
                               self.get_ancestors()))
        if self.poser == GlassmakerIndent :
            tab_count += 1
        tabs = "" if tab_count == 0 else "\t"*tab_count

        return "\n".join([tabs+arg.to_string(true_mode) for arg in self.get_args()])

class GlassmakerBlockCommand(GlypherPhraseGroup) :
    def __init__(self, parent, area=(0,0,0,0), lhs=None, rhs=None) :
        GlypherPhraseGroup.__init__(self, parent, [], area, 'command')
        self.add_properties({'gravity' : True})
        self.mes.append('space_array')
        self.mes.append('blockcommand')

        command = GlypherPhrase(parent)
        self.add_target(command, 'command')
        self.set_lhs_target('command')
        self.append(command)

        self.set_recommending(command)

        colon = GlypherSymbol(parent, ":")
        self.append(colon)

        block_phr = GlypherPhrase(parent)
        self.add_target(block_phr, 'block')
        self.set_rhs_target('block')
        self.append(block_phr, row=1)

        block = GlassmakerBlock(parent, num_ops=1, indent=True)
        # if from XML, block will be wiped when Parser empties block_phr
        block_phr.adopt(block)

        if rhs is  not None :
            block.poss[0].adopt(rhs)
            self.set_recommending(rhs)
        else :
            self.set_recommending(block.poss[0])
        
        if lhs is not None :
            self['command'].adopt(lhs)
        else :
            self.set_recommending(self['command'])

    def get_python(self, true_mode="python") :
        return self['command'].to_string(true_mode) + u":\n" + self.get_entities()[-1].to_string(true_mode)

g.add_phrasegroup_by_class('command', GlassmakerCommand)
g.add_phrasegroup_by_class('block', GlassmakerBlock)
g.add_phrasegroup_by_class('blockcommand', GlassmakerBlockCommand)
Parser.binary_expression_names['command'] = ' '
Parser.binary_expression_names['block'] = ''
Parser.binary_expression_names['blockcommand'] = ':'
Parser.binary_expression_properties_for_symbol[':'] = {'gravity' : True}
Parser.binary_expression_properties_for_symbol[''] = {'gravity' : True}
#Parser.binary_expression_properties_for_symbol[' '] = {'gravity' : True}
