from gi.repository import Gtk
import re
from aobject.utils import *
from gi.repository import Pango
from gi.repository import GObject
from .. import glypher
import copy
from lxml import etree
import cairo
from aobject import aobject
from aobject.paths import *
from ..tablemaker import *

import StringIO

from LanternWidget import *

class LanternCodeView(LanternWidget) :
    ty = 'codeview'
    initial_font_size = 25
    #code = "<pre class='brush:js'>\nfunction foo() {\n\talert('Code here...');\n}\n</pre>"
    code = "test\ntest"
    language = "js"

    def get_auto_aesthete_properties(self) :
        d = LanternWidget.get_auto_aesthete_properties(self)
        d.update({'code' : (str,), 'font' : (str,), 'gutter': (bool,), 'language': (str,)})
        return d

    def get_font(self) :
        return self.font_desc.to_string()

    def restore_to_layouts(self, n) :
        if not self.attached :
            self.layout.execute_script("add_design_codeview('%s', %d)" %\
                                       (self.get_js_id(), n))
            self._code_to_layout()
        self.attached = True

    def change_code(self, code) :
        debug_print(code)
        code = re.sub(r'<p>', "", code)
        code = re.sub(r'</p>', "", code)
        code = re.sub(r'<pre>', "", code)
        code = re.sub(r'</pre>', "", code)
        code = re.sub(r'<br />', "\n", code)
        code = re.sub(r'\\n', "\n", code)
        debug_print(code)
        self.code = code
        if self.attached :
            self._code_to_layout()

    def _get_code_class(self) :
        code_class = ['brush: '+self.get_language()]
        code_class.append('gutter: '+('true' if self.get_gutter() else 'false'))
        return ';'.join(code_class)

    def change_language(self, val) :
        self.language = val
        if self.attached :
            self._code_to_layout()

    def change_gutter(self, val) :
        self.gutter = val
        if self.attached :
            self._code_to_layout()

    # n is available as a js var
    def get_loader_js(self) :
        code = self.get_code()
        code = re.sub(r'\\', r'\\\\', code)
        code = re.sub(r'\n', r'\\n', code)
        code = re.sub(r'\t', r'\\t', code)
        code = re.sub(r'"', r'\\"', code)
        name = self.get_js_id()
        js = '''
        add_presentation_codeview('%s', n);
        fill_code("%s-div", "%s", "%s");
        ''' % (name, name, self._get_code_class(), code)
        return js

    def _code_to_layout(self) :
        code = self.get_code()
        code = re.sub(r'\\', r'\\\\', code)
        code = re.sub(r'\n', r'\\n', code)
        code = re.sub(r'\t', r'\\t', code)
        code = re.sub(r'"', r'\\"', code)
        debug_print(code)
        self.layout.execute_script('fill_code("%s-div", "%s", "%s")' % \
                                   (self.get_js_id(), self._get_code_class(), code))
        #self.layout.execute_script("document.getElementById('%s-code').value='%s'" % \
        #                           (self.get_aname(), self.get_code()))
        #self.layout.execute_script("document.getElementById('%s-code').onchange()" % \
        #                           (self.get_aname()))

    def remove_from_layouts(self) :
        if self.attached :
            self.layout.execute_script("remove_by_id('%s')"%\
                    self.get_js_id())
        self.attached = False

    def make_action_panel(self) :
        codeview_table_maker = PreferencesTableMaker()
        codeview_table_maker.append_row("Font",
            self.aes_method_font_button('font'))
        codeview_table_maker.append_row("Line Numbers",
            self.aes_method_toggle_button('gutter'))
        codeview_table_maker.append_row("Language",
            self.aes_method_entry_update('language'))
        win = codeview_table_maker.make_table()
        win.aes_title = "Glancer View"
        win.show_all()

        return win

    def change_font(self, val) :
        self.font_desc = Pango.FontDescription(val)
        self.presentation_widget.font_desc = self.font_desc

    def __init__(self, slide, design_layout, env=None) :
        self.font_desc = Pango.FontDescription(\
            "Linux Libertine %d"%int(self.initial_font_size))
        LanternWidget.__init__(self,
                               slide,
                               design_layout,
                               name_root='LanternCodeView',
                                 env=env)
        self.action_panel = self.make_action_panel()
