from gi.repository import Gtk
from re import escape
from aobject.utils import *
from gi.repository import Pango
from gi.repository import GObject
from .. import glypher
import copy
from lxml import etree
import cairo
from aobject import aobject
from aobject.paths import *
from ..tablemaker import *

from aobject.aobject import *
import StringIO

from LanternWidget import *

class LanternGlancer(LanternWidget) :
    ty = "glancer"
    glancer = None
    dpi = 72

    #PROPERTIES
    def get_auto_aesthete_properties(self) :
        d = LanternWidget.get_auto_aesthete_properties(self)
        d.update({
            'glancer' : (str, (AOBJECT_CAN_NONE,)),
            'dpi' : (int,),
        })
        return d
    #BEGIN PROPERTIES FUNCTIONS
    #def get_dpi(self) :
    #    return self.design_widget.dpi
    #def change_dpi(self, val) :
    #    self.design_widget.dpi = val
    #    self.design_widget.check_canvas_size()
    #    self.emit('redraw-request')

    def change_glancer(self, val) :
        if val == '' :
            val = None

        if val != self.glancer :
            if self.glancer is not None :
                glancer = aobject.get_object_from_dictionary(self.glancer)
                glancer.plotter.canvas.disconnect(self.glancer_changed_conn)
            elif val is not None :
                glancer = aobject.get_object_from_dictionary(val)
                self.glancer_changed_conn =\
                    glancer.connect("aesthete-property-change", lambda w, e, s,t:
                                     self._glancer_to_layout())

        self.glancer = val

        self._glancer_to_layout()

    def get_loader_js(self) :
        if not self.attached or self.get_glancer() is None :
            return ""

        glancer = aobject.get_object_from_dictionary(self.get_glancer())

        if glancer is None :
            return ""

        canvas = glancer.fig.canvas
        w, h = glancer.plotter.get_size_inches()
        c_dpi = canvas.get_width_height()[0]/w
        f_dpi = self.get_dpi() # points per inch is const
        svg = StringIO.StringIO()
        ims = cairo.SVGSurface(svg, int(f_dpi*w), int(f_dpi*h))
        cr = cairo.Context(ims)

        cr.scale(f_dpi/c_dpi, f_dpi/c_dpi)

        glancer._render_to_canvas(cr)

        ims.finish()

        svg = escape(svg.getvalue())
        return "fill_html('%s-div', '%s')" % \
                                   (self.get_aname(), svg)

    def _glancer_to_layout(self) :
        if not self.attached or self.get_glancer() is None :
            return

        glancer = aobject.get_object_from_dictionary(self.get_glancer())

        if glancer is None :
            return

        canvas = glancer.fig.canvas
        w, h = glancer.plotter.get_size_inches()
        c_dpi = canvas.get_width_height()[0]/w
        f_dpi = self.get_dpi() # points per inch is const
        svg = StringIO.StringIO()
        ims = cairo.SVGSurface(svg, int(f_dpi*w), int(f_dpi*h))
        cr = cairo.Context(ims)

        cr.scale(f_dpi/c_dpi, f_dpi/c_dpi)

        glancer._render_to_canvas(cr)

        ims.finish()

        svg = escape(svg.getvalue())
        self.layout.execute_script("fill_html('%s-div', '%s')" % \
                                   (self.get_aname(), svg))

    def _render_to_canvas(self, cr) :
        canvas = self.fig.canvas
        renderer = RendererCairo (canvas.figure.dpi)
        renderer.set_width_height (*canvas.get_width_height())
        renderer.gc.ctx = cr
        canvas.figure.draw (renderer)
        self.emit('redraw-request')
    #END PROPERTIES FUNCTIONS

    def restore_to_layouts(self, n) :
        if not self.attached :
            self.layout.execute_script("add_glancer('%s', %d)" %\
                                       (self.get_aname(), n))
            self._glancer_to_layout()
        self.attached = True

    def remove_from_layouts(self) :
        if self.attached :
            self.layout.execute_script("remove_by_id('%s')"%\
                    self.get_aname())
        self.attached = False

    def __init__(self, slide, design_layout, env=None) :
        LanternWidget.__init__(self,
                               slide,
                               design_layout,
                               name_root='LanternGlancer',
                                 env=env)
        self.action_panel = self.make_action_panel()

    def make_action_panel(self) :
        win = Gtk.VBox()
        line_table_maker = PreferencesTableMaker()

        glancer_cmbo = Gtk.ComboBox()
        glancer_cmbo.set_model(aobject.get_object_dictionary().get_liststore_by_am('Glancer') ) 
        glancer_cllr = Gtk.CellRendererText()
        glancer_cmbo.pack_start(glancer_cllr, 0) 
        self._glancer_cmbo = glancer_cmbo
        glancer_cmbo.add_attribute(glancer_cllr, 'text', 1) 
        glancer_cmbo.connect("changed", lambda o : \
            self.set_glancer(glancer_cmbo.get_model().get(o.get_active_iter(),0)[0]))

        line_table_maker.append_row("Plot", glancer_cmbo)

        line_table_maker.append_row("DPI", self.aes_method_entry_update('dpi'))

        win.aes_title = "Glancer View"
        win.pack_start(line_table_maker.make_table(), True, True, 0)
        win.show_all()
        return win

    def get_action_panel(self) :
        return self.action_panel

class LanternPresentationGlancer(Gtk.DrawingArea) :
    __gsignals__ = { "draw" : "override", "key-release-event" : "override" ,
         "redraw-request" : (GObject.SignalFlags.RUN_LAST, None, () )}
    glancer = None
    default_blank_size = 100
    basic_canvas_scaling = 1
    dpi = 30
    canvas_width = None
    canvas_height = None

    def do_key_release_event(self, event):

        keyname = Gdk.keyval_name(event.keyval)
        m_control = bool(event.get_state() & Gdk.ModifierType.CONTROL_MASK)
        m_shift = bool(event.get_state() & Gdk.ModifierType.SHIFT_MASK)
        m_alt = bool(event.get_state() & Gdk.ModifierType.MOD1_MASK)
        m_super = bool(event.get_state() & Gdk.EventMask.SUPER_MASK)
        return False

    def __init__(self, env=None) :
        GObject.GObject.__init__(self)
        self.add_events(Gdk.EventMask.KEY_RELEASE_MASK)
        self.set_property("can-focus", True)

    def do_draw(self, cr):
        #cr = self.window.cairo_create()
        #cr.rectangle ( event.area.x, event.area.y, event.area.width, event.area.height)
        #cr.clip()

        self.presentation_draw(cr)

    cache = None
    def check_canvas_size(self) :
        if self.glancer is not None :
            glancer = aobject.get_object_from_dictionary(self.glancer)
            if glancer is not None and glancer.plotter is not None :
                canvas = glancer.plotter.canvas
                w1, h1 = glancer.plotter.fig.get_size_inches()
                w1 *= self.dpi*self.basic_canvas_scaling
                h1 *= self.dpi*self.basic_canvas_scaling
                if self.canvas_width != w1 or self.canvas_height != h1 :
                    self.canvas_width = w1
                    self.canvas_height = h1
                    self.set_size_request(int(self.canvas_width),
                                          int(self.canvas_height))
                    self.cache = cairo.ImageSurface(cairo.FORMAT_ARGB32,
                                                    int(w1/self.basic_canvas_scaling),
                                                    int(h1/self.basic_canvas_scaling))
                    self.recache_canvas()
                    return
        else :
            self.set_size_request(self.default_blank_size, self.default_blank_size)
            self.canvas_width = None
            self.canvas_height = None

    def set_recache_queue(self) :
        self.recache_queue=True
    recache_queue = False
    def recache_canvas(self, cr=None) :
        self.recache_queue = False
        if self.glancer is not None :
            glancer = aobject.get_object_from_dictionary(self.glancer)
            if glancer is not None and glancer.plotter is not None :
                canvas = glancer.plotter.canvas
                w1, h1 = glancer.plotter.fig.get_size_inches()
                w1 *= self.dpi
                h1 *= self.dpi
                #w1, h1 = glancer.plotter.canvas.get_width_height()
                if cr is None :
                    if self.cache is None :
                        self.cache = cairo.ImageSurface(cairo.FORMAT_ARGB32, w1, h1)
                    cr = cairo.Context(self.cache)
                cr.save()
                rat = canvas.figure.dpi/float(self.dpi)
                cr.scale(1/rat, 1/rat)
                renderer = RendererCairo (canvas.figure.dpi)
                renderer.set_width_height (*canvas.get_width_height())
                renderer.gc.ctx = cr
                canvas.figure.draw (renderer)
                cr.restore()
                return
        self.cache = None

    def presentation_draw(self, cr, scaling=None, final=False) :
        if self.recache_queue == True :
            self.recache_canvas()
            self.emit("redraw-request")

        if self.glancer is not None :
            glancer = aobject.get_object_from_dictionary(self.glancer)
            if glancer is not None and glancer.plotter is not None :
                canvas = glancer.plotter.canvas
                cr.scale(self.basic_canvas_scaling, self.basic_canvas_scaling)
                if final :
                    self.recache_canvas(cr)
                else :
                    if self.cache is None :
                        self.recache_canvas()
                    cr.set_source_surface(self.cache, 0, 0)
                    cr.paint()
        else :
            cr.save()
            cr.rectangle(self.allocation.x,
                         self.allocation.y,
                         self.allocation.width,
                         self.allocation.height)
            cr.set_source_rgba(0.4, 0.4, 0.4, 1.0)
            cr.fill()
            cr.restore()
