from gi.repository import Gtk
import re
from aobject.utils import *
from gi.repository import Pango
from gi.repository import GObject
from .. import glypher
import copy
from lxml import etree
import cairo
from aobject import aobject
from aobject.paths import *
from ..tablemaker import *

import StringIO

from LanternWidget import *

class LanternImage(LanternWidget) :
    ty = 'image'
    fallback_design = 'no-image.design.svg'
    fallback_presentation = 'no-image.presentation.svg'
    alignment = 'left'


    def get_auto_aesthete_properties(self) :
        d = LanternWidget.get_auto_aesthete_properties(self)
        d.update({'filename' : (str,(aobject.AOBJECT_CAN_NONE,)),
                  'size' : (aobject.string_to_int_tup,(aobject.AOBJECT_CAN_NONE,)),
                  'offset' : (aobject.string_to_float_tup,(aobject.AOBJECT_CAN_NONE,)),
                  'alignment' : (str,)})
        return d

    def restore_to_layouts(self, n) :
        if not self.attached :
            self.layout.execute_script("add_design_image_object('%s', '%s', %d)" %\
                                       (self.get_js_id(), self.get_filename(), n))
            self._image_to_layout()
        self.attached = True

    def change_offset(self, offset) :
        self.offset = offset
        debug_print(offset)
        if self.attached :
            self._image_to_layout()

    def change_size(self, size) :
        self.size = size
        if self.attached :
            self._image_to_layout()

    def change_filename(self, filename) :
        self.filename = filename
        if self.attached :
            self._image_to_layout()

    # n is available as a js var
    def get_loader_js(self) :
        image = self.get_filename()
        if image is None :
            image = self.fallback_presentation
        name = self.get_js_id()
        js = '''
        add_presentation_image_object('%s', '%s', n);
        document.getElementById("%s").style.textAlign = "%s";
        ''' % (name, image, name, self.get_alignment())

        lt = self.get_offset()
        if lt is not None :
            js += '''
                document.getElementById("%s-div").style.left = %lf;
                document.getElementById("%s-div").style.top = %lf;
            ''' % (name, lt[0], name, lt[1])

        wh = self.get_size()
        if wh is not None :
            js += '''
        document.getElementById("%s-img").style.width = "%d";
        document.getElementById("%s-img").style.height = "%d";
            ''' % (name, wh[0], name, wh[1])
        return js

    def _image_to_layout(self) :
        image = self.get_filename()
        if image is None :
            image = get_share_location()+'images/lantern/'+self.fallback_design
        name = self.get_js_id()
        self.layout.execute_script('document.getElementById("%s-img").src = "%s"' % \
                                   (name, image))
        self.layout.execute_script('document.getElementById("%s").style.textAlign = "%s"' % \
                                   (name, self.get_alignment()))
        lt = self.get_offset()
        if lt is not None :
            self.layout.execute_script('''
                $("#%s-div").css("left", %lf);
                $("#%s-div").css("top", %lf);
            ''' % (name, lt[0], name, lt[1]))
        wh = self.get_size()
        if wh is not None :
            self.layout.execute_script('''
                document.getElementById("%s-img").style.width = "%d";
                document.getElementById("%s-img").style.height = "%d";
            ''' % (name, wh[0], name, wh[1]))
        self.layout.execute_script('update()')

    def remove_from_layouts(self) :
        if self.attached :
            self.layout.execute_script("remove_by_id('%s')"%\
                    self.get_js_id())
        self.attached = False

    def make_action_panel(self) :
        image_table_maker = PreferencesTableMaker()
        image_table_maker.append_row("Filename",
            self.aes_method_file_button('filename'))
        image_table_maker.append_row("Offset",
            self.aes_method_entry_update('offset'))
        image_table_maker.append_row("Size",
            self.aes_method_entry_update('size'))

        win = image_table_maker.make_table()
        win.aes_title = "Image"
        win.show_all()

        return win

    def __init__(self, slide, design_layout, env=None) :
        LanternWidget.__init__(self,
                               slide,
                               design_layout,
                               name_root='LanternImage',
                                 env=env)
        self.action_panel = self.make_action_panel()
        slide.deployment_manifest.add((self.fallback_presentation,
                                      'images/lantern/'+self.fallback_presentation))
