import time
from _aesthete_c import get_svg_from_clipboard
import pickle
from gi.repository import Gtk
import cgi
import re
from gi.repository import WebKit
import tarfile
from matplotlib.backends.backend_gtk3cairo import RendererGTK3Cairo as RendererCairo
from aobject.utils import *
from gi.repository import Pango
from gi.repository import GObject, Gdk, GdkPixbuf
import os
import shutil
from .. import glypher
import copy
import urllib
from lxml import etree
from aobject import aobject
from aobject.paths import *
from ..tablemaker import *

import StringIO
from SVG import *
from Image import *
from Video import *
from TextView import *
from CodeView import *
from Glancer import *
from Glypher import *
from Containers import *

GLOSSER_PRESENTATION = True
GLOSSER_DESIGN = False

SPRONK_SETTING = 1
SPRONK_ACTION = 2

black_pattern = cairo.SolidPattern(0.,0.,0.,1.)
paper_pattern = cairo.ImageSurface.create_from_png(get_share_location()+'images/patterns/paper.png')
if paper_pattern is not None :
    paper_pattern = cairo.SurfacePattern(paper_pattern)
    paper_pattern.set_extend(cairo.EXTEND_REPEAT)

layout_background_pattern = { \
                GLOSSER_PRESENTATION : black_pattern,
                GLOSSER_DESIGN : paper_pattern,
                            }
widget_types = {
           'video' : ('video', Gtk.STOCK_MEDIA_PLAY, LanternVideo,
                       lambda s : render_stock(s,Gtk.STOCK_JUSTIFY_FILL)),
           'image' : ('Image', Gtk.STOCK_MEDIA_RECORD, LanternImage,
                       lambda s : render_stock(s,Gtk.STOCK_JUSTIFY_FILL)),
           'svg' : ('SVG', Gtk.STOCK_MEDIA_RECORD, LanternSVG,
                       None),
            'textview' : ('Text', Gtk.STOCK_JUSTIFY_FILL, LanternTextView,
                        lambda s : render_stock(s,Gtk.STOCK_JUSTIFY_FILL)),
            'codeview' : ('Code', Gtk.STOCK_CONVERT, LanternCodeView,
                        lambda s : render_stock(s,Gtk.STOCK_CONVERT)),
            'glyphentry' : ('Equation', 'aes-glypher', LanternGlyphEntry,
                            lambda s : render_stock(s,'aes-glypher')),
            'glancer' : ('Plot', 'aes-glancer', LanternGlancer,
                         lambda s : render_stock(s,'aes-glancer')),
            'vbox' : ('VBox', Gtk.STOCK_GO_DOWN, LanternVBox,
                         lambda s : render_stock(s,Gtk.STOCK_GO_DOWN)),
            'bullets' : ('VBox', Gtk.STOCK_SORT_ASCENDING, LanternBullets,
                         lambda s : render_stock(s,Gtk.STOCK_SORT_ASCENDING)) }

aname_root_to_ty = {}
for ty in widget_types.keys() :
    aname_root_to_ty[widget_types[ty][2].__name__] = ty

def _make_lambda_iw(obj, w) :
    return lambda b : obj.insert_widget(w)

def _px_to_float(px_str) :
    w = px_str
    if w.endswith('px') :
        f = float(w[:-2])
    else :
        f = float(w)
    return f
GLOSSER_SPRONK_CELL_NORMAL = 0
GLOSSER_SPRONK_CELL_OPEN = 1
GLOSSER_SPRONK_CELL_CLOSE = 2
GLOSSER_SPRONK_CELL_INITIAL = 3
GLOSSER_SPRONK_CELL_FINAL = 4
GLOSSER_SPRONK_CELL_OVERVIEW = 5
GLOSSER_SPRONK_CELL_BLANK = 6

lantern_spronk_cell_colours = {\
        GLOSSER_SPRONK_CELL_OPEN : ( (.7,.7,.7), (1.,.7,.5)),
        GLOSSER_SPRONK_CELL_INITIAL : ( (.6,.6,.6), (1.,.5,0.) ),
        GLOSSER_SPRONK_CELL_NORMAL : ( (.6,.6,.6), (1.,.5,0.) ),
        GLOSSER_SPRONK_CELL_FINAL : ( (.7,.7,.7), (1.,.5,0.) ),
        GLOSSER_SPRONK_CELL_CLOSE : ( (.7,.7,.7), (1.,.7,.5) ),
        GLOSSER_SPRONK_CELL_OVERVIEW : ( (.7,.7,.7), (1.,.5,0.) ),
        GLOSSER_SPRONK_CELL_BLANK : ( (1.,1.,1.), (1.,1.,1.) ),
                              }

# Workaround known WebKit bug: 79586
def _fix_svg(html) :
    new_svg = ""
    svg_from = 0
    for n in re.finditer(r'(<svg.*?<\/svg>)', html) :
        new_svg += html[svg_from:n.start(0)]
        new_svg += re.sub(r' xlink=', r' xmlns:xlink=',
                    re.sub(r' href=', r'xlink:href=',
                    n.group(0)))
        svg_from = n.end(0)
    new_svg += html[svg_from:]
    return new_svg

class LanternSpronkCellRenderer(Gtk.CellRenderer) :
    __gproperties__ = {
        "text": (GObject.TYPE_STRING, "text", "text", "", GObject.PARAM_READWRITE),
        "active": (GObject.TYPE_BOOLEAN, "active", "active", False, GObject.PARAM_READWRITE),
        "special": (GObject.TYPE_INT, "special", "special",0,6,
                    GLOSSER_SPRONK_CELL_NORMAL, GObject.PARAM_READWRITE),
        "dimmed": (GObject.TYPE_BOOLEAN, "dimmed", "dimmed", False, GObject.PARAM_READWRITE),
    }

    def __init__(self) :
        Gtk.CellRenderer.__init__(self)
        self.spronk_font_face = cairo.ToyFontFace("Linux Libertine 8")

    def do_set_property(self, pspec, value) :
        setattr(self, pspec.name, value)

    def do_get_property(self, pspec) :
        return getattr(self, pspec.name)

    def do_render(self, cr, widget, background_area, cell_area,
                  flags) :
        surface = cr.get_target()
        cr = cairo.Context(surface)
        
        cr.save()
        cr.translate(cell_area.x, cell_area.y)

        cr.rectangle(0, 0, cell_area.width, cell_area.height)
        opacity = 0.5 if self.dimmed else 1
        colour = Gdk.Color(\
          *map(lambda x:x*65535*opacity,
               lantern_spronk_cell_colours[self.special][1 if self.active else 0]))
        colour_list = [colour.red/65535.,
                       colour.green/65535.,
                       colour.blue/65535.,
                       1.]

        colour_list[3] = .5
        cr.set_source_rgba(*colour_list)
        cr.fill()
        colour_list[3] = 1.

        if self.special == GLOSSER_SPRONK_CELL_OPEN :
            cr.move_to(cell_area.width/2, cell_area.height)
            cr.rel_line_to(-10, 0)
            cr.rel_line_to(10, -cell_area.height)
            cr.rel_line_to(10, cell_area.height)
            cr.close_path()
            colour_list[3] = 1.
            cr.set_source_rgba(*colour_list)
            cr.fill()
        elif self.special == GLOSSER_SPRONK_CELL_CLOSE :
            cr.move_to(cell_area.width/2, 0)
            cr.rel_line_to(-10, 0)
            cr.rel_line_to(10, cell_area.height)
            cr.rel_line_to(10, -cell_area.height)
            cr.close_path()
            cr.set_source_rgba(*colour_list)
            cr.fill()
        elif self.special == GLOSSER_SPRONK_CELL_INITIAL :
            cr.move_to(cell_area.width/2,cell_area.height)
            cr.arc(cell_area.width/2, cell_area.height, 10, 3.14159, 0)
            cr.close_path()
            cr.set_source_rgba(*colour_list)
            cr.fill()
        elif self.special == GLOSSER_SPRONK_CELL_FINAL :
            cr.move_to(cell_area.width/2,0)
            cr.arc(cell_area.width/2, 0, 10, 0, 3.14159)
            cr.close_path()
            cr.set_source_rgba(*colour_list)
            cr.fill()
        elif self.special == GLOSSER_SPRONK_CELL_OVERVIEW :
            cr.rectangle(cell_area.width/2-10, 0, 20, cell_area.height)
            cr.set_source_rgba(*colour_list)
            cr.fill()
            cr.set_source_rgba(1,1,1,0.8)
            cr.move_to(3*cell_area.width/8,cell_area.height/2)
            cr.arc(3*cell_area.width/8, cell_area.height/2, cell_area.width/8, 0,
                   2*3.14159)
            cr.close_path()
            cr.stroke()
            cr.move_to(5*cell_area.width/8,cell_area.height/2)
            cr.arc(5*cell_area.width/8, cell_area.height/2, cell_area.width/8,
                   0, 2*3.14159)
            cr.close_path()
            cr.stroke()
        elif self.special == GLOSSER_SPRONK_CELL_NORMAL :
            cr.rectangle(cell_area.width/2-10, 0, 20, cell_area.height)
            cr.set_source_rgba(*colour_list)
            cr.fill()

        cr.restore()

    line_height = 100
    def do_get_size(self, widget, cell_area=None):
        x = 0
        y = 0

        if cell_area is not None :
            x = cell_area.x
            y = cell_area.y

        w = 30
        h = 10

        return (x, y, w, h)

GObject.type_register(LanternSpronkCellRenderer)

class LanternSlideCellRenderer(Gtk.CellRenderer) :
    __gproperties__ = {
        "slide": (GObject.TYPE_PYOBJECT, "slide", "Slide", GObject.PARAM_READWRITE),
    }
    height = 100.
    dims_render = (800., 600.)

    def __init__(self) :
        Gtk.CellRenderer.__init__(self)
        self.slide = None
        self.spronk_font_face = cairo.ToyFontFace("Linux Libertine 5")

    def do_set_property(self, pspec, value) :
        setattr(self, pspec.name, value)

    def do_get_property(self, pspec) :
        return getattr(self, pspec.name)

    def do_render(self, cr, widget, background_area, cell_area,
                  flags) :
        surface = cr.get_target()
        cr = cairo.Context(surface)
        
        cr.save()
        cr.translate(cell_area.x, cell_area.y)

        if self.slide is None or self.template is None :
            return

        ci = self.slide.thumb
        if cr is not None and ci is not None :
            cr.save()
            cr.translate(15, 0)
            cr.scale(self.height/self.dims_render[1],
                     self.height/self.dims_render[1])
            cr.set_source_surface(ci, 0, 0)
            cr.paint()
            cr.restore()
            spnum = len(self.slide.spronks)
            if spnum > 1:
                cr.save()
                mid = self.height/2
                cr.set_source_rgb(.5,.5,.5)
                if spnum > 6 :
                    text = str(len(self.slide.spronks))
                    cr.set_font_face(self.spronk_font_face)
                    cr.set_font_size(8.)
                    for i in range(0, len(text)) :
                        cr.move_to(0, mid-4*len(text)+8*i)
                        cr.show_text(text[i])
                else :
                    cr.save()
                    for i in range(0, spnum) :
                        cr.move_to(0, mid-3*spnum+6*i)
                        cr.rel_line_to(0, 6)
                        cr.rel_line_to(3, -3)
                        cr.close_path()
                        cr.fill()
                    cr.restore()
                cr.restore()
        cr.restore()

    line_height = 100
    def do_get_size(self, widget, cell_area=None):
        x = 0
        y = 0

        if cell_area is not None :
            x = cell_area.x
            y = cell_area.y

        if self.slide is None :
            w = self.line_height
            h = self.line_height
        else :
            w = self.dims_render[0] * self.height/self.dims_render[1]
            h = self.height
            #w = int(self.slide.thumb.get_width())
            #h = int(self.slide.thumb.get_height())

        return (x, y, w, h)

        #ims = self.main_phrase.cairo_cache_image_surface
GObject.type_register(LanternSlideCellRenderer)


def _make_text_tag() :
    el = etree.Element('tspan', role='line')
    return el

class LanternWebkit(WebKit.WebView) :
    _script_queue = None
    def __init__(self) :
        WebKit.WebView.__init__(self)
        self.set_full_content_zoom(True)
        self._script_queue = []
    def get_value(self, id) :
        self.execute_script('get_value("%s")'%id)
        text = self.get_main_frame().get_title()
        if text == '{NULL}' :
            return None
        return text
    def get_text_content(self, id) :
        self.execute_script('text_to_title("%s")' % id)
        text = self.get_main_frame().get_title()
        if text == '{NULL}' :
            return None
        return text
    def get_html(self, js_object=None) :
        if js_object is None :
            js_object = 'document.body.innerHTML'
        self.execute_script('''
                            if (window["send_html_to_title"]!==undefined)
                                send_html_to_title(%s);
                            else
                                document.title = "";
                            ''' % js_object)
        html = self.get_main_frame().get_title()
        if html is None :
            html = ""
        html = re.sub(r'\\n', r"\n", html)
        html = re.sub(r'\\\\', r"\\", html)

        return html
    def queue_script(self, script) :
        self._script_queue.append(script)
        self._actuate_script()
    _reload_queued = False
    def _actuate_script(self) :
        if not self.get_property('load-status') == WebKit.LoadStatus.FINISHED and\
           not self.get_property('load-status') == WebKit.LoadStatus.FIRST_VISUALLY_NON_EMPTY_LAYOUT :
            ssconn = self.connect("onload-event", lambda w, f : \
                (self.disconnect(ssconn), self._actuate_script()))
        elif len(self._script_queue) > 0 and self._reload_queued is False :
            self.execute_script(self._script_queue.pop(0))

class LanternLayout(LanternWebkit) :

    def update_text(self, filled_text) :
        for nodeid in filled_text :
            self.execute_script('javascript:fill_text("%s","%s")' % \
                (nodeid, filled_text[nodeid]))

    __gsignals__ =     { \
                "size-allocate" : "override",
                "navigation-policy-decision-requested" : "override",
                "content-update" : (GObject.SignalFlags.RUN_LAST, None, ()),
                "management-request" : (GObject.SignalFlags.RUN_LAST, None,
                                        (GObject.TYPE_STRING,
                                         GObject.TYPE_PYOBJECT)),
            }

    handle = None
    width = None
    height = None
    line_num = 0
    line_indicators = ()
    default_height = 100.
    lines = None
    prefix = "design"
    internal_dims = (1400.,1100.)

    def do_submit_form(self, request) :
        debug_print(type(request))
    def do_navigation_policy_decision_requested(self, frame, request,
                                                nav_action, pol_dec) :
        uri = request.get_uri()
        debug_print(uri)
        if uri.startswith('lantern:') :
            arr, args = translate_aobject_web_request('lantern', uri)
            if arr[1] == 'property' :
                self.emit("management-request", "set-property", (arr[2], arr[3],":".join(arr[4:])))
            elif arr[1] == 'update' :
                self.emit('content-update')
            elif arr[1] == 'active' :
                self.emit("management-request", "change-line", arr[2])
            elif arr[1] == 'function' :
                if arr[2] == 'mce_save' :
                    #textview = get_object_from_dictionary(args['mce_id'])
                    #textview.set_text(args['mce_textarea'])
                    debug_print(args)
                    self.emit("management-request", "set-property",
                              (args['mce_id'], args['mce_property'], args['mce_textarea']))
            #elif arr[1] == 'method' :
            #    obj = get_object_from_dictionary(arr[1])
            #    if arr[2] == 'mce_save' :
            #        obj

            #    print uri
            #    print arr
            #    print args
                
            pol_dec.ignore()
            return 1
        return 0

    def do_size_allocate(self, allocation) :
        LanternWebkit.do_size_allocate(self, allocation)
        self._resize_to_allocation(allocation)

    def __init__(self, presentation=False):
        LanternWebkit.__init__(self)

        if presentation :
            self.prefix = "presentation"
        else :
            self.prefix = "design"

        self.load_uri('file://'+get_share_location()+'html/lantern/'+self.prefix+'.html')

        self.reload_bypass_cache()

        self.pilcrow_font_face = cairo.ToyFontFace("Linux Libertine 20")
        self.am_presentation = presentation

    def add_svg(self, variant, svg, rect, style='') :
        #svg = re.escape(svg)
        svg = escape_js(svg)
        style = escape_js(style)

        debug_print(style)

        self.execute_script('add_svg(\'%s\', \'%s\',%lf,%lf,%lf,%lf,\'%s\')'%\
                            (variant,svg, rect['x'], rect['y'],
                             rect['width'],rect['height'], style))
        #self.execute_script('document.title = document.documentElement.innerHTML')

    def get_svg(self) :
        svg = self.get_html('document.getElementById("background").innerHTML')
        return svg

    def get_rat(self) :
        al = self.get_allocation()
        swidth = al.width
        sheight = al.height
        twidth = self.width
        theight = self.height
        rath = sheight/float(theight)
        ratw = swidth/float(twidth)
        rat = min(rath, ratw)
        if self.show_decoration :
            rat *= 0.8
        return rat

    def translate_dist(self, d, rev=False) :
        rat = self.get_rat()
        if rev :
            d /= rat
        else :
            d *= rat

        return d

    def translate_dim(self, w, h) :
        rat = self.get_rat()

        w *= rat
        h *= rat

        return w, h

    def translate_body_pos(self, x, y, rev=False) :
        if not rev :
            x += self.body_x
            y += self.body_y
        pos = self.translate_pos(x, y, rev=rev)
        if rev :
            pos = (pos[0]-self.body_x, pos[1]-self.body_y)
        return pos

    def translate_pos(self, x, y, rev=False) :
        al = self.get_allocation()
        swidth = al.width
        sheight = al.height
        twidth = self.width
        theight = self.height
        rat = self.get_rat()

        if rev :
            x -= .5*(swidth-rat*twidth)
            y -= .5*(sheight-rat*theight)
            x /= rat
            y /= rat
        else :
            x *= rat
            y *= rat
            x += .5*(swidth-rat*twidth)
            y += .5*(sheight-rat*theight)

        return x, y

    def draw(self, cr, swidth, sheight):
        cr.save()
        cr.set_source(layout_background_pattern[self.am_presentation])
        cr.rectangle(0, 0, swidth, sheight); cr.fill()

        if self.handle is not None :
            rat = self.get_rat()
            twidth = self.width
            theight = self.height
            #rat = swidth/float(twidth)
            #if self.show_decoration :
            #    rat *= 0.8
            cr.translate(0.5*(swidth-rat*twidth),
                         0.5*(sheight-rat*theight))
            cr.scale(rat, rat)

            if self.show_decoration :
                blurrad = 2.
                for i in range(1,10) :
                    cr.rectangle(-blurrad*i,-blurrad*i,twidth+2*blurrad*i,theight+2*blurrad*i)
                    cr.set_source_rgba(0.,0.,0.,0.5/float(i))
                    cr.fill()

            self.handle.render_cairo(cr)

            if self.show_decoration :
                y = 0
                i = 0
                current_y = None
                current_height = None
                for line in self.lines :
                    if line is None :
                        continue
                    cr.set_source_rgba(0.85,0.85,0.85,1.)
                    cr.rectangle(self.body_x-13, self.body_y+y, 10., line[1].h)
                    cr.fill()

                    if i == self.line_num :
                        current_y = y
                        current_height = line[1].h

                    y += line[1].h
                    i += 1

                if current_y is None :
                    current_y = y

                if self.current_widget is not None :
                    cr.set_source_rgba(0.5,0.5,0.5,1.)
                    cr.rectangle(self.body_x-18,
                                 self.body_y+self.current_widget.y, 5,
                                 self.current_widget.h)
                    cr.fill()

                rx, ry = (self.body_x-13, self.body_y+current_y)

                if current_height is not None :
                    linear_gradient = cairo.LinearGradient(rx, ry, rx,
                                                           ry+current_height)
                    linear_gradient.add_color_stop_rgba(0, 0.85, 0.85, 0.85, 1.)
                    linear_gradient.add_color_stop_rgba(0.25, 0.7, 0.7, 0.7, .8)
                    linear_gradient.add_color_stop_rgba(0.75, 0.7, 0.7, 0.7, .8)
                    linear_gradient.add_color_stop_rgba(1, 0.85, 0.85, 0.85, 1.)
                else :
                    current_height = self.default_height
                    linear_gradient = cairo.LinearGradient(rx, ry, rx,
                                                           ry+current_height)
                    linear_gradient.add_color_stop_rgba(0, 0.85, 0.85, 0.85, 1.)
                    linear_gradient.add_color_stop_rgba(0.5, 0.7, 0.7, 0.7, .2)
                    linear_gradient.add_color_stop_rgba(1, 0.85, 0.85, 0.85, 0.)

                cr.set_source(linear_gradient)

                cr.rectangle(rx, ry, 10., current_height)
                cr.fill()
        
                cr.move_to(self.body_x-30, self.body_y+current_y)
                cr.set_source_rgba(0.8,0.8,0.8,1.)
                cr.set_font_face(self.pilcrow_font_face)
                cr.set_font_size(20.)
                cr.show_text(u'\u00b6')
                #cr.set_source_rgba(0.4,0.4,0.4,1.)
                i = 0
                for ind in self.line_indicators :
                    cr.save()
                    cr.translate(self.body_x-35,
                                 self.body_y+current_y+\
                                 i*20)

                    if isinstance(ind, GdkPixbuf.Pixbuf) :
                        cr.set_source_pixbuf(ind, 0, 0)
                        cr.paint()
                        cr.stroke()
                    else :
                        cr.move_to(4, 20)
                        cr.show_text(ind)
                    cr.restore()
                    i += 1

                #cr.set_source_rgba(0.8,0.8,0.8,1.)
                #cr.set_line_width(10.)
                #cr.move_to(self.body_x-8, self.body_y+current_y)
                #cr.rel_line_to(0, current_height)
                #cr.stroke()
        
        cr.restore()

    def _resize_to_allocation(self, allocation=None) :
        self.set_zoom_level(min(allocation.width/self.internal_dims[0],
                                allocation.height/self.internal_dims[1]))

class LanternSlide(aobject.AObject) :
    title = None
    num = None
    thumb_html = ""
    thumb = None
    lines = None
    spronks = None
    spronk_specials = None
    variant = 'basic'

    _filled_text = None
    def _fill_text(self, nodeid, text) :
        self._filled_text[nodeid] = text
        #node = root.xpath('//*[@id=\'%s\']'%nodeid)
    
        #if len(node) != 1 :
        #    return
        #    #raise RuntimeError('Malformed template SVG :'+\
        #    #                   ' Expected 1 id to be '+nodeid)
    
        #node = node[0]
    
        ##for child in node :
        ##    node.remove(child)
    
        ##text_node = _make_text_tag()
        #text_node = node[0]
        #text_node.text = text
        #node.append(text_node)

    def aes_get_parameters(self) :
        return {'num' : self.get_num()}

    def pre_spronk(self) :
        for l in self.lines :
            if l is None :
                continue
            l[1].initial_spronk()
            if l[1].container :
                for w in l[1].contained :
                    w.initial_spronk()

    def execute_all_spronks(self) :
        self.pre_spronk()
        for n in range(0, len(self.spronks)) :
            self.execute_spronk(n)

    def execute_spronk(self, num, reverse=False) :
        n = 1 if reverse else 0
        for action_pair in self.spronks[num] :
            if action_pair[n] is not None :
                action_pair[n]()

    @staticmethod
    def _name_from_item(item) :
        if isinstance(item, aobject.AObject) :
            name = item.get_js_id()
        else :
            name = str(item)
        return name

    def _get_spronk_group(self, num) :
        if isinstance(num, int) :
            return self.spronks[num]
        elif num in self.spronk_specials :
            return self.spronk_specials[num]
        return None

    def unset_for_spronk(self, num, item, setting) :
        spronks, index = self.find_for_spronk(num, SPRONK_SETTING, item, setting)
        if index is not None :
            del spronks[index]
        self._update_lsto_for_spronk(spronks)

    def set_for_spronk(self, num, item, setting, value) :
        spronks, index = self.find_for_spronk(num, SPRONK_SETTING, item, setting)

        name = self._name_from_item(item)
        if spronks is None :
            raise RuntimeError('Add spronks first!')

        if index is None :
            spronks.append((SPRONK_SETTING, name, setting, value))
        else :
            spronks[index] = (SPRONK_SETTING, name, setting, value)
        self._update_lsto_for_spronk(spronks)

    def _update_lsto_for_spronk(self, spronk) :
        if spronk != self.spronk_setting_spronk :
            return
        self.spronk_setting_lsto.clear()
        for line in spronk :
            if line[0] == SPRONK_SETTING :
                self.spronk_setting_lsto.append((line[1],line[2],line[3]))
        self.spronk_setting_lsto.append(('','',''))

    spronk_setting_spronk = None
    def spronk_to_action_panel(self, num) :
        spronk = self._get_spronk_group(num)
        if spronk is None :
            return

        self.spronk_setting_spronk = spronk
        if self.spronk_win.get_parent() is not None :
            self.spronk_win.get_parent().remove(self.spronk_win)

        self._update_lsto_for_spronk(spronk)
        self.spronk_win.show_all()

        self.get_aenv().action_panel.to_action_panel(self.spronk_win)

    def _update_ids(self, from_id, to_id) :
        for spronk in self.spronks+self.spronk_specials.values() :
            for action_tuple in spronk :
                if action_tuple[1] == from_id :
                    spronk[spronk.index(action_tuple)] = \
                        (action_tuple[0], to_id, action_tuple[2], action_tuple[3])
            self._update_lsto_for_spronk(spronk)

    @staticmethod
    def _spronk_to_js(num, l, spronk) :
        enter_js = ""
        leave_js = ""
        for n in range(len(spronk)) :
            action_tuple = spronk[n]
            js_obj = "document.getElementById('%s')" % action_tuple[1]
            if action_tuple[0] == SPRONK_SETTING :
                enter_js += "lantern_stored_%d_%s_%d = %s.%s;\n" % \
                        (num, l, n, js_obj, action_tuple[2])
                enter_js += "%s.%s = '%s';\n" % \
                        (js_obj,action_tuple[2],action_tuple[3])
                leave_js += "%s.%s = lantern_stored_%d_%s_%d;\n" % \
                        (js_obj,action_tuple[2], num, l, n)
        return enter_js, leave_js

    def spronks_to_js(self) :
        total_js = "current_spronk = null;\nfunction get_%d_spronk_count() { return %d; }\n" % (self.num, len(self.spronks))

        total_enter_js = "function enter_%d_spronk(spronk) {\nleave_spronk(current_spronk);\n" % self.num
        total_leave_js = "function leave_%d_spronk(spronk) {\n" % self.num
        for l in range(len(self.spronks)) :
            spronk = self.spronks[l]
            if len(spronk) == 0 :
                continue

            enter_js, leave_js = self._spronk_to_js(self.num, str(l), spronk)
            total_enter_js += """
              if (spronk == %d) {
                %s
              }
            """ % (l, enter_js)
            total_leave_js += """
              if (spronk == %d) {
                %s
              }
            """ % (l, leave_js)

        enter_js, leave_js = self._spronk_to_js(self.num, 'overview',self.spronk_specials['overview'])
        total_enter_js += """
          if (spronk == 'overview') {
            %s
          }
        """ % enter_js
        total_leave_js += """
          if (spronk == 'overview') {
            %s
          }
        """ % leave_js

        total_enter_js += "current_spronk = spronk;\n}\n"
        total_leave_js += "}\n"
        total_js += total_enter_js + total_leave_js

        return total_js
            
    def get_for_spronk(self, num, item, setting) :
        spronks, index = self.find_for_spronk(num, SPRONK_SETTING, item, setting)
        if index is None :
            return None
        return spronks[index][3]

    def find_for_spronk(self, num, stype, item, setting) :
        name = self._name_from_item(item)
        spronks = self._get_spronk_group(num)

        if spronks is None :
            return None, None

        for action_tuple in spronks :
            if action_tuple[0] == stype and action_tuple[1] == name \
                    and action_tuple[2] == setting :
                return spronks, spronks.index(action_tuple)
        return spronks, None

    def remove_from_spronk(self, num, action) :
        spronks = self.spronks[num]
        for action_pair in spronk :
            if action == action_pair[0] :
                spronk.remove(action_pair)

    def remove_from_spronks(self, action) :
        for spronk in self.spronks :
            for action_pair in spronk :
                if action == action_pair[0] :
                    spronk.remove(action_pair)

    #PROPERTIES
    def get_auto_aesthete_properties(self):
        return {
            'num' : (int,),
            'title' : (str,),
            'variant' : (str,),
            'widgets' : (string_to_aobject_tup,),
            'spronks_pickled' : (str,),
            'spronk_specials_pickled' : (str,),
        }
    #BEGIN PROPERTIES FUNCTIONS
    def get_widgets(self) :
        return tuple([l[1] for l in filter(lambda l:l is not None, self.lines)])

    def get_spronks_pickled(self) :
        return pickle.dumps(self.spronks)
    def get_spronk_specials_pickled(self) :
        return pickle.dumps(self.spronk_specials)
    def change_spronks_pickled(self, var) :
        self.spronks = pickle.loads(var)
        self.reload()
    def change_spronk_specials_pickled(self, var) :
        self.spronk_specials = pickle.loads(var)
        self.reload()
    def get_title(self) :
        return self.title if self.title is not None else ''
    def change_title(self, val) :
        self.title = val if val != "" else None
        self.reload()
    def change_variant(self, val) :
        self.variant = val
        self.reload()
    #END PROPERTIES FUNCTIONS

    def reload(self) :
        self.reload_slide(self.num, change_current=False)

    def _set_spronk_from_lsto(self) :
        if self.spronk_setting_spronk is None :
            return
        del self.spronk_setting_spronk[:]
        for vals in self.spronk_setting_lsto :
            if vals[0] != '' :
                self.spronk_setting_spronk.append((SPRONK_SETTING, vals[0], vals[1], vals[2]))
        self._update_lsto_for_spronk(self.spronk_setting_spronk)
        self.reload()

    def do_spronk_trvw_edited(self, cell, path, new_text, col) :
        self.spronk_setting_lsto[path][col] = new_text
        self._set_spronk_from_lsto()

    def _setup_spronk_action_panel(self) :
        spronk_trvw = Gtk.TreeView(self.spronk_setting_lsto)
        spronk_scwd = Gtk.ScrolledWindow()
        spronk_scwd.add(spronk_trvw)
        self.spronk_win.append_page(spronk_scwd, Gtk.Label("Settings"))

        spronk_name_trvc = Gtk.TreeViewColumn('ID')
        spronk_trvw.append_column(spronk_name_trvc)
        spronk_name_crtx = Gtk.CellRendererText()
        spronk_name_crtx.set_property('editable', True)
        spronk_name_crtx.connect_after('edited', self.do_spronk_trvw_edited, 0)
        spronk_name_trvc.pack_start(spronk_name_crtx, True)
        spronk_name_trvc.add_attribute(spronk_name_crtx, 'text', 0)

        spronk_setting_trvc = Gtk.TreeViewColumn('Setting')
        spronk_trvw.append_column(spronk_setting_trvc)
        spronk_setting_crtx = Gtk.CellRendererText()
        spronk_setting_crtx.set_property('editable', True)
        spronk_setting_crtx.connect_after('edited', self.do_spronk_trvw_edited,1)
        spronk_setting_trvc.pack_start(spronk_setting_crtx, True)
        spronk_setting_trvc.add_attribute(spronk_setting_crtx, 'text', 1)

        spronk_value_trvc = Gtk.TreeViewColumn('Value')
        spronk_trvw.append_column(spronk_value_trvc)
        spronk_value_crtx = Gtk.CellRendererText()
        spronk_value_crtx.set_property('editable', True)
        spronk_value_crtx.connect_after('edited', self.do_spronk_trvw_edited, 2)
        spronk_value_trvc.pack_start(spronk_value_crtx, True)
        spronk_value_trvc.add_attribute(spronk_value_crtx, 'text', 2)

        spronk_action_scwd = Gtk.ScrolledWindow()
        self.spronk_win.append_page(spronk_action_scwd, Gtk.Label("Actions"))
        spronk_action_vbox = Gtk.VBox()
        spronk_action_update_butt = Gtk.Button("Update")
        spronk_action_update_butt.connect("clicked", self.do_spronk_action_update_butt_clicked)
        spronk_action_vbox.pack_start(spronk_action_update_butt, False, False, 0)
        spronk_action_scwd.add_with_viewport(spronk_action_vbox)

        self.spronk_action_boxes = {}
        for act in self.spronk_actions :
            spronk_action_vbox.pack_start(Gtk.Label(act), False, False, 0)
            self.spronk_action_boxes[act] = Gtk.TextView()
            scrw = Gtk.ScrolledWindow()
            scrw.set_size_request(-1, 300)
            scrw.add(self.spronk_action_boxes[act])
            spronk_action_vbox.pack_start(scrw, False, False, 0)

        self.spronk_win.show_all()

    def do_spronk_action_update_butt_clicked(self, butt) :
        pass

    spronk_actions = ('stepin', 'stepout', 'jumpin', 'jumpout')
    def __init__(self, num, variant, reload_slide, variant_lsto,
                 deployment_manifest, env=None) :
        self._filled_text = {}
        self.reload_slide = reload_slide
        self.deployment_manifest = deployment_manifest
        self.num = num
        self.thumb = cairo.ImageSurface(cairo.FORMAT_ARGB32, 800, 600)
        self.lines = []
        self.spronks = [ [] ]
        self.spronk_specials = { 'overview' : [ ] }
        self.variant_lsto = variant_lsto

        self.spronk_setting_lsto = Gtk.ListStore(str, str, str)
        self.spronk_win = Gtk.Notebook()
        self.spronk_win.aes_title = "Spronk settings"

        self._setup_spronk_action_panel()

        aobject.AObject.__init__(self, name_root='LanternSlide',
                                 env=env,
                                 view_object=False,
                                 elevate=False)
        self.thumb_manager = PresentationHTMLManager(background="#FFEEDD")
        self.set_variant(variant)
        self.action_panel = self.make_action_panel()

    def render_widgets_to_cairo(self, cr, ignore_spronks=True) :
        cr.save()
        for l in self.lines :
            cr.save()
            cr.translate(l[1].x, l[1].y)
            cr.restore()
        cr.restore()

    def remove_line(self, n) :
        self.remove_widget(self.lines[n][1])
        self.lines.remove(self.lines[n])

    def remove_widget(self, widget) :
        widget.remove_from_layouts()

        for spronk in self.spronks+self.spronk_specials.values() :
            for action_tuple in spronk :
                if action_tuple[1] == widget.get_js_id() :
                    del spronk[spronk.index(action_tuple)]
            self._update_lsto_for_spronk(spronk)

    def insert_line(self, n, ty, widget, line_height, alignment) :
        debug_print(0, SPRONK_SETTING, widget,'style.visibility')
        debug_print(self.find_for_spronk(0, SPRONK_SETTING, widget,'style.visibility'))
        if self.find_for_spronk(0, SPRONK_SETTING, widget, 'style.visibility')[1] is None:
            self.set_for_spronk(0, widget, 'style.visibility', 'visible')
            self.set_for_spronk('overview', widget, 'style.visibility', 'visible')

        for i in range(len(self.lines), n+1) :
            self.lines.append(None)

        if self.lines[n] is None :
            self.lines[n] = [ty, widget, line_height, alignment]
        else :
            self.lines.insert(n, [ty, widget, line_height, alignment])

    def set_title(self, title='') :
        self.change_property('title', title)
    def get_num(self) :
        return self.num

    def set_thumb_html(self, thumb_html) :
        self.thumb_html = thumb_html
    def get_thumb_html(self) :
        return self.thumb_html

    def fill_in_text(self) :
        title_text = self.title
        if title_text is None :
            title_text = '[NO TITLE]'
        self._fill_text('lantern_title', title_text)

        num_text = str(self.get_num()+1)
        self._fill_text('lantern_number', num_text)

    def make_action_panel(self) :
        win = Gtk.VBox()
        line_table_maker = PreferencesTableMaker()
        line_table_maker.append_row("Title",
                                    self.aes_method_entry_update("title"))

        variant_cmbo = Gtk.ComboBox()
        variant_cmbo.set_model(self.variant_lsto)
        variant_crtx = Gtk.CellRendererText()
        variant_cmbo.pack_start(variant_crtx, True)
        variant_cmbo.add_attribute(variant_crtx, 'text', 1)
        line_table_maker.append_row('Variant',
                                    variant_cmbo)
        self.aes_method_automate_combo(variant_cmbo, 'variant', 0)

        win.aes_title = "Slide %s" % self.get_num()
        win.pack_start(line_table_maker.make_table(), True, True, 0)
        win.show_all()
        return win

class LanternBasic(Gtk.VBox) :
    presentation = None

    def go_previous(self) :
        self.set_slide(self.slides.index(self.current_slide)-1)
    def go_next(self) :
        self.set_slide(self.slides.index(self.current_slide)+1)
    def go_first(self) :
        self.set_slide(0)
    def go_last(self) :
        self.set_slide(len(self.slides)-1)

    def do_key_release_event(self, event):

        keyname = Gdk.keyval_name(event.keyval)
        state = event.get_state()[1]
        m_control = bool(state & Gdk.ModifierType.CONTROL_MASK)
        m_shift = bool(state & Gdk.ModifierType.SHIFT_MASK)
        m_alt = bool(state & Gdk.ModifierType.MOD1_MASK)
        m_super = bool(state & Gdk.ModifierType.SUPER_MASK)
        
        if m_control and not m_alt and not m_shift :
            if keyname == 'Home' :
                self.go_first()
                return True
            if keyname == 'End' :
                self.go_last()
                return True
        if keyname == 'Page_Up' :
            self.go_previous()
        if keyname == 'Page_Down' :
            self.go_next()
        if keyname == 'Escape' and self.presentation is not None :
            self.get_parent().unfullscreen()
            self.get_parent().hide()
            self.presentation.grab_focus()
            return True

        return False

    __gsignals__ =     { \
                 "key-release-event" : "override",\
            }
    slide_list = None
    
    default_template = 'basic'
    current_slide = None
    current_xml = None
    current_variant = ""

    template = None

    conference = None
    date = None
    institute = None
    presenter = None
    layout_update_conn = None

    def set_current_line_alignment(self, al) :
        if self.current_line in range(0, len(self.current_slide.lines)) and \
               self.current_slide.lines[self.current_line] is None :
            self.current_slide.lines[self.current_line][3] = al
            self.do_widget_pos_update(self, None)

    def do_layout_update(self, layout) :
        if self.current_slide is not None :
            self.update_slide_thumb(self.current_slide.num, self.slide_list_trvw.queue_draw)

    def __init__(self, layout=None, template=template, slides=None, complete_widgets_list=None):
        GObject.GObject.__init__(self)
        if layout is None :
            layout = LanternLayout()
        self.layout = layout
        self.connect_after("size-allocate", self.do_widget_pos_update)
        self.layout_update_conn = self.layout.connect_after("content-update", self.do_layout_update)
        self.set_property("can-focus", True)

        self.add_events(Gdk.EventMask.KEY_RELEASE_MASK)

        if template is None :
            template = { 'name' : self.default_template }
        self.template = template

        if slides is None :
            self.slides = []
        else :
            self.slides = slides

        self.layout.show_decoration = False

        self.pack_start(self.layout, True, True, 0)

        self.grab_focus()

        self.set_slide(0)

        self.show_all()

    def do_widget_pos_update(self, widget, req) :
        if self.current_slide is None :
            return 

        y = 0
        for l in self.current_slide.lines :
            if l is None :
                continue
            widget = l[1]
            #new_pos = map(int,self.layout.translate_pos(
            #        self.template['body']['x'],
            #        self.template['body']['y']+y))
            if l[3] == 'left' :
                offset = 0
            elif l[3] == 'centre' :
                offset = self.template["body"]["width"]-widget.w
                offset *= 0.5
            elif l[3] == 'right' :
                offset = self.template["body"]["width"]-widget.w
            #new_pos[0] += int(offset)
            #new_x, new_y = map(int,self.layout.translate_pos(new_pos[0],
            #                                               new_pos[1],
            #                                         rev=True))
            #new_x, new_y = self.template["body"]["x"], self.template["body"]["y"]+y
            new_x = int(offset)
            new_y = y

            #offset = self.layout.translate_pos(body_x, body_y+y)
            if abs(new_x-widget.x) > 5 or abs(new_y-widget.y) > 5:
                widget.move(new_x, new_y)

            #if widget.current_offset[subwidget] != \
            #        offset or\
            #   widget.current_scaling[subwidget] != \
            #        self.layout.translate_dist(1) or \
            #  (widget.x, widget.y) != (new_x,new_y) :
            #    widget.offset(subwidget, *offset)
            #    widget.rescale(subwidget, self.layout.translate_dist(1))
            #    widget.move(new_x, new_y)

            y += widget.h


    def set_template_variant_xml(self, variant, force=False) :
        if variant == self.current_variant and not force :
            return
        self.current_variant = variant
        self.current_xml = copy.deepcopy(self.template['variants'][variant]['xml'])

        root = self.current_xml.getroot()
        body_node = root.xpath('//*[@id=\'lantern_rect\']')

        if len(body_node) != 1 :
            raise RuntimeError('Malformed template SVG :'+\
                               ' Expected one "rect" id to be lantern_rect')

        body_node = body_node[0]
        
        body = { 'x' : _px_to_float(body_node.get('x')),
                 'y' : _px_to_float(body_node.get('y')),
                 'width' : _px_to_float(body_node.get('width')),
                 'height' : _px_to_float(body_node.get('height')) }
        self.template['body'] = body

        body_node.getparent().remove(body_node)

    def fill_details_for_slide(self, slide) :
        conf_text = self.conference
        if conf_text is None : conf_text = ''
        slide._fill_text('lantern_conference', conf_text)
        slide._fill_text('lantern_total', str(len(self.slides)))

        attributes = ['presenter', 'institute', 'date', 'conference']
        for attribute in attributes :
            val = self.__getattribute__(attribute)
            if val is None :
                val = ''
            slide._fill_text('lantern_'+attribute, str(val))
        
        slide._fill_text('lantern_date_big', self.date)
        slide._fill_text('lantern_presenter_institute', "%s (%s)" % (\
                                            self.presenter,
                                            self.institute))

    old_slide = None
    def set_slide(self, n=None, change_current=True, xml=None) :
        have_changed = False
        if n is not None :
            if n < 0 or n >= len(self.slides) :
                return
            slide = self.slides[n]
            if change_current and slide != self.current_slide :
                have_changed = True
                self.current_slide = slide
        else :
            slide = self.current_slide

        # Even if we do not want to change TO this slide, we may still need to
        # update it
        change_current = change_current or slide==self.current_slide

        if slide is None :
            return

        n = slide.num

        if self.current_xml is None :
            return

        self.set_template_variant_xml(slide.variant)

        if xml is None :
            if change_current :
                xml = self.current_xml
            else :
                xml = copy.deepcopy(self.current_xml)

        if have_changed and self.presentation is None and self.get_aenv() is not None and \
                self.get_aenv().action_panel is not None :
            self.get_aenv().action_panel.to_action_panel(slide.action_panel)

        slide.fill_in_text()
        self.fill_details_for_slide(slide)

        self._reload_slide = slide
        self._reload_xml = xml
        self.layout._reload_queued = True
        if self.layout.get_property('load-status') == WebKit.LoadStatus.FINISHED or\
           self.layout.get_property('load-status') == WebKit.LoadStatus.FIRST_VISUALLY_NON_EMPTY_LAYOUT :
            self._actuate_slide()

    def do_export_html(self) :
        self.html_manager = PresentationHTMLManager(background='black')

        chooser = Gtk.FileChooserDialog(\
                      title="Export HTML Presentation",
                                        action=Gtk.FileChooserAction.SELECT_FOLDER,
                      buttons=(Gtk.STOCK_CANCEL,Gtk.ResponseType.CANCEL,Gtk.STOCK_SAVE,Gtk.ResponseType.OK))
        chooser.set_default_response(Gtk.ResponseType.OK)

        all_ffil = Gtk.FileFilter()
        all_ffil.set_name("All files")
        all_ffil.add_pattern("*")
        chooser.add_filter(all_ffil)

        resp = chooser.run()
        self.grab_focus()

        if resp == Gtk.ResponseType.OK :
            uri = chooser.get_uri()
            chooser.destroy()
        else :
            chooser.destroy()
            return

        dir_file = Gio.file_new_for_uri(uri)
        if not dir_file.query_exists(None) :
            dir_file.make_directory(None)

        for deployment_filename, deployment_location in self.deployment_manifest :
            deployment_file = Gio.file_new_for_uri('file:///'+get_share_location()+deployment_location)
            dest_file = Gio.file_new_for_uri(uri+'/'+deployment_filename)
            try :
                deployment_file.copy(dest_file,
                                     Gio.FileCopyFlags.OVERWRITE,
                                     None,
                                     lambda complete, total, data : 1,
                                     None)
            except StandardError as e :
                self.log(0, "(Exporting HTML) %s" % str(e))

        self.html_manager.get_html(self.slides[0], self.template, self,
                 slide_transition='internal', self_launching=True,
                 callback=lambda : self._enact_export_html(uri+'/index.html'))

    def _enact_export_html(self, uri) :
        gf = Gio.file_new_for_uri(uri=uri)
        f = gf.replace('', False, 0, None)
        sf = StringIO.StringIO()

        body_html = self.html_manager.layout.get_html()
        pr = open(get_share_location()+'html/lantern/presentation.html', 'r')
        pr_contents = pr.read()
        pr.close()

        st = re.search(r'^<!--STARTBODY-->', pr_contents, re.MULTILINE).start()
        en = re.search(r'^<!--ENDBODY-->', pr_contents, re.MULTILINE).end()

        html = pr_contents[:st] + body_html + pr_contents[en:]

        html = _fix_svg(html)

        f.write(html, None)

        f.close(None)

        gf = Gio.file_new_for_uri(uri=uri+'.design.html')
        f = gf.replace('', False, 0, None)
        sf = StringIO.StringIO()

        body_html = self.layout.get_html()
        pr = open(get_share_location()+'html/lantern/presentation.html', 'r')
        pr_contents = pr.read()
        pr.close()

        st = re.search(r'^<!--STARTBODY-->', pr_contents, re.MULTILINE).start()
        en = re.search(r'^<!--ENDBODY-->', pr_contents, re.MULTILINE).end()

        html = pr_contents[:st] + body_html + pr_contents[en:]

        html = _fix_svg(html)

        f.write(html, None)

        f.close(None)
    def _actuate_slide(self) :
        debug_print('load finished')
        debug_print_stack()
        debug_print(self.layout._reload_queued)
        if not self.layout._reload_queued :
            return
        self.layout._reload_queued = False

        slide = self._reload_slide
        xml = self._reload_xml

        self._reload_queued = False
        svg = StringIO.StringIO()
        xml.write(svg)

        
        self.layout.add_svg(slide.variant, svg.getvalue(),
                            self.template['body'],
                            self.template['variants'][slide.variant]['fgstyle'])
        self.layout.update_text(slide._filled_text)

        if self.old_slide is not None :
            for line in self.old_slide.lines :
                if line is None :
                    continue
                line[1].remove_from_layouts()
        self.layout.execute_script('reset_slide()');

        for line in slide.lines :
            if line is None :
                continue 
            widget = line[1]
            line[1].restore_to_layouts(slide.lines.index(line))

        debug_print(slide.num)
        self.layout.queue_draw()
        
        self.old_slide = slide

        #slide.pre_spronk()

        #self.do_widget_pos_update(self, None)

        #self.redraw(slide, xml, change_current=change_current)

    def redraw(self, slide=None, xml=None, change_current=True) :
        if slide is None :
            slide = self.current_slide

        if xml is None :
            self.set_template_variant_xml(slide.variant)
            xml = self.current_xml

        if xml is not None and \
                slide is not None :
            svg = StringIO.StringIO()
            xml.write(svg)
            handle = rsvg.Handle()
            handle.write(self.layout.get_svg())
            handle.close()


            if change_current :
                self.layout.width = self.template['width']
                self.layout.height = self.template['height']
                self.layout.handle = handle
                #self.layout.line_num = self.current_line
                self.layout.lines = slide.lines
                #self.layout.current_widget = self.current_widget
                #if self.current_line < len(slide.lines) :
                #    ty = slide.lines[self.current_line][0]
                #    self.layout.line_indicators = \
                #        [ widget_types[ty][3](self.get_style())]
                #    if self.current_widget is not None and self.current_widget != slide.lines[self.current_line][1]:
                #        ty = self.current_widget.ty
                #        self.layout.line_indicators.append(widget_types[ty][3](self.get_style()))
                #else :
                #    self.layout.line_indicators = ( u'\u273c', )
                self.layout.set_foreground_location(self.template['body'])

            self.layout.queue_draw()

class LanternPresentation() :
    presentation = None
    current_spronk = 0

    def go_first(self) :
        self.set_slide(0)
        self.current_spronk = 0
    def go_last(self) :
        self.set_slide(len(self.slides)-1)
        self.current_slide.execute_all_spronks()
        self.current_spronk = len(self.current_slide.spronks)

    def go_previous(self) :
        if self.current_spronk == 0 :
            self.set_slide(self.slides.index(self.current_slide)-1)
        else :
            self.current_spronk -= 1
            self.current_slide.execute_spronk(self.current_spronk, reverse=True)
    def go_next(self) :
        if self.current_spronk == len(self.current_slide.spronks) :
            if self.current_slide == self.slides[-1] :
                return
            else :
                self.set_slide(self.slides.index(self.current_slide)+1)
                self.current_spronk = 0
        else :
            self.current_slide.execute_spronk(self.current_spronk, reverse=False)
            self.current_spronk += 1

    def do_key_release_event(self, event):

        keyname = Gdk.keyval_name(event.keyval)
        state = event.get_state()
        m_control = bool(state & Gdk.ModifierType.CONTROL_MASK)
        m_shift = bool(state & Gdk.ModifierType.SHIFT_MASK)
        m_alt = bool(state & Gdk.ModifierType.MOD1_MASK)
        m_super = bool(state & Gdk.ModifierType.SUPER_MASK)
        
        if keyname == 'Escape' :
            self.window.unfullscreen()
            self.window.hide()
            self.lantern.grab_focus()
            return True

        return False

    def set_slide(self, n) :
        #self.presentation_manager.get_html(self.lantern.slides[n],
        #                                   self.lantern.template,
        #                                   self.lantern,
        #                                   'notify')
        self.presentation_manager.get_html(self.lantern.slides[n],
                                           self.lantern.template,
                                           self.lantern,
                                           'internal')

    def show(self) :
        self.window.show()

    def __init__(self, lantern) :
        self.window = Gtk.Window()
        self.window.show_all()
        self.window.connect_after("show", lambda w : (w.fullscreen(), w.grab_focus()))
        self.window.connect("key-release-event", lambda w, e : self.do_key_release_event(e))
        self.lantern = lantern
        self.presentation_manager = PresentationHTMLManager(self.window,
                                                            background='black')
        self.presentation_manager.layout.connect('navigation-policy-decision-requested',
                            self.do_layout_navigation_policy_decision_requested)
        self.window.hide()

    def do_layout_navigation_policy_decision_requested(self, layout, frame, request,
                                                nav_action, pol_dec) :
        uri = request.get_uri()
        if uri.startswith('lantern:') :
            arr, args = translate_aobject_web_request('lantern', uri)
            if arr[1] == 'slide-change' :
                self.set_slide(int(arr[2]))
            return True

def modify_xml() :
    pass
    #root = current_xml.getroot()
    #body_node = root.xpath('//*[@id=\'lantern_rect\']')

    #if len(body_node) != 1 :
    #    raise RuntimeError('Malformed template SVG :'+\
    #                       ' Expected one "rect" id to be lantern_rect')

    #body_node = body_node[0]
    #
    #body = { 'x' : _px_to_float(body_node.get('x')),
    #         'y' : _px_to_float(body_node.get('y')),
    #         'width' : _px_to_float(body_node.get('width')),
    #         'height' : _px_to_float(body_node.get('height')) }

    #body_node.getparent().remove(body_node)

    #xml = current_xml

    #slide.fill_in_xml(xml)
    #root = xml.getroot()

    #conf_text = lantern.conference
    #if conf_text is None : conf_text = ''
    #slide._fill_text('lantern_conference', conf_text)
    #slide._fill_text('lantern_total', str(len(lantern.slides)))

    #attributes = ['presenter', 'institute', 'date', 'conference']
    #for attribute in attributes :
    #    val = lantern.__getattribute__(attribute)
    #    if val is None :
    #        val = ''
    #    slide._fill_text('lantern_'+attribute, str(val))
    #
    #slide._fill_text('lantern_date_big', lantern.date)
    #slide._fill_text('lantern_presenter_institute', "%s (%s)" % (\
    #                                    lantern.presenter,
    #                                    lantern.institute))

    #svg = StringIO.StringIO()
    #current_xml.write(svg)
    #layout.body_x = body['x']
    #layout.body_y = body['y']
    #layout.body_w = body['width']
    #layout.body_h = body['height']
    #layout.set_size_request(600, 600)
    #layout.set_svg(svg.getvalue())
    #layout.update_text(slide._filled_text)

def escape_js(js) :
            js = re.sub(r'\\', r'\\\\', js)
            js = re.sub(r'"', r'\\"', js)
            js = re.sub(r'\n', r'\\n', js)
            return js

class PresentationHTMLManager() :
    phconn_loaded = None
    phconn_display = None
    _inited = False
    background = "none"

    def __init__(self, window=None, background="none") :
        self.layout = LanternLayout(presentation=GLOSSER_PRESENTATION)
        if window is None :
            self.offwin = Gtk.OffscreenWindow()
            self.offwin.set_default_size(800, 600)
        else :
            self.offwin = window
        self.offwin.add(self.layout)
        self.offwin.show_all()
        self.layout_conn = None
        self.background = background

    def init_slide(self) :
        self.layout.execute_script('init_presentation()')
        self.layout.execute_script('init_slide()')
        self._inited = True
    def go_to_overview(self) :
        if not self._inited :
            self.init_slide()
        self.layout.execute_script('enter_spronk("overview")')

    def get_html(self,slide, template, lantern,
                 slide_transition='external',
                 self_launching=False,
                 callback=None,
                 display_callback=None) :
        if self.phconn_display is not None :
            self.offwin.disconnect(self.phconn_display)
    
        layout = self.layout
    
        self.phconn_loaded = layout.connect_after("onload-event", lambda w, f :\
                self.on_loaded(slide, template, lantern, slide_transition,
                               self_launching, callback, display_callback))
        layout.reload_bypass_cache()
        #layout.execute_script('console.log(document.body.innerHTML)');
    
    def on_loaded(self, slide, template, lantern, slide_transition,
                  self_launching, callback,
                                      display_callback) :
        layout = self.layout
        layout.disconnect(self.phconn_loaded)
    
        if slide is None :
            return
    
        n = slide.num
    
        variants = []
        if slide_transition == 'internal' :
            variants = [ slide.variant for slide in lantern.slides ]
        else :
            variants = [ slide.variant ]

        for variant in variants :
            current_xml = copy.deepcopy(template['variants'][variant]['xml'])
    
            root = current_xml.getroot()
            body_node = root.xpath('//*[@id=\'lantern_rect\']')
    
            if len(body_node) != 1 :
                raise RuntimeError('Malformed template SVG :'+\
                                   ' Expected one "rect" id to be lantern_rect')
    
            body_node = body_node[0]
            body = { 'x' : _px_to_float(body_node.get('x')),
                     'y' : _px_to_float(body_node.get('y')),
                     'width' : _px_to_float(body_node.get('width')),
                     'height' : _px_to_float(body_node.get('height')) }
    
            body_node.getparent().remove(body_node)
    
            slide.fill_in_text()
            lantern.fill_details_for_slide(slide)

            svg = StringIO.StringIO()
            current_xml.write(svg)
            layout.add_svg(variant, svg.getvalue(), body, template['variants'][variant]['fgstyle'])
    
        if slide_transition is not 'internal' :
            layout.execute_script('append_js("function get_N(){return %d;}")' % len(slide.lines))
            layout.execute_script('append_js("function get_slide_number(){return %d;}")' % slide.num)
            layout.execute_script('append_js("function get_slide_variant(){return \'%s\';}")' %slide.variant)
            slides = [ slide ]
        else :
            js_get_N = "function get_N() {\n"
            js_get_variant = "function get_slide_variant() {\n"
            for slide in lantern.slides :
                js_get_N += "  if (get_slide_number()==%d) return %d;\n" % (slide.num, len(slide.lines))
                js_get_variant += "  if (get_slide_number()==%d) return '%s';\n" % (slide.num, slide.variant)
            js_get_N += "}\n"
            js_get_variant += "}\n"

            layout.execute_script('append_js("%s")' % escape_js(js_get_N))
            layout.execute_script('append_js("%s")' % escape_js(js_get_variant))
            layout.execute_script('append_js("slide_num = %d; function get_slide_number(){return slide_num;}")' % slide.num)
            slides = lantern.slides

        for slide in slides :
            for line in slide.lines :
                if line is None :
                    continue
                widget = line[1]
                n = slide.lines.index(line)
                layout.execute_script('append_slot_loader(%d, %d, "%s")' %
                                      (slide.num,
                                       n,escape_js(line[1].get_loader_js())))
                layout.execute_script('append_slot_unloader(%d, %d, "%s")' %
                                      (slide.num,
                                       n,escape_js(line[1].get_unloader_js())))
            set_details_js = "function set_details_%d() {\n" % slide.num
            for nodeid in slide._filled_text :
                set_details_js += '  fill_text("%s","%s");' % \
                        (nodeid, slide._filled_text[nodeid])
                set_details_js += "\n"
            set_details_js += "}\n"
            layout.execute_script('append_js("%s")'%escape_js(set_details_js))

            layout.execute_script('append_js("%s")'%escape_js(slide.spronks_to_js()))

    

        layout.execute_script('append_js("function get_slide_total(){return %d;}")' % len(lantern.slides))
        layout.execute_script('append_js("function get_slide_change_method(){return \'%s\';}")' % slide_transition)

        if display_callback is not None :
            self.phconn_display = self.offwin.connect_after("damage-event", lambda w, e :\
                display_callback(layout.get_html()))


        if self_launching :
            js = """
            function launch(){
                init_presentation();
                init_slide();
            }
            """
            layout.execute_script('append_js("%s");'%escape_js(js))
            layout.execute_script('append_js("window.addEventListener(\'load\',launch,false)");')
        else :
            layout.execute_script('document.documentElement.style.background="%s"'%self.background)
            self.init_slide()
        layout.queue_draw()
    
        if callback is not None :
            if not self.layout.get_property('load-status') == WebKit.LoadStatus.FINISHED and\
               not self.layout.get_property('load-status') == WebKit.LoadStatus.FIRST_VISUALLY_NON_EMPTY_LAYOUT :
                ssconn = self.layout.connect("onload-event", lambda w, f : \
                    (self.disconnect(ssconn), debug_print(ssconn), callback()))
            else :
                callback()

lantern_signals = aobject.asignals.copy()
class Lantern(LanternBasic, aobject.AObject) :
    deployment_manifest = (
        ('presentation.css', 'html/lantern/presentation.css'),
        ('presentation.js', 'html/lantern/presentation.js'),
        ('standard.css', 'html/lantern/standard.css'),
        ('standard.js', 'html/lantern/standard.js'),
    )

    can_save = True
    current_widget = None
    #def change_current_line(self, n=None) :
    #    ret = LanternBasic.change_current_line(self, n=n)

    #    if not ret :
    #        return False

    #    if n < len(self.current_slide.lines) :
    #        self.change_current_widget(self.current_slide.lines[n][1])

    #    self.redraw()

    _offwin_finished_conn = None
    def update_slide_thumb(self, n, draw) :
        slide = self.slides[n]
        if slide is None :
            return
        slide.thumb_manager.get_html(slide, self.template, self,
                              'external', False,
                              slide.thumb_manager.go_to_overview,
                              lambda h:self._update_slide_thumb_from_offwin(slide,draw,h))

    def _update_slide_thumb_from_offwin(self, slide, draw, html) :
        cr = cairo.Context(slide.thumb)
        ci = slide.thumb_manager.offwin.get_surface()
        if cr is not None and ci is not None :
            cr.save()
            cr.set_source_surface(ci, 0, 0)
            cr.paint()
            cr.restore()
        draw()

    __gsignals__ = lantern_signals
    from_spronk = None
    def update_spronk_strip(self) :
        if self.current_slide is None :
            return

        slide = self.current_slide

        self.spronk_buffer.clear()

        w = self.current_widget

        cwh = w is None or (slide.get_for_spronk(0, w, 'style.visibility')=='hidden')
        cwd = (slide.get_for_spronk(0, w, 'style.opacity') is not None)
        self.spronk_buffer.insert(0, (GLOSSER_SPRONK_CELL_OPEN, not
                                      cwh,u'\u21c8', cwd,""))
        self.spronk_buffer.insert(1, (GLOSSER_SPRONK_CELL_INITIAL, not
                                      cwh,'Initial', cwd,"0"))

        for l in range(2, len(slide.spronks)+1) :
            cwh = w is None or (slide.get_for_spronk(l-1, w, 'style.visibility')=='hidden')
            cwd = (slide.get_for_spronk(l-1, w, 'style.opacity') is not None)

            text = ''

            ty = GLOSSER_SPRONK_CELL_NORMAL if len(slide.spronks) > l else \
                 GLOSSER_SPRONK_CELL_FINAL
            self.spronk_buffer.insert(l, (ty, not cwh, text, cwd, str(l-1)))

        self.spronk_buffer.insert(len(slide.spronks)+2,
                                  (GLOSSER_SPRONK_CELL_CLOSE, not cwh,
                                   u'\u21ca', cwd, ""))

        self.spronk_buffer.insert(len(slide.spronks)+3,(GLOSSER_SPRONK_CELL_BLANK,
                                                        False, '', False, ""))

        cwh = w is None or (slide.get_for_spronk('overview', w, 'style.visibility')=='hidden')
        cwd = (slide.get_for_spronk('overview', w, 'style.opacity') is not None)
        self.spronk_buffer.insert(len(slide.spronks)+4,(GLOSSER_SPRONK_CELL_OVERVIEW,
                                                        not cwh,'Overview',
                                                        cwd,'overview'))

        self.slide_list_trvw.queue_draw()

    current_widget = None
    def change_current_line(self, n=None) :

        if self.current_slide is None :
            self.current_line = 0
            return False

        if n is None :
            n = self.current_line

        if n > len(self.current_slide.lines) :
            n = len(self.current_slide.lines)
        elif n < 0 :
            n = 0

        self.current_line = n

        if self.get_current_widget_in_line() is None :
            self.change_current_widget(None)
        elif self.get_current_widget_in_line() == -1 and self.current_slide.lines[n] is not None :
            self.change_current_widget(self.current_slide.lines[n][1])

        self.layout.queue_script('_set_active_slot(%d)' % n)
        self.update_spronk_strip()

    def get_current_widget_in_line(self) :
        if self.current_slide is None :
            return None
        if 0 > self.current_line or len(self.current_slide.lines) <=  self.current_line:
            return None
        if self.current_slide.lines[self.current_line] is None :
            return None
        if self.current_widget == self.current_slide.lines[self.current_line][1] :
            return -1
        if not self.current_slide.lines[self.current_line][1].container :
            return -1
        if self.current_widget not in self.current_slide.lines[self.current_line][1].contained:
            return -1
        return self.current_slide.lines[self.current_line][1].contained.index(self.current_widget)

    current_line = 0
    def change_current_widget(self, widget) :
        if widget == self.current_widget :
            return

        if widget is not None and widget.slide != self.current_slide :
            return

        self.current_widget = widget
        self.update_spronk_strip()

        if widget is not None :
            widget.grab_focus()
            widget.to_action_panel()
        elif self.current_slide is not None :
            self.current_slide.to_action_panel()

    def do_key_release_event(self, event):

        keyname = Gdk.keyval_name(event.keyval)
        state = event.get_state()[1]
        m_control = bool(state & Gdk.ModifierType.CONTROL_MASK)
        m_shift = bool(state & Gdk.ModifierType.SHIFT_MASK)
        m_alt = bool(state & Gdk.ModifierType.MOD1_MASK)
        m_super = bool(state & Gdk.ModifierType.SUPER_MASK)
        
        if m_control and not m_alt and not m_shift :
            if keyname == 'Down' :
                widpos = self.get_current_widget_in_line()

                if widpos is None :
                    self.change_current_line(self.current_line+1)
                else :
                    line_widget = self.current_slide.lines[self.current_line][1]
                    if not line_widget.container or widpos == len(line_widget.contained)-1 :
                        self.change_current_line(self.current_line+1)
                    else :
                        self.change_current_widget(line_widget.contained[widpos+1])
                #if self.current_line < 0 or \
                #   self.current_line >= len(self.current_slide.lines) :
                #    self.change_current_line(self.current_line+1)
                #if line_widget.container and len(line_widget.contained) > 0 :
                #    if self.current_widget is None :
                #        self.current_widget = line_widget.contained[0]
                #        self.redraw()
                #    elif line_widget.contained[-1] == self.current_widget :
                #        self.change_current_line(self.current_line+1)
                #    else :
                #        self.current_widget = \
                #            line_widget.contained[1+
                #            line_widget.contained.index(self.current_widget)]
                #        self.redraw()
                #else :
                #    self.change_current_line(self.current_line+1)
                return True
            elif keyname == 'Up' :
                widpos = self.get_current_widget_in_line()

                if widpos is None or widpos == -1 :
                    self.change_current_line(self.current_line-1)
                else :
                    line_widget = self.current_slide.lines[self.current_line][1]
                    if widpos == 0 :
                        self.change_current_widget(line_widget)
                    else :
                        self.change_current_widget(line_widget.contained[widpos-1])
                #if self.current_line < 0 or \
                #   self.current_line >= len(self.current_slide.lines) :
                #    self.change_current_line(self.current_line-1)
                #line_widget = self.current_slide.lines[self.current_line][1]
                #if line_widget.container and len(line_widget.contained) > 0 :
                #    if self.current_widget is None :
                #        self.change_current_line(self.current_line-1)
                #    elif line_widget.contained[0] == self.current_widget :
                #        self.current_widget = None
                #        self.redraw()
                #    else :
                #        self.current_widget = \
                #            line_widget.contained[-1+
                #            line_widget.contained.index(self.current_widget)]
                #        self.redraw()
                #else :
                #    self.change_current_line(self.current_line-1)
                #return True
            elif keyname == 'BackSpace' :
                self.remove_current_line()
                return True
            elif keyname == 'v' :
                self.do_paste()
                return True

        return LanternBasic.do_key_release_event(self, event)

    def do_paste(self) :
        svg = get_svg_from_clipboard()
        if svg != "" :
            widget = self.insert_widget('svg')
            widget.set_svg(svg)
        print svg

    def remove_current_line(self) :
        if self.current_line in range(0,len(self.current_slide.lines)) :
            self.current_slide.remove_line(self.current_line)
        self.change_current_line()
        self.set_slide()

    def __del__(self) :
        aobject.AObject.__del__(self)

    #PROPERTIES
    def get_auto_aesthete_properties(self):
        return {
            'conference' : (str,(aobject.AOBJECT_CAN_NONE,)),
            'date' : (str,(aobject.AOBJECT_CAN_NONE,)),
            'presenter' : (str,(aobject.AOBJECT_CAN_NONE,)),
            'institute' : (str,(aobject.AOBJECT_CAN_NONE,)),
            'template' : (str,),
        }
    #BEGIN PROPERTIES FUNCTIONS
    def get_template(self, val=None) :
        return self.template['name'] if self.template is not None else ''
    def change_template(self, val) :
        if val == '' :
            val = None
        self.set_template(val)
        self.set_slide()
    def change_date(self, val) :
        if val == '' :
            self.date = None
        else :
            self.date = val
            #self.presentation_instance.date = val
        self.set_slide()
    def change_presenter(self, val) :
        if val == '' :
            self.presenter = None
        else :
            self.presenter = val
            #self.presentation_instance.presenter = val
        self.set_slide()
    def change_institute(self, val) :
        if val == '' :
            self.institute = None
        else :
            self.institute = val
            #self.presentation_instance.institute = val
        self.set_slide()
    def change_conference(self, val) :
        if val == '' :
            self.conference = None
        else :
            self.conference = val
            #self.presentation_instance.conference = val
        self.set_slide()
    #END PROPERTIES FUNCTIONS

    def make_action_panel(self) :
        win = Gtk.VBox()
        line_table_maker = PreferencesTableMaker()
        line_table_maker.append_row("Conf",
                                    self.aes_method_entry_update("conference"))
        line_table_maker.append_row("Presenter",
                                    self.aes_method_entry_update("presenter"))
        line_table_maker.append_row("Date",
                                    self.aes_method_entry_update("date"))
        line_table_maker.append_row("Institute",
                                    self.aes_method_entry_update("institute"))

        template_cmbo = Gtk.ComboBox()
        template_cmbo.set_model(template_lsto)
        template_crtx = Gtk.CellRendererText()
        template_cmbo.pack_start(template_crtx, True)
        template_cmbo.add_attribute(template_crtx, 'text', 1)
        line_table_maker.append_row("Template", template_cmbo)
        self.aes_method_automate_combo(template_cmbo, 'template', 0)

        win.aes_title = "Presentation"
        win.pack_start(line_table_maker.make_table(), True, True, 0)
        win.show_all()

        return win

    def _get_object_by_id(self, js_id) :
        for slide in self.slides :
            for line in slide.lines :
                if line is None :
                    continue
                if line[1].get_js_id() == js_id :
                    return line[1]
        return None

    def do_layout_management_request(self, layout, request_type, request) :
        if request_type == "change-line" :
            self.change_current_line(int(request))
        if request_type == "set-property" :
            debug_print(request)
            obj = self._get_object_by_id(request[0])
            obj.change_property(request[1], request[2])

    def do_slide_list_drag_end(self, slide_list_trvw, dc) :
        new_slides = []
        for i in range(len(self.slides)) :
            slide = self.slide_list[i][1]
            slide.num = i
            self.slide_list[i][0] = slide.num+1
            new_slides.append(slide)
        debug_print(new_slides)
        self.slides = new_slides

    def __init__(self, slides=None, complete_widgets_list=None, env=None):
        self.slide_list = Gtk.ListStore(GObject.TYPE_INT, GObject.TYPE_PYOBJECT)
        self.slide_list.set_sort_column_id(0, Gtk.SortType.ASCENDING)

        self.variant_lsto = Gtk.ListStore(str, str)

        self.template = {'name':self.default_template}
        self.deployment_manifest = set(self.deployment_manifest)

        if slides is None :
            self.slides = []
        else :
            self.slides = slides
        self.layout = LanternLayout(presentation=GLOSSER_DESIGN)
        self.layout.connect("onload-event", lambda w, f : self._actuate_slide())
        self.layout_update_conn = self.layout.connect_after("content-update", self.do_layout_update)
        self.layout.connect("management-request",
                            self.do_layout_management_request)

        self.presentation_instance = LanternPresentation(self)

        GObject.GObject.__init__(self)

        aobject.AObject.__init__(self, name_root="Lantern", env=env,
                                 view_object=True,
                                 elevate=True)

        self.connect_after("size-allocate", self.do_widget_pos_update)
        self.add_events(Gdk.EventMask.KEY_RELEASE_MASK)

        self.layout.show_decoration = True

        self.set_template()
        #self.new_slide(0, 'basic')

        toolbar = Gtk.Toolbar()

        present_tbut = Gtk.ToolButton('aes-lantern')
        present_tbut.connect("clicked", lambda w :(\
                             self.presentation_instance.go_first(),
                             self.presentation_instance.show()))
        toolbar.insert(present_tbut, -1)

        pdf_tbut = Gtk.ToolButton(Gtk.STOCK_SAVE_AS)
        pdf_tbut.connect("clicked", lambda w :\
                             self.do_export_html())
        toolbar.insert(pdf_tbut, -1)
        toolbar.insert(Gtk.SeparatorToolItem(), -1)

        for w in widget_types :
            wt = widget_types[w]
            if wt[3] is not None :
                new_tbut = Gtk.ToolButton(wt[1])
                new_tbut.set_tooltip_text(wt[0])
                new_tbut.connect('clicked', _make_lambda_iw(self, w))
                toolbar.insert(new_tbut, -1)

        toolbar.insert(Gtk.SeparatorToolItem(), -1)

        left_tbut = Gtk.ToolButton(Gtk.STOCK_JUSTIFY_LEFT)
        left_tbut.connect("clicked", lambda w : \
                          self.set_current_line_alignment('left'))
        toolbar.insert(left_tbut, -1)
        centre_tbut = Gtk.ToolButton(Gtk.STOCK_JUSTIFY_CENTER)
        centre_tbut.connect("clicked", lambda w : \
                          self.set_current_line_alignment('centre'))
        toolbar.insert(centre_tbut, -1)
        right_tbut = Gtk.ToolButton(Gtk.STOCK_JUSTIFY_RIGHT)
        right_tbut.connect("clicked", lambda w : \
                          self.set_current_line_alignment('right'))
        toolbar.insert(right_tbut, -1)

        self.pack_start(toolbar, False, False, 0)

        self.layout_hbox = Gtk.HBox()

        self.spronk_buffer = Gtk.ListStore(int, bool, str, bool, str)
        spronk_strip = Gtk.TreeView(self.spronk_buffer)
        spronk_strip.connect("row-activated", self.do_spronk_strip_row_activated)
        spronk_strip.connect_after("button-release-event",
                             self.do_spronk_strip_button_release_event)
        spronk_strip.get_selection().connect("changed",
                                          self.do_spronk_strip_selection_changed)
        spronk_strip.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        spronk_column = Gtk.TreeViewColumn('Spronk')
        spronk_strip.append_column(spronk_column)
        spronk_tvcr = LanternSpronkCellRenderer()
        spronk_column.pack_start(spronk_tvcr, True)
        spronk_column.add_attribute(spronk_tvcr, 'special', 0)
        spronk_column.add_attribute(spronk_tvcr, 'active', 1)
        spronk_column.add_attribute(spronk_tvcr, 'text', 2)
        spronk_column.add_attribute(spronk_tvcr, 'dimmed', 3)

        self.layout_scrw = Gtk.ScrolledWindow();
        self.layout_scrw.add(self.layout);
        self.layout_hbox.pack_start(self.layout_scrw, True, True, 0)
        self.layout_hbox.pack_start(spronk_strip, False, False, 0)

        self.pack_start(self.layout_hbox, True, True, 0)

        self.grab_focus()

        self.set_slide(0)
        #ssconn = self.layout.connect("load-finished", lambda w, f : \
        #    (self.layout.disconnect(ssconn), self.set_slide(0)))
        #self.presentation_instance.set_slide(0)

        self.show_all()

        self.action_panel = self.make_action_panel()
        if self.get_aenv() is not None :
            self.get_aenv().action_panel.to_action_panel(self.action_panel)

    old_sel = None
    old_old_sel = None
    def do_spronk_strip_selection_changed(self, selection) :
        if self.current_slide is None or self.current_widget is None :
            return

        self.old_old_sel = self.old_sel
        self.old_sel = selection.get_selected_rows()

    def do_spronk_strip_button_release_event(self, spronk_strip, event) :
        button = event.get_button()[1]

        state = event.get_state()
        m_control = bool(state & Gdk.ModifierType.CONTROL_MASK)
        selected_rows = spronk_strip.get_selection().get_selected_rows()[1]

        if m_control :
            if button == 1:
                for row in selected_rows :
                    l = self.spronk_buffer.get(self.spronk_buffer.get_iter(row),4)[0]
                    if l.isdigit() :
                        l = int(l)

                    if l != "" :
                        if self.current_slide.get_for_spronk(l,self.current_widget,'style.opacity') is None:
                            self.current_slide.set_for_spronk(l,self.current_widget,'style.opacity','0.2')
                        else :
                            self.current_slide.unset_for_spronk(l,self.current_widget,'style.opacity')

                    if l == 'overview' :
                        self.do_layout_update(None)
                self.update_spronk_strip()
        else :
            if button == 3:
                row = selected_rows[0]
                l = self.spronk_buffer.get(self.spronk_buffer.get_iter(row),4)[0]
                if l.isdigit() :
                    l = int(l)
                self.current_slide.spronk_to_action_panel(l)

    def do_spronk_strip_row_activated(self, spronk_strip, path, column) :
        if self.current_slide is None :
            return
        spronks = self.current_slide.spronks

        l = self.spronk_buffer.get(self.spronk_buffer.get_iter(path),4)[0]
        if l.isdigit() :
            l = int(l)

        if self.old_old_sel is not None :
            selected_rows = self.old_old_sel[1]
        else :
            selected_rows = spronk_strip.get_selection().get_selected_rows()[1]

        if len(selected_rows) == 1 and path == Gtk.TreePath(0) :
            spronks.insert(0, copy.deepcopy(spronks[0]))
        elif len(selected_rows) == 1 and path == Gtk.TreePath(len(spronks)+1):
            spronks.insert(len(spronks), copy.deepcopy(spronks[-1]))
        elif self.current_widget is not None :
            #self.current_slide.remove_from_spronks(self.current_widget.show)
            #self.current_slide.remove_from_spronks(self.current_widget.hide)

            if isinstance(l, int) :
                for n in range(0, len(self.current_slide.spronks)) :
                    if ((Gtk.TreePath(n+1) in selected_rows)) :
                        self.current_slide.set_for_spronk(n,self.current_widget,'style.visibility','visible')
                    else :
                        self.current_slide.set_for_spronk(n,self.current_widget,'style.visibility','hidden')
            else :
                visibility = self.current_slide.get_for_spronk(l,self.current_widget,'style.visibility')
                if visibility == 'visible' :
                    self.current_slide.set_for_spronk(l,self.current_widget,'style.visibility','hidden')
                else :
                    self.current_slide.set_for_spronk(l,self.current_widget,'style.visibility','visible')

            if l == 'overview' :
                self.do_layout_update(None)

        self.update_spronk_strip()

    def do_widget_size_allocate(self, widget, req) :
        line = widget.line
        line[2] = self.layout.translate_dist(req.height, rev=True)

    line_max = 5
    def insert_widget(self, ty, slide_num=None, line=None) :
        if slide_num is not None :
            slide = self.slides[slide_num]
        else :
            slide = self.current_slide

        if self.template is None or slide is None :
            return
        w = self.template['body']['width']
        h = self.template['body']['height']

        widget = widget_types[ty][2](slide,
                                     self.layout,
                                     env=self.get_aenv())
        #for sw in widget.subwidgets :
        #    sw.modify_bg(Gtk.StateType.NORMAL,
        #                 Gdk.Color(self.template['params']['bgcolor']))
        self.absorb_properties(widget, as_self=False)
        #design_widget = widget.subwidgets[GLOSSER_WIDGET_DESIGN]
        #design_widget.set_property('width-request', int(self.layout.translate_dist(w)))
        line_height = h/self.line_max

        if line is None :
            line = self.current_line

        if line < len(slide.lines) and \
           slide.lines[line] is not None and \
           slide.lines[line][1].container :
            cntr = slide.lines[line][1]
            cntr.append(widget)
        else :
            slide.insert_line(line,
                                       ty,
                                       widget,
                                       line_height,
                                       'left')
            widget.line = slide.lines[line]
        self.set_slide()
        self.layout.grab_focus()
        self.change_current_widget(widget)

        #self.do_widget_size_allocate(widget, design_widget.allocation)
        #self.do_widget_pos_update(self, None)
        #design_widget.grab_focus()

        return widget

    def new_slide(self, num, variant, insert=True) :
        num = int(num)
        for i in range(len(self.slides), num+1) :
            self.slides.append(None)
        slide = LanternSlide(num,
                                             variant=variant,
                                             reload_slide=self.set_slide,
                                             variant_lsto=self.variant_lsto,
                                             deployment_manifest=self.deployment_manifest,
                                             env=self.get_aenv())
        if insert and self.slides[num] is not None :
            self.slides.insert(num, slide)
        else :
            self.slides[num] = slide

        self.absorb_properties(self.slides[num], as_self=False)
        self.slides[num].reload()
        self.set_slide(num)
        debug_print(self.slides)
        #self.presentation_instance.set_slide(num)
        self._update_slide_list_from_slides()
        self.update_slide_thumb(num, self.slide_list_trvw.queue_draw)
        return self.slides[num]

    def _update_slide_list_from_slides(self) :
        self.slide_list.clear()
        for slide in self.slides :
            if slide is None :
                num = -1
            else :
                num = slide.num+1
            self.slide_list.append((num, slide))

    def set_slide(self, n=None, change_current=True, xml=False) :

        #self.presentation_instance.set_slide(n,
        #                                         change_current=change_current)
        LanternBasic.set_slide(self, n=n, change_current=change_current)

        if change_current :
            self.update_spronk_strip()
            self.change_current_line()

    def run_chooser(self, env = None) :
        chooser = Gtk.FileChooserDialog(title="Export PDF",
                                        parent=self.get_toplevel(),
                                        action=Gtk.FileChooserAction.SAVE,
                                        buttons=(Gtk.STOCK_CANCEL,Gtk.ResponseType.CANCEL,Gtk.STOCK_SAVE,Gtk.ResponseType.OK))

        chooser.set_default_response(Gtk.ResponseType.OK)

        resp = chooser.run()

        if resp == Gtk.ResponseType.OK :
            filename = chooser.get_filename()
            chooser.destroy()
        else :
            chooser.destroy()
            return
        
        self.render_to_pdf(filename)

    def render_to_pdf(self, filename) :
        pdfsurface = cairo.PDFSurface(filename,
                                   int(self.template['width']),
                                   int(self.template['height']))
        pdf_cr = cairo.Context(pdfsurface)

        for slide in self.slides :
            self.set_template_variant_xml(slide.variant)
            xml = copy.deepcopy(self.current_xml)
            #self.presentation_instance.set_slide(n=self.slides.index(slide),
            #               xml=xml,
            #               change_current=False)
            slide.pre_spronk()

            for n in range(0, len(slide.spronks)+1) :
                pdf_cr.save()
                svg = StringIO.StringIO()
                xml.write(svg)
                handle = rsvg.Handle()
                handle.write(svg.getvalue())
                handle.close()

                handle.render_cairo(pdf_cr)
                pdf_cr.translate(self.template["body"]["x"],
                                   self.template["body"]["y"])
                slide.render_widgets_to_cairo(pdf_cr, ignore_spronks=False)
                pdf_cr.restore()
                pdf_cr.show_page()

                if n < len(slide.spronks) :
                    slide.execute_spronk(n)

    thumb_rat = .2
    def redraw(self, slide=None, xml=None, change_current=True, thumb_only=False) :

        #if not thumb_only :
        #    self.presentation_instance.redraw(slide, xml, change_current=False)

        if slide is None :
            slide = self.current_slide
        if xml is None :
            self.set_template_variant_xml(slide.variant)
            xml = self.current_xml

        if xml is not None and \
                slide is not None :
            svg = StringIO.StringIO()
            xml.write(svg)
            #self.layout.set_svg(svg.getvalue())
            #self.layout.update_text(slide._filled_text)
            handle = rsvg.Handle()
            handle.write(self.layout.get_svg())
            handle.close()

            if change_current and not thumb_only :
                self.layout.width = self.template['width']
                self.layout.height = self.template['height']
                self.layout.handle = handle
                self.layout.line_num = self.current_line
                self.layout.lines = slide.lines
                self.layout.current_widget = self.current_widget
                #if self.current_line < len(slide.lines) and slide.lines[self.current_line] is not None :
                #    ty = slide.lines[self.current_line][0]
                #    self.layout.line_indicators = \
                #        [ widget_types[ty][3](self.get_style())]
                #    if self.current_widget != slide.lines[self.current_line][1]:
                #        ty = self.current_widget.ty
                #        self.layout.line_indicators.append(widget_types[ty][3](self.get_style()))
                #else :
                #    self.layout.line_indicators = ( u'\u273c', )
                self.layout.set_foreground_location(self.template['body'])

            self.slide_list_trvw.queue_draw()
            self.layout.queue_draw()

    def _slide_to_action_panel(self) :
        if self.current_slide is None :
            return
        self.get_aenv().action_panel.to_action_panel(self.current_slide.action_panel)

    def _remake_slide_menu(self) :
        children = self.new_slide_menu.get_children()
        for r in range(1, len(children)) :
            self.new_slide_menu.remove(children[r])

        if self.template is None or 'variants' not in self.template :
            return

        #FIXME: migrate to UI manager
        for variant in self.template['variants'] :
            variant_settings = self.template['variants'][variant]
            new_variant_butt = Gtk.MenuItem()
            new_variant_butt.remove(new_variant_butt.get_child())
            new_variant_butt.add(self.template['variants'][variant]['sample'])
            new_variant_butt.set_tooltip_text(variant_settings['title'])
            new_variant_butt.connect('activate',
                                     variant_settings['new_slide'])
            self.new_slide_menu.append(new_variant_butt)

        self.new_slide_menu.show_all()

    def aes_add_a(self, aname_root, **parameters) :
        if aname_root == 'LanternSlide' :
            return self.new_slide(parameters['num'], 'basic', insert=False)
        elif aname_root in aname_root_to_ty :
            return self.insert_widget(aname_root_to_ty[aname_root],
                                      slide_num=int(parameters["on_slide"]),
                                      line=int(parameters["line"]))

        return aobject.AObject.aes_add_a(self, aname_root, **parameters)

    def get_method_window(self) :
        vbox = Gtk.VBox()

        butt_hbox = Gtk.HBox()
        new_slide_butt = Gtk.ToggleButton()
        new_slide_butt.set_image(Gtk.Image.new_from_stock(Gtk.STOCK_NEW,
                                                    Gtk.IconSize.BUTTON))

        new_slide_butt.set_tooltip_text("New slide")
        new_slide_menu = Gtk.Menu()
        new_slide_menu.attach_to_widget(new_slide_butt, None)
        new_slide_butt.connect("button_press_event", lambda w, e :\
            new_slide_menu.popup(None, None, None, None, e.button, e.time) \
                if not w.get_active() else \
            new_slide_menu.popdown())

        new_slide_menu.set_title('New Slide')
        self.new_slide_menu = new_slide_menu
        new_slide_menu.modify_bg(Gtk.StateType.NORMAL, Gdk.Color(65535, 65535, 65535))

        new_slide_labl = Gtk.Label()
        new_slide_labl.set_markup('<b>New...</b>')
        new_slide_meni = Gtk.MenuItem()
        new_slide_meni.remove(new_slide_meni.get_child())
        new_slide_meni.add(new_slide_labl)
        new_slide_menu.append(new_slide_meni)

        self._remake_slide_menu()

        settings_butt = Gtk.Button()
        settings_butt.set_image(Gtk.Image.new_from_stock(Gtk.STOCK_PREFERENCES,
                                                    Gtk.IconSize.BUTTON))
        slide_settings_butt = Gtk.Button()
        slide_settings_butt.set_image(Gtk.Image.new_from_stock(Gtk.STOCK_PAGE_SETUP,
                                                    Gtk.IconSize.BUTTON))

        if self.get_aenv() is not None and \
                self.get_aenv().action_panel is not None :
            settings_butt.connect('clicked', lambda b : \
                self.get_aenv().action_panel.to_action_panel(self.action_panel))
            slide_settings_butt.connect('clicked', lambda b : \
                self._slide_to_action_panel())
        else :
            win = Gtk.Window()
            win.add(action_panel)
            win.hide()
            settings_butt.connect('clicked', lambda b : win.show())
        butt_hbox.pack_start(settings_butt, False, False, 0)
        butt_hbox.pack_start(slide_settings_butt, False, False, 0)

        butt_hbox.pack_start(new_slide_butt, False, False, 0)

        vbox.pack_start(butt_hbox, False, False, 0)
        vbox.pack_start(Gtk.Label('Slides'), True, True, 0)
        slide_list_trvw = Gtk.TreeView(self.slide_list)
        slide_list_trvw.connect("drag-end",
                                self.do_slide_list_drag_end)
        slide_list_trvw.set_reorderable(True)
        slide_list_trvw.set_can_focus(True)
        slide_list_cllr = LanternSlideCellRenderer()
        slide_list_num_cllr = Gtk.CellRendererText()
        slide_list_num_tvcl = Gtk.TreeViewColumn('#', slide_list_num_cllr)
        slide_list_num_tvcl.add_attribute(slide_list_num_cllr, 'text', 0)
        slide_list_tvcl = Gtk.TreeViewColumn('Slide', slide_list_cllr)
        slide_list_tvcl.add_attribute(slide_list_cllr, 'slide', 1)
        slide_list_trvw.append_column(slide_list_num_tvcl)
        slide_list_trvw.append_column(slide_list_tvcl)
        slide_list_scrw = Gtk.ScrolledWindow()
        slide_list_scrw.add_with_viewport(slide_list_trvw)
        slide_list_scrw.set_size_request(-1, 300)
        self.slide_list_trvw = slide_list_trvw
        slide_list_cllr.template = self.template
        slide_list_cllr.lantern = self

        slide_list_trvw.connect('row-activated', self._set_slide_from_trvw)

        vbox.pack_start(slide_list_scrw, True, True, 0)
        vbox.show_all()
        return vbox

    def _set_slide_from_trvw(self, tv, pa, col) :
        it = self.slide_list.get_iter(pa)
        self.set_slide(self.slide_list.get_value(it, 0)-1)

    def _make_new_slide_lambda(self, variant) :
        return lambda c: self.new_slide(len(self.slides), variant)

    def set_template(self, template_name=None, variant='basic') :
        if template_name is None :
            template_name = self.default_template

        self.template['name'] = template_name

        self.template['tar'] = tarfile.open(\
            get_share_location()+'templates/lantern/'+template_name+'.tgz', 'r')
        self.template['params_xml'] = \
            etree.parse(self.template['tar'].extractfile('parameters.xml'))

        self.template['params'] = {}
        params = ['bgcolor']
        root = self.template['params_xml'].getroot()

        param_root = root.find('params')
        for p in params :
            node = param_root.find(p)
            if node is not None :
                self.template['params'][p] = node.text.strip()

        self.variant_lsto.clear()

        variants_root = root.find('variants')
        self.template['variants'] = {}

        for variant_node in variants_root :
            dets = {
                'title' : variant_node.get('title'),
                'filename' : variant_node.text.strip(),
                'new_slide' : self._make_new_slide_lambda(variant_node.tag),
                'fgstyle' : '',
            }

            debug_print(variant_node.get('fgstyle'))
            if variant_node.get('fgstyle') is not None :
                dets['fgstyle'] = variant_node.get('fgstyle')

            self.template['variants'][variant_node.tag] = dets
            var_dict = self.template['variants'][variant_node.tag]
            fobj = var_dict['filename']
            fobj = self.template['tar'].extractfile(fobj)
            file_contents = fobj.read()

            var_dict['xml'] =\
                etree.ElementTree(etree.fromstring(file_contents))

            loader = GdkPixbuf.PixbufLoader.new_with_type("svg")
            loader.write(file_contents)
            loader.set_size(50, 50)
            loader.close()

            var_dict['sample'] =\
                    Gtk.Image.new_from_pixbuf(loader.get_pixbuf())

            self.variant_lsto.append((variant_node.tag,
                                      var_dict['title']))

        if not self.presentation :
            self._remake_slide_menu()

        basic_xml = self.template['variants']['basic']['xml']

        w = basic_xml.getroot().get('width')
        self.template['width'] = _px_to_float(w)
        h = basic_xml.getroot().get('height')
        self.template['height'] = _px_to_float(h)

        self.set_template_variant_xml(variant, force=True)

        # Redraw thumbs
        for n in range(0, len(self.slides)) :
            self.update_slide_thumb(n, self.slide_list_trvw.queue_draw)
            #self.set_slide(n, change_current=False)

        # Update slide in view
        #self.set_slide()

all_templates = {}
template_files = os.listdir(get_share_location()+'templates/lantern/')
template_lsto = Gtk.ListStore(str, str, str)

for t in template_files :
    if t.endswith('.tgz') :
        template_name = t[:-4]
        all_templates[template_name] = {}

        tar = tarfile.open(\
            get_share_location()+'templates/lantern/'+template_name+'.tgz', 'r')
        params_xml = \
            etree.parse(tar.extractfile('parameters.xml'))

        root = params_xml.getroot()
        param_root = root.find('params')
        for node in param_root :
            all_templates[template_name][node.tag] = node.text.strip()

        template_lsto.append((template_name,
                              all_templates[template_name]['title'],
                              all_templates[template_name]['description'],))

aobject.aname_root_catalog['Lantern'] = Lantern
