from sympy.core.function import Lambda as aes_Lambda
from aobject import aobject
from VTK import SourceVTK, VTK_factory
from CSV import CSV, CSV_factory
import lxml.etree as ET
from gi.repository import Gtk
from gi.repository import Gdk
import cairo
import os
from gi.repository import GObject
from aobject import details
from aobject.utils import debug_print
from .. import glypher
from Source import *

class SourceImporter(Gtk.FileChooserDialog) :
    last_dir = os.path.expanduser('~')

    def __init__(self) :
        Gtk.FileChooserDialog.__init__(self,
                      title="Import Source", action=Gtk.FileChooserAction.OPEN,
                      buttons=(Gtk.STOCK_CANCEL,Gtk.ResponseType.CANCEL,Gtk.STOCK_OPEN,Gtk.ResponseType.OK))
        self.set_local_only(False)

    @classmethod
    def run_chooser(cls, env = None) :
        chooser = cls()
        if cls.last_dir is not None :
            chooser.set_current_folder(cls.last_dir)
        chooser.set_default_response(Gtk.ResponseType.OK)
        chooser.set_select_multiple(True)

        all_ffil = Gtk.FileFilter()
        all_ffil.set_name("All files")
        all_ffil.add_pattern("*")
        chooser.add_filter(all_ffil)

        aes_ffil = Gtk.FileFilter()
        aes_ffil.set_name("CSV")
        aes_ffil.add_mime_type("text/csv")
        aes_ffil.add_pattern("*.csv")
        chooser.add_filter(aes_ffil)

        extra_hbox = Gtk.HBox()
        #extra_hbox.pack_start(Gtk.Label("Name as", True, True, 0), False)
        #extra_entr = Gtk.Entry()
        #extra_entr.set_activates_default(True)
        extra_chkb = Gtk.CheckButton("Regex")
        extra_chkb.set_active(False)
        extra_hbox.pack_start(extra_chkb, False, False, 0)
        extra_cmbe = Gtk.ComboBoxText()
        extra_hbox.pack_start(extra_cmbe, True, True, 0)
        extra_hbox.show_all()
        chooser.set_extra_widget(extra_hbox)

        resp = chooser.run()

        #new_name = extra_entr.get_text()
        #if len(new_name) == 0:
        #    new_name = None

        cls.last_dir = chooser.get_current_folder()

        if resp == Gtk.ResponseType.OK :
            if extra_chkb.get_active() :
                debug_print(extra_cmbe.get_active_text())
            else :
                uris = chooser.get_uris()
            chooser.destroy()
        else :
            chooser.destroy()
            return

        for uri in uris :
            load_from_file(uri, None, env)

def parse_params(pfn, header = True):
    pfile = open(pfn, 'r')
    params = {}
    for line in pfile:
        if cmp(line[0],'=') == 0:
         if header :
          header = False
          continue
         else :
          break
        if header: continue
        commpos = line.find('#')
        if commpos > 0 : line = line[0:commpos-1]
        line = line.strip()
        if commpos == 0 or len(line) == 0: continue
        var,val = line.split('=')
        var = var.strip(); val = val.strip()
        params[var] = val
    return params

class SourceDictionary(GObject.GObject) :
    def __init__(self) :
        GObject.GObject.__init__(self)
        dic = aobject.get_object_dictionary()
        dic.connect("add", self.check_source)
        dic.connect("remove", self.check_source)
    def check_source(self, ob, n, r) :
        ob = aobject.get_object_from_dictionary(n)
        if ob == None : ob = aobject.get_removing_from_dictionary(n)
        if ob == None or ob.am("Source") :
            self.emit("aesthete-source-change", n, r)

source_dictionary = SourceDictionary()
def get_source_dictionary() : return source_dictionary

def do_source_rename(dlog, resp) :
    if resp == Gtk.ResponseType.ACCEPT :
        dlog.object.set_aname_nice(dlog.get_text())

def make_source_rename_dialog(parent, object) :
    rename_dlog = Gtk.Dialog("Rename source", parent, 0,\
        (Gtk.STOCK_CANCEL    , Gtk.ResponseType.REJECT,\
         Gtk.STOCK_OK        , Gtk.ResponseType.ACCEPT))
    rename_hbox = Gtk.HBox()
    rename_dlog.vbox.pack_start(rename_hbox, True, True, 0)
    rename_hbox.pack_start(Gtk.Label("Name", True, True, 0), False, False, 15)
    rename_entr = Gtk.Entry()
    rename_entr.set_text(object.get_aname_nice())
    rename_hbox.pack_start(rename_entr, True, True, 0)
    rename_dlog.get_text = rename_entr.get_text
    rename_dlog.object = object
    rename_dlog.connect("response", do_source_rename)
    rename_dlog.show_all()
    rename_dlog.run()
    rename_dlog.destroy()

def do_source_treeview_popup(trvw, event) :
    if event.button != 3 : return
    item = trvw.get_path_at_pos(int(event.x), int(event.y))
    if item == None : return
    time = event.time

    context_menu = Gtk.Menu()
    obj = aobject.get_object_from_dictionary(trvw.get_model()[item[0]][0])

    context_menu_aname_nice = Gtk.MenuItem(obj.get_aname_nice())
    context_menu_aname_nice.set_sensitive(False)
    context_menu.add(context_menu_aname_nice)
    
    context_menu_rename = Gtk.MenuItem("Rename")
    context_menu_rename.connect("button-press-event", lambda o, e : make_source_rename_dialog(trvw.get_toplevel(), obj))
    context_menu.add(context_menu_rename)

    context_menu_rename = Gtk.MenuItem("Remove")
    context_menu_rename.connect("button-press-event",
                                lambda o, e : obj.aes_remove())
    context_menu.add(context_menu_rename)

    if (obj.reloadable) :
        context_menu_reload = Gtk.MenuItem("Reload")
        context_menu_reload.connect("button-press-event", lambda o, e : obj.source_reload())
        context_menu.add(context_menu_reload)

    context_menu_settings = Gtk.MenuItem("Settings...")
    context_menu_settings.connect("button-press-event", lambda o, e :\
                                obj.env.action_panel.to_action_panel(obj.get_action_panel()))
    context_menu.add(context_menu_settings)

    context_menu.popup( None, None, None, None, event.button, time )
    context_menu.show_all()

connected = []
_entry_server = None

def init_entry_server() :
    global _entry_server
    if _entry_server is None :
        _entry_server = glypher.Widget.GlyphEntry(position=(0,0))
        root = glypher.Word.make_word('hi',
                                                     _entry_server.main_phrase).get_xml()
        xml = ET.ElementTree(root)
        _entry_server.set_xml(xml)
        _entry_server.main_phrase.set_is_caching(True)
        _entry_server.main_phrase.set_font_size(5)
        _entry_server.main_phrase.background_colour = (0.3,0.5,0.4)

def status_cdf(c, cr, tm, it, data) :
    obj = aobject.get_object_from_dictionary(tm.get_value(it,0))
    if obj is None : return
    cr.set_property('status', obj.source_get_status())

def cdf(c, cr, tm, it, data) :
    obj = aobject.get_object_from_dictionary(tm.get_value(it,0))
    if obj is None : return
    #cr.set_property('xml', tm.get_value(it, 2))
    #cr.set_property('text', tm.get_value(it, 1))
    cr.set_property('obj', obj)
    #cr.set_property('foreground', 'red' if obj.is_needs_reloaded() else 'black')

    #if obj.get_aname_xml() is None :
    #    root = glypher.Word.make_word(obj.get_aname_nice(),
    #                                                 _entry_server.main_phrase).get_xml()
    #else :
    #    root = obj.get_aname_xml()
    #xml = ET.ElementTree(root)
    ##_entry_server.set_xml(xml)

    #_entry_server.main_phrase.draw()
    #cis = _entry_server.main_phrase.cairo_cache_image_surface
    #pixbuf = \
    #    GdkPixbuf.Pixbuf.new_from_data(cis.get_data(),
    #                                 GdkPixbuf.Colorspace.RGB,
    #                                 True,
    #                                 8,
    #                                 cis.get_width(),
    #                                 cis.get_height(),
    #                                 cis.get_stride())

    #cr.set_property('pixbuf', pixbuf)
def on_needs_reloaded_status_change(ob, val, tm, i, tv, cr) :
    tv.get_column(1).set_cell_data_func(cr, cdf)
    tv.queue_draw()

def connect_sources_reload_signal(tm, p, i, tv, cr) :
    obj = aobject.get_object_from_dictionary(tm.get_value(i,0))
    if obj.get_aname() in connected : return
    obj.connect("aes_needs_reloaded_status_change", on_needs_reloaded_status_change, tm, i, tv, cr)
    connected.append(obj.get_aname())

sources_status_cell_colours = {\
        SOURCES_STATUS_NORMAL : (.8,.8,.8),
        SOURCES_STATUS_NEEDS_RELOADED : (1.,0.,0.),
        SOURCES_STATUS_UNAVAILABLE : (.4,.4,.4),
                              }

class SourcesStatusCellRenderer(Gtk.CellRenderer) :
    __gproperties__ = {
        "status": (GObject.TYPE_INT, "status", "status",0,2,
                    SOURCES_STATUS_NORMAL, GObject.PARAM_READWRITE),
    }

    def do_set_property(self, pspec, value) :
        setattr(self, pspec.name, value)

    def do_get_property(self, pspec) :
        return getattr(self, pspec.name)

    def do_render(self, cr, widget, background_area, cell_area,
                  flags) :
        surface = cr.get_target()
        cr = cairo.Context(surface)
        
        cr.save()
        cr.translate(cell_area.x, cell_area.y)

        cr.rectangle(0, 0, cell_area.width, cell_area.height)
        colour = Gdk.Color(\
            *[65535*i for i in sources_status_cell_colours[self.status]])
        colour_list = [colour.red,
                       colour.green,
                       colour.blue,
                       1.]

        colour_list[3] = .5
        cr.set_source_rgba(*colour_list)
        cr.fill()
        colour_list[3] = 1.

        if self.status == SOURCES_STATUS_NEEDS_RELOADED :
            cr.move_to(cell_area.width/2, cell_area.height)
            cr.rel_line_to(-10, 0)
            cr.rel_line_to(10, -cell_area.height)
            cr.rel_line_to(10, cell_area.height)
            cr.close_path()
            colour_list[3] = 1.
            cr.set_source_rgba(*colour_list)
            cr.fill()
        else :
            cr.rectangle(cell_area.width/2-10, 0, 20, cell_area.height)
            cr.set_source_rgba(*colour_list)
            cr.fill()

        cr.restore()

    line_height = 100
    def do_get_size(self, widget, cell_area=None):
        x = 0
        y = 0

        if cell_area is not None :
            x = cell_area.x
            y = cell_area.y

        w = 30
        h = 10

        return (x, y, w, h)

GObject.type_register(SourcesStatusCellRenderer)

def _source_trvw_on_cursor_changed(trvw, sources_lsst) :
    cursor = trvw.get_cursor()[0]
    if cursor is not None :
        value = sources_lsst.get_value(sources_lsst.get_iter(trvw.get_cursor()[0]), 0)
        aobject.get_object_dictionary().selected_source_change(value)

def make_source_treeview() :
    init_entry_server()
    sources_lsst = aobject.get_object_dictionary().get_liststore_by_am('Source')
    sources_trvw = Gtk.TreeView(sources_lsst)
    sources_trvw.set_can_focus(True)

    status_trvc = Gtk.TreeViewColumn()
    status_cllr = SourcesStatusCellRenderer()
    status_trvc.pack_start(status_cllr, True)
    status_trvc.set_cell_data_func(status_cllr, status_cdf)
    sources_trvw.append_column(status_trvc)
    sources_trvw.set_size_request(150, -1)

    sources_trvc = Gtk.TreeViewColumn('Sources'); sources_trvw.append_column(sources_trvc)
    #sources_cllr = Gtk.CellRendererPixbuf(); sources_trvc.pack_start(sources_cllr, True)
    sources_cllr = glypher.Widget.GlyphCellRenderer(); sources_trvc.pack_start(sources_cllr, True)
    #sources_trvc.set_attributes(sources_cllr, text=1)
    sources_trvc.set_cell_data_func(sources_cllr, cdf)
    sources_trvw.connect("button-press-event", do_source_treeview_popup)
    sources_trvw.connect("row-activated", lambda t, p, c : \
                         aobject.get_object_dictionary().try_active_source_action())
    sources_trvw.connect("cursor-changed", _source_trvw_on_cursor_changed, sources_lsst)
    sources_trvw.set_tooltip_column(3)
    sources_lsst.connect("row-changed", connect_sources_reload_signal, sources_trvw, sources_cllr)
    sources_lsst.connect("row-changed", \
            lambda tm, p, i : sources_trvw.set_cursor(p, sources_trvc))
    return sources_trvw
def reload_sources(ness = True) :
    d = aobject.get_object_dictionary().dictionary
    for key in d :
        obj = d[key]
        if obj.am('Source') and obj.reloadable and ( not ness or obj.is_needs_reloaded() ) :
            debug_print("Reloading", obj.get_aname_nice())
            obj.source_reload()

def load_from_file(uri, new_name, env) :
    #if filename.startswith('file://') :
    #    filename = filename[7:]

    suff = uri.split('.')
    suff = suff[-1].lower() if len(suff) > 1 else None

    if suff == 'csv' :
        source = CSV_factory(uri, env)
        mime_type = 'text/csv'
    elif suff == 'vtu' :
        source = VTK_factory(uri, env)
        mime_type = 'text/csv'
    else :
        raise RuntimeError('Unrecognized file type : ' + uri)

    if source is None :
        return

    #if new_name is not None :
    #    source.set_aname_nice(new_name)

    recman = Gtk.RecentManager.get_default()
    #if not filename.startswith('file://') :
    #    uri = 'file://'+filename
    #else :
    #    uri = filename

    #recman.add_full(uri, {'mime_type' : mime_type,
    #                           'app_name' : details.get_name(),
    #                           'app_exec' : details.get_command(),
    #                           'display_name' : source.get_aname_nice(),
    #                           'description' : 'Aesthete imported source',
    #                           'is_private' : False,
    #                           'groups' : ['AestheteSrc']})

GObject.signal_new("aesthete-source-change", SourceDictionary, GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_STRING, GObject.TYPE_STRING))
