import aesthete.paths as paths
paths.setup("aesthete")

import codecs
import os
import subprocess
import argparse
import lxml.etree as ET
from aesthete import details
from xml.parsers.expat import ExpatError
import time
import re

def wrap_wiki_text(ctime, wikiname, text) :
    text = text.replace('\n', '%0a')

    return """version=pmwiki-2.2.27 ordered=1 urlencoded=1
agent=Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.9.2.24)
Gecko/20111107 Ubuntu/10.04 (lucid) Firefox/3.6.24
author=
charset=UTF-8
csum=auto-generated content
ctime=%d
host=localhost
name=%s
rev=1
targets=
text=%s
""" % (ctime, wikiname, text)

def process_code(member, code) :
    return re.sub(\
       r"['\"]? *\+? *self\[['\"]([^\"']*)['\"]].%s\(\) *\+? *u?(nicode)?['\"]?" %member,
       " ''\\1'' ", code)
phrasegroups = os.listdir('../share/defs')
cats = {}
for u in phrasegroups :
    try :
        filename = "../share/defs/%s"%u
        tree = ET.parse(filename)

        root = tree.getroot()
        name = root.tag.lower()
        print '\t%s : %s' % (name, u)

        wikiname = 'Manual.Phrasegroups%s' % name
        f = codecs.open('wiki/%s' % wikiname, 'w', encoding='utf-8')

        ctime = int(time.time())

        string_dict = {'name':name}

        targets = root.find('targets')

        if targets is not None :
            targettext = ''
            for target in targets :
                targetname = target.get('name')
                targettext += ":''%s'':" % targetname
                if targetname == root.get('lhs') :
                    targettext += " '''[LHS]'''"
                if targetname == root.get('rhs') :
                    targettext += " '''[RHS]'''"
                targettext += '%0a'
            string_dict['targettext'] = targettext
        else :
            string_dict['targettext'] = ''

        string_code = root.find("string")
        if string_code is not None : 
            string_dict['stringtext'] = process_code('to_string',
                                                     string_code.text.strip())
        else :
            string_dict['stringtext'] = "'''None'''"

        latex_code = root.find("latex")
        if latex_code is not None : 
            string_dict['latextext'] = process_code('to_latex',
                                                     latex_code.text.strip())
        else :
            string_dict['latextext'] = "'''None'''"

        sympy_code = root.find("sympy")
        if sympy_code is not None : 
            string_dict['sympytext'] = process_code('get_sympy',
                                                     sympy_code.text.strip())
        else :
            string_dict['sympytext'] = "'''None'''"

        #latex_code = root.find("latex")
        ##FIXME: it strikes me that this isn't very safe
        #if latex_code is not None : 
        #    new_phrasegroup.to_latex_code = compile(latex_code.text.strip(), '<string>', 'eval')
        #info_text = root.find("info")
        #if info_text is not None :
        #    if top :
        #        new_phrasegroup.indicate_info = True
        #    new_phrasegroup.set_info_text(info_text.text.strip())
        title_text = root.get("title")
        if title_text is not None :
            string_dict['title'] = title_text
        else :
            string_dict['title'] = ''

        type_text = root.get("type")
        if type_text is not None and type_text != 'phrasegroup' :
            string_dict['type'] = type_text
        else :
            string_dict['type'] = ''

        info_text = root.find("info")
        if info_text is not None :
            string_dict['info'] = info_text.text.strip()
        else :
            string_dict['info'] = ''

        wiki_text = root.find("wiki")
        if wiki_text is not None :
            string_dict['wiki'] = wiki_text.text.strip()
        else :
            string_dict['wiki'] = ''

        alternatives_text = root.find("alternatives")
        if alternatives_text is not None :
            string_dict['altcat'] = alternatives_text.text.strip()
        else :
            string_dict['altcat'] = ''

        symbol_text = root.find("symbol")
        if symbol_text is not None :
            string_dict['symbol'] = symbol_text.text.strip()
        else :
            string_dict['symbol'] = ''

        category_text = root.find("category")
        if category_text is not None :
            string_dict['category'] = category_text.text.strip()
            if string_dict['category'] not in cats :
                cats[string_dict['category']] = []
            cats[string_dict['category']].append((name, wikiname,
                                                  string_dict['symbol']))
        else :
            string_dict['category'] = ''

        latex_text = root.get("latex_name")
        if latex_text is not None :
            string_dict['texcut'] = latex_text
        else :
            string_dict['texcut'] = ''

        shortcut_text = root.find("shortcut")
        if shortcut_text is not None :
            string_dict['shortcut'] = shortcut_text.text.strip()
        else :
            string_dict['shortcut'] = ''

        for key in string_dict.keys() :
            string_dict[key] = string_dict[key].encode('ascii',
                                                       'xmlcharrefreplace')
        text = '''
Return to [[Manual.Phrasegroups|Phasegroups list]] or [[Manual]].

! {name}

||border="1" style="border-collapse:collapse" cellpadding="5" width=66%
||Description ||Name       ||Symbol || Category || Shortcut || Wikipedia ||
||{title} ||{name} || {symbol} || {category} || @@{shortcut}@@ || [[http://en.wikipedia.org/wiki/{wiki}|{wiki}]] ||

||border="1" style="border-collapse:collapse" cellpadding="5" width=66%
||Since || TeX shortcut || Alternatives Category || Definition || Type ||
|| 0.4 || {texcut} || {altcat} || XML || {type} ||

!!!! Image

%width:50% Attach:Main./aesthete-phrasegroup-{name}.svg

!!!! Description

{info}

!!!! Targets

{targettext}

!!!! LaTeX

@@ {latextext} @@

!!!! String

@@ {stringtext} @@

!!!! Sympy

@@ {sympytext} @@
'''.format(**string_dict)

        
        f.write(wrap_wiki_text(ctime, wikiname, text))
        f.close()
    except (ExpatError, ET.XMLSyntaxError) as e :
        print "WARNING : Could not parse PG '%s', continuing without: %s" % (u, e)

ctime = int(time.time())
f = codecs.open('wiki/Manual.Phrasegroups', 'w', encoding='utf-8')
fstr = ""
for cat in cats :
    fstr += u"* %s\n" % cat
    for pg in cats[cat] :
        fstr += u"** '''%s''' - [[%s|%s]]\n" % (pg[2], pg[1], pg[0])
fstr = fstr.encode('ascii', 'xmlcharrefreplace')
f.write(wrap_wiki_text(ctime, 'Manual.Phrasegroups', fstr))
f.close()
